@@ -56,7 +56,7 @@ static __inline__ int set_bit(int nr, unsigned long *addr)
 }
 
 /* clear bit nr in bitmap addr; returns previous value of bit nr*/
-static __inline__ int clear_bit(int nr, unsigned long *addr)
+static __inline__ int test_and_clear_bit(int nr, unsigned long *addr)
 {
         unsigned long mask;
 
@@ -67,6 +67,8 @@ static __inline__ int clear_bit(int nr, unsigned long *addr)
         return nr;
 }
 
+#define clear_bit(n, a) test_and_clear_bit(n, a)
+
 static __inline__ int test_bit(int nr, const unsigned long *addr)
 {
         return ((1UL << (nr & (BITS_PER_LONG - 1))) &
