@@ -30,7 +30,8 @@
 #define LNET_DEV_MAJOR 10
 #define LNET_DEV_MINOR 240
 #define OBD_DEV_ID 1
-#define OBD_DEV_PATH "/dev/obd"
+#define OBD_DEV_NAME "obd"
+#define OBD_DEV_PATH "/dev/" OBD_DEV_NAME
 #define OBD_DEV_MAJOR 10
 #define OBD_DEV_MINOR 241
 #define SMFS_DEV_ID  2
