@@ -1,23 +1,39 @@
 /* -*- mode: c; c-basic-offset: 8; indent-tabs-mode: nil; -*-
  * vim:expandtab:shiftwidth=8:tabstop=8:
  *
+ * GPL HEADER START
+ *
+ * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
+ *
+ * This program is free software; you can redistribute it and/or modify
+ * it under the terms of the GNU General Public License version 2 only,
+ * as published by the Free Software Foundation.
+ *
+ * This program is distributed in the hope that it will be useful, but
+ * WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+ * General Public License version 2 for more details (a copy is included
+ * in the LICENSE file that accompanied this code).
+ *
+ * You should have received a copy of the GNU General Public License
+ * version 2 along with this program; If not, see
+ * http://www.sun.com/software/products/lustre/docs/GPLv2.pdf
+ *
+ * Please contact Sun Microsystems, Inc., 4150 Network Circle, Santa Clara,
+ * CA 95054 USA or visit www.sun.com if you need additional information or
+ * have any questions.
+ *
+ * GPL HEADER END
+ */
+/*
+ * Copyright  2008 Sun Microsystems, Inc. All rights reserved
+ * Use is subject to license terms.
+ *
  * Copyright (c) 2003 Los Alamos National Laboratory (LANL)
- * Copyright (C) 2005 Cluster File Systems, Inc. All rights reserved.
- *
- *   This file is part of Lustre, http://www.lustre.org/
- *
- *   Lustre is free software; you can redistribute it and/or
- *   modify it under the terms of version 2 of the GNU General Public
- *   License as published by the Free Software Foundation.
- *
- *   Lustre is distributed in the hope that it will be useful,
- *   but WITHOUT ANY WARRANTY; without even the implied warranty of
- *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- *   GNU General Public License for more details.
- *
- *   You should have received a copy of the GNU General Public License
- *   along with Lustre; if not, write to the Free Software
- *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
+ */
+/*
+ * This file is part of Lustre, http://www.lustre.org/
+ * Lustre is a trademark of Sun Microsystems, Inc.
  */
 
 #include "gmlnd.h"
@@ -62,9 +78,34 @@ gmnal_tunables_t gmnal_tunables = {
 };
 
 #if defined(CONFIG_SYSCTL) && !CFS_SYSFS_MODULE_PARM
+
+#ifndef HAVE_SYSCTL_UNNUMBERED
+
+enum {
+        GMLND_PORT = 1,
+        GMLND_NTX,
+        GMLND_CREDITS,
+        GMLND_PEERCREDITS,
+        GMLND_NLARGE_TX_BUFS,
+        GMLND_NRX_SMALL,
+        GMLND_NRX_LARGE
+};
+
+#else
+
+#define GMLND_PORT              CTL_UNNUMBERED
+#define GMLND_NTX               CTL_UNNUMBERED
+#define GMLND_CREDITS           CTL_UNNUMBERED
+#define GMLND_PEERCREDITS       CTL_UNNUMBERED
+#define GMLND_NLARGE_TX_BUFS    CTL_UNNUMBERED
+#define GMLND_NRX_SMALL         CTL_UNNUMBERED
+#define GMLND_NRX_LARGE         CTL_UNNUMBERED
+
+#endif
+
 static cfs_sysctl_table_t gmnal_ctl_table[] = {
         {
-                .ctl_name = 1,
+                .ctl_name = GMLND_PORT,
                 .procname = "port",
                 .data     = &port,
                 .maxlen   = sizeof (int),
@@ -72,7 +113,7 @@ static cfs_sysctl_table_t gmnal_ctl_table[] = {
                 .proc_handler = &proc_dointvec
         },
         {
-                .ctl_name = 2,
+                .ctl_name = GMLND_NTX,
                 .procname = "ntx",
                 .data     = &ntx,
                 .maxlen   = sizeof (int),
@@ -80,7 +121,7 @@ static cfs_sysctl_table_t gmnal_ctl_table[] = {
                 .proc_handler = &proc_dointvec
         },
         {
-                .ctl_name = 3,
+                .ctl_name = GMLND_CREDITS,
                 .procname = "credits",
                 .data     = &credits,
                 .maxlen   = sizeof (int),
@@ -88,7 +129,7 @@ static cfs_sysctl_table_t gmnal_ctl_table[] = {
                 .proc_handler = &proc_dointvec
         },
         {
-                .ctl_name = 4,
+                .ctl_name = GMLND_PEERCREDITS,
                 .procname = "peer_credits",
                 .data     = &peer_credits,
                 .maxlen   = sizeof (int),
@@ -96,7 +137,7 @@ static cfs_sysctl_table_t gmnal_ctl_table[] = {
                 .proc_handler = &proc_dointvec
         },
         {
-                .ctl_name = 5,
+                .ctl_name = GMLND_NLARGE_TX_BUFS,
                 .procname = "nlarge_tx_bufs",
                 .data     = &nlarge_tx_bufs,
                 .maxlen   = sizeof (int),
@@ -104,7 +145,7 @@ static cfs_sysctl_table_t gmnal_ctl_table[] = {
                 .proc_handler = &proc_dointvec
         },
         {
-                .ctl_name = 6,
+                .ctl_name = GMLND_NRX_SMALL,
                 .procname = "nrx_small",
                 .data     = &nrx_small,
                 .maxlen   = sizeof (int),
@@ -112,7 +153,7 @@ static cfs_sysctl_table_t gmnal_ctl_table[] = {
                 .proc_handler = &proc_dointvec
         },
         {
-                .ctl_name = 7,
+                .ctl_name = GMLND_NRX_LARGE,
                 .procname = "nrx_large",
                 .data     = &nrx_large,
                 .maxlen   = sizeof (int),
@@ -124,7 +165,7 @@ static cfs_sysctl_table_t gmnal_ctl_table[] = {
 
 static cfs_sysctl_table_t gmnal_top_ctl_table[] = {
         {
-                .ctl_name = 207,
+                .ctl_name = CTL_GMLND,
                 .procname = "gmnal",
                 .data     = NULL,
                 .maxlen   = 0,
@@ -175,6 +216,6 @@ gmnal_unload(void)
 module_init(gmnal_load);
 module_exit(gmnal_unload);
 
-MODULE_AUTHOR("Cluster File Systems, Inc. <info@clusterfs.com>");
+MODULE_AUTHOR("Sun Microsystems, Inc. <http://www.lustre.org/>");
 MODULE_DESCRIPTION("Kernel GM LND v1.01");
 MODULE_LICENSE("GPL");
