@@ -301,6 +301,7 @@ kptllnd_active_rdma(kptl_rx_t *rx, lnet_msg_t *lntmsg, int type,
                 kptllnd_peer_close(peer, -EIO);
                 /* Everything (including this RDMA) queued on the peer will
                  * be completed with failure */
+                kptllnd_schedule_ptltrace_dump();
         }
 
         return 0;
@@ -319,22 +320,27 @@ kptllnd_send(lnet_ni_t *ni, void *private, lnet_msg_t *lntmsg)
         lnet_kiov_t      *payload_kiov = lntmsg->msg_kiov;
         unsigned int      payload_offset = lntmsg->msg_offset;
         unsigned int      payload_nob = lntmsg->msg_len;
+        kptl_net_t       *net = ni->ni_data;
         kptl_peer_t      *peer;
         kptl_tx_t        *tx;
         int               nob;
         int               nfrag;
         int               rc;
 
+        LASSERT (net->net_ni == ni);
+        LASSERT (!net->net_shutdown);
         LASSERT (payload_nob == 0 || payload_niov > 0);
         LASSERT (payload_niov <= LNET_MAX_IOV);
         LASSERT (payload_niov <= PTL_MD_MAX_IOV); /* !!! */
         LASSERT (!(payload_kiov != NULL && payload_iov != NULL));
         LASSERT (!in_interrupt());
 
-        rc = kptllnd_find_target(&peer, target);
+        rc = kptllnd_find_target(net, target, &peer);
         if (rc != 0)
                 return rc;
         
+        /* NB peer->peer_id does NOT always equal target, be careful with
+         * which one to use */
         switch (type) {
         default:
                 LBUG();
@@ -364,7 +370,7 @@ kptllnd_send(lnet_ni_t *ni, void *private, lnet_msg_t *lntmsg)
                 tx->tx_lnet_msg = lntmsg;
                 tx->tx_msg->ptlm_u.rdma.kptlrm_hdr = *hdr;
                 kptllnd_init_msg (tx->tx_msg, PTLLND_MSG_TYPE_PUT,
-                                  sizeof(kptl_rdma_msg_t));
+                                 target, sizeof(kptl_rdma_msg_t));
 
                 CDEBUG(D_NETTRACE, "%s: passive PUT p %d %p\n",
                        libcfs_id2str(target),
@@ -393,8 +399,7 @@ kptllnd_send(lnet_ni_t *ni, void *private, lnet_msg_t *lntmsg)
                         goto out;
                 }
 
-                tx->tx_lnet_replymsg =
-                        lnet_create_reply_msg(kptllnd_data.kptl_ni, lntmsg);
+                tx->tx_lnet_replymsg = lnet_create_reply_msg(ni, lntmsg);
                 if (tx->tx_lnet_replymsg == NULL) {
                         CERROR("Failed to allocate LNET reply for %s\n",
                                libcfs_id2str(target));
@@ -415,7 +420,7 @@ kptllnd_send(lnet_ni_t *ni, void *private, lnet_msg_t *lntmsg)
                 tx->tx_lnet_msg = lntmsg;
                 tx->tx_msg->ptlm_u.rdma.kptlrm_hdr = *hdr;
                 kptllnd_init_msg (tx->tx_msg, PTLLND_MSG_TYPE_GET,
-                                  sizeof(kptl_rdma_msg_t));
+                                 target, sizeof(kptl_rdma_msg_t));
 
                 CDEBUG(D_NETTRACE, "%s: passive GET p %d %p\n",
                        libcfs_id2str(target),
@@ -467,7 +472,7 @@ kptllnd_send(lnet_ni_t *ni, void *private, lnet_msg_t *lntmsg)
         }
         
         nob = offsetof(kptl_immediate_msg_t, kptlim_payload[payload_nob]);
-        kptllnd_init_msg(tx->tx_msg, PTLLND_MSG_TYPE_IMMEDIATE, nob);
+        kptllnd_init_msg(tx->tx_msg, PTLLND_MSG_TYPE_IMMEDIATE, target, nob);
 
         CDEBUG(D_NETTRACE, "%s: immediate %s p %d %p\n",
                libcfs_id2str(target),
@@ -687,6 +692,22 @@ kptllnd_watchdog(void *arg)
         /* threads shut down in phase 2 after all peers have been destroyed */
         while (kptllnd_data.kptl_shutdown < 2) {
 
+                /* add a check for needs ptltrace
+                 * yes, this is blatant hijacking of this thread
+                 * we can't dump directly from tx or rx _callbacks as it deadlocks portals
+                 * and takes out the node
+                */
+
+                if (atomic_read(&kptllnd_data.kptl_needs_ptltrace)) {
+#ifdef CRAY_XT3
+                        kptllnd_dump_ptltrace();
+                        /* we only dump once, no matter how many pending */
+                        atomic_set(&kptllnd_data.kptl_needs_ptltrace, 0);
+#else
+                        LBUG();
+#endif
+                }
+
                 timeout = (int)(deadline - jiffies);
 
                 if (timeout <= 0) {
