@@ -1,24 +1,41 @@
 /* -*- mode: c; c-basic-offset: 8; indent-tabs-mode: nil; -*-
  * vim:expandtab:shiftwidth=8:tabstop=8:
  *
- * Copyright (C) 2004 Cluster File Systems, Inc.
- *   Author: Eric Barton <eric@bartonsoftware.com>
+ * GPL HEADER START
  *
- *   This file is part of Lustre, http://www.lustre.org.
+ * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
  *
- *   Lustre is free software; you can redistribute it and/or
- *   modify it under the terms of version 2 of the GNU General Public
- *   License as published by the Free Software Foundation.
+ * This program is free software; you can redistribute it and/or modify
+ * it under the terms of the GNU General Public License version 2 only,
+ * as published by the Free Software Foundation.
  *
- *   Lustre is distributed in the hope that it will be useful,
- *   but WITHOUT ANY WARRANTY; without even the implied warranty of
- *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- *   GNU General Public License for more details.
+ * This program is distributed in the hope that it will be useful, but
+ * WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+ * General Public License version 2 for more details (a copy is included
+ * in the LICENSE file that accompanied this code).
  *
- *   You should have received a copy of the GNU General Public License
- *   along with Lustre; if not, write to the Free Software
- *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
+ * You should have received a copy of the GNU General Public License
+ * version 2 along with this program; If not, see
+ * http://www.sun.com/software/products/lustre/docs/GPLv2.pdf
  *
+ * Please contact Sun Microsystems, Inc., 4150 Network Circle, Santa Clara,
+ * CA 95054 USA or visit www.sun.com if you need additional information or
+ * have any questions.
+ *
+ * GPL HEADER END
+ */
+/*
+ * Copyright  2008 Sun Microsystems, Inc. All rights reserved
+ * Use is subject to license terms.
+ */
+/*
+ * This file is part of Lustre, http://www.lustre.org/
+ * Lustre is a trademark of Sun Microsystems, Inc.
+ *
+ * lnet/klnds/ralnd/ralnd_modparams.c
+ *
+ * Author: Eric Barton <eric@bartonsoftware.com>
  */
 
 #include "ralnd.h"
@@ -72,9 +89,35 @@ kra_tunables_t kranal_tunables = {
 };
 
 #if CONFIG_SYSCTL && !CFS_SYSFS_MODULE_PARM
+
+#ifndef HAVE_SYSCTL_UNNUMBERED
+enum {
+        KRANAL_N_CONND = 1,
+        KRANAL_RECONNECT_MIN,
+        KRANAL_RECONNECT_MAX,
+        KRANAL_NTX,
+        KRANAL_CREDITS,
+        KRANAL_PEERCREDITS,
+        KRANAL_FMA_CQ_SIZE,
+        KRANAL_TIMEOUT,
+        KRANAL_IMMEDIATE_MAX
+};
+#else
+
+#define KRANAL_N_CONND          CTL_UNNUMBERED
+#define KRANAL_RECONNECT_MIN    CTL_UNNUMBERED
+#define KRANAL_RECONNECT_MAX    CTL_UNNUMBERED
+#define KRANAL_NTX              CTL_UNNUMBERED
+#define KRANAL_CREDITS          CTL_UNNUMBERED
+#define KRANAL_PEERCREDITS      CTL_UNNUMBERED
+#define KRANAL_FMA_CQ_SIZE      CTL_UNNUMBERED
+#define KRANAL_TIMEOUT          CTL_UNNUMBERED
+#define KRENAL_IMMEDIATE_MAX    CTL_UNNUMBERED
+#endif
+
 static cfs_sysctl_table_t kranal_ctl_table[] = {
         {
-                .ctl_name = 1,
+                .ctl_name = KRANAL_N_CONND,
                 .procname = "n_connd",
                 .data     = &n_connd,
                 .maxlen   = sizeof(int),
@@ -82,7 +125,7 @@ static cfs_sysctl_table_t kranal_ctl_table[] = {
                 .proc_handler = &proc_dointvec
         },
         {
-                .ctl_name = 2,
+                .ctl_name = KRANAL_RECONNECT_MIN,
                 .procname = "min_reconnect_interval",
                 .data     = &min_reconnect_interval,
                 .maxlen   = sizeof(int),
@@ -90,7 +133,7 @@ static cfs_sysctl_table_t kranal_ctl_table[] = {
                 .proc_handler = &proc_dointvec
         },
         {
-                .ctl_name = 3,
+                .ctl_name = KRANAL_RECONNECT_MAX,
                 .procname = "max_reconnect_interval",
                 .data     = &max_reconnect_interval,
                 .maxlen   = sizeof(int),
@@ -98,7 +141,7 @@ static cfs_sysctl_table_t kranal_ctl_table[] = {
                 .proc_handler = &proc_dointvec
         },
         {
-                .ctl_name = 4,
+                .ctl_name = KRANAL_NTX,
                 .procname = "ntx",
                 .data     = &ntx,
                 .maxlen   = sizeof(int),
@@ -106,7 +149,7 @@ static cfs_sysctl_table_t kranal_ctl_table[] = {
                 .proc_handler = &proc_dointvec
         },
         {
-                .ctl_name = 5,
+                .ctl_name = KRANAL_CREDITS,
                 .procname = "credits",
                 .data     = &credits,
                 .maxlen   = sizeof(int),
@@ -114,7 +157,7 @@ static cfs_sysctl_table_t kranal_ctl_table[] = {
                 .proc_handler = &proc_dointvec
         },
         {
-                .ctl_name = 6,
+                .ctl_name = KRANAL_PEERCREDITS,
                 .procname = "peer_credits",
                 .data     = &peer_credits,
                 .maxlen   = sizeof(int),
@@ -122,7 +165,7 @@ static cfs_sysctl_table_t kranal_ctl_table[] = {
                 .proc_handler = &proc_dointvec
         },
         {
-                .ctl_name = 7,
+                .ctl_name = KRANAL_FMA_CQ_SIZE,
                 .procname = "fma_cq_size",
                 .data     = &fma_cq_size,
                 .maxlen   = sizeof(int),
@@ -130,7 +173,7 @@ static cfs_sysctl_table_t kranal_ctl_table[] = {
                 .proc_handler = &proc_dointvec
         },
         {
-                .ctl_name = 8,
+                .ctl_name = KRANAL_TIMEOUT,
                 .procname = "timeout",
                 .data     = &timeout,
                 .maxlen   = sizeof(int),
@@ -138,7 +181,7 @@ static cfs_sysctl_table_t kranal_ctl_table[] = {
                 .proc_handler = &proc_dointvec
         },
         {
-                .ctl_name = 9,
+                .ctl_name = KRANAL_IMMEDIATE_MAX,
                 .procname = "max_immediate",
                 .data     = &max_immediate,
                 .maxlen   = sizeof(int),
@@ -150,7 +193,7 @@ static cfs_sysctl_table_t kranal_ctl_table[] = {
 
 static cfs_sysctl_table_t kranal_top_ctl_table[] = {
         {
-                .ctl_name = 202,
+                .ctl_name = CTL_KRANAL,
                 .procname = "ranal",
                 .data     = NULL,
                 .maxlen   = 0,
@@ -193,4 +236,3 @@ kranal_tunables_fini ()
 }
 
 #endif
-
