@@ -103,6 +103,8 @@ enum {
         PSDEV_LNET_DAEMON_FILE,   /* spool kernel debug buffer to file */
         PSDEV_LNET_DEBUG_MB,      /* size of debug buffer */
         PSDEV_LNET_DEBUG_LOG_UPCALL, /* debug log upcall script */
+        PSDEV_LNET_WATCHDOG_RATELIMIT,  /* ratelimit watchdog messages  */
+        PSDEV_LNET_FORCE_LBUG,    /* hook to force an LBUG */
 };
 #else
 #define CTL_LNET                        CTL_UNNUMBERED
@@ -123,10 +125,12 @@ enum {
 #define PSDEV_LNET_DAEMON_FILE          CTL_UNNUMBERED
 #define PSDEV_LNET_DEBUG_MB             CTL_UNNUMBERED
 #define PSDEV_LNET_DEBUG_LOG_UPCALL     CTL_UNNUMBERED
+#define PSDEV_LNET_WATCHDOG_RATELIMIT   CTL_UNNUMBERED
+#define PSDEV_LNET_FORCE_LBUG           CTL_UNNUMBERED
 #endif
 
 
-static int
+int
 proc_call_handler(void *data, int write,
                   loff_t *ppos, void *buffer, size_t *lenp,
                   int (*handler)(void *data, int write,
@@ -188,6 +192,9 @@ static int __proc_dobitmasks(void *data, int write,
 
 DECLARE_PROC_HANDLER(proc_dobitmasks)
 
+static int min_watchdog_ratelimit = 0;          /* disable ratelimiting */
+static int max_watchdog_ratelimit = (24*60*60); /* limit to once per day */
+
 static int __proc_dump_kernel(void *data, int write,
                               loff_t pos, void *buffer, int nob)
 {
@@ -325,6 +332,13 @@ int LL_PROC_PROTO(proc_console_backoff)
         return rc;
 }
 
+int LL_PROC_PROTO(libcfs_force_lbug)
+{
+        if (write)
+                LBUG();
+        return 0;
+}
+
 static cfs_sysctl_table_t lnet_table[] = {
         /*
          * NB No .strategy entries have been provided since sysctl(8) prefers
@@ -454,6 +468,23 @@ static cfs_sysctl_table_t lnet_table[] = {
                 .mode     = 0644,
                 .proc_handler = &proc_debug_mb,
         },
+        {
+                .ctl_name = PSDEV_LNET_WATCHDOG_RATELIMIT,
+                .procname = "watchdog_ratelimit",
+                .data     = &libcfs_watchdog_ratelimit,
+                .maxlen   = sizeof(int),
+                .mode     = 0644,
+                .proc_handler = &proc_dointvec_minmax,
+                .extra1   = &min_watchdog_ratelimit,
+                .extra2   = &max_watchdog_ratelimit,
+        },
+        {       .ctl_name = PSDEV_LNET_FORCE_LBUG,
+                .procname = "force_lbug",
+                .data     = NULL,
+                .maxlen   = 0,
+                .mode     = 0200,
+                .proc_handler = &libcfs_force_lbug
+        },
         {0}
 };
 
