@@ -1,27 +1,48 @@
 /* -*- mode: c; c-basic-offset: 8; indent-tabs-mode: nil; -*-
  * vim:expandtab:shiftwidth=8:tabstop=8:
  *
- * Copyright (C) 2005 Cluster File Systems, Inc.
+ * GPL HEADER START
  *
- *   This file is part of Lustre, http://www.lustre.org.
+ * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
  *
- *   Lustre is free software; you can redistribute it and/or
- *   modify it under the terms of version 2 of the GNU General Public
- *   License as published by the Free Software Foundation.
+ * This program is free software; you can redistribute it and/or modify
+ * it under the terms of the GNU General Public License version 2 only,
+ * as published by the Free Software Foundation.
  *
- *   Lustre is distributed in the hope that it will be useful,
- *   but WITHOUT ANY WARRANTY; without even the implied warranty of
- *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- *   GNU General Public License for more details.
+ * This program is distributed in the hope that it will be useful, but
+ * WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+ * General Public License version 2 for more details (a copy is included
+ * in the LICENSE file that accompanied this code).
  *
- *   You should have received a copy of the GNU General Public License
- *   along with Lustre; if not, write to the Free Software
- *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
+ * You should have received a copy of the GNU General Public License
+ * version 2 along with this program; If not, see
+ * http://www.sun.com/software/products/lustre/docs/GPLv2.pdf
+ *
+ * Please contact Sun Microsystems, Inc., 4150 Network Circle, Santa Clara,
+ * CA 95054 USA or visit www.sun.com if you need additional information or
+ * have any questions.
+ *
+ * GPL HEADER END
+ */
+/*
+ * Copyright  2008 Sun Microsystems, Inc. All rights reserved
+ * Use is subject to license terms.
+ */
+/*
+ * This file is part of Lustre, http://www.lustre.org/
+ * Lustre is a trademark of Sun Microsystems, Inc.
  */
 
-#ifndef __KERNEL__
-#ifndef REDSTORM
+#if !defined(__KERNEL__) || !defined(REDSTORM)
+
+#include <libcfs/libcfs.h>
+#include <libcfs/kp30.h>
+
 #include <sys/socket.h>
+#ifdef	HAVE_NETINET_IN_H
+#include <netinet/in.h>
+#endif
 #include <netinet/tcp.h>
 #include <sys/ioctl.h>
 #include <unistd.h>
@@ -38,9 +59,6 @@
 #include <sys/syscall.h>
 #endif
 
-#include <libcfs/libcfs.h>
-#include <libcfs/kp30.h>
-
 /*
  * Functions to get network interfaces info
  */
@@ -49,8 +67,8 @@ int
 libcfs_sock_ioctl(int cmd, unsigned long arg)
 {
         int fd, rc;
-        
-        fd = socket(AF_INET, SOCK_STREAM, 0);        
+
+        fd = socket(AF_INET, SOCK_STREAM, 0);
 
         if (fd < 0) {
                 rc = -errno;
@@ -600,5 +618,4 @@ int libcfs_sock_readv(int fd, const struct iovec *vector, int count)
         return rc;
 }
 
-#endif /* !defined(REDSTORM) */
-#endif /* !__KERNEL__ */
+#endif /* !__KERNEL__ || !defined(REDSTORM) */
