@@ -182,19 +182,19 @@ static inline lvfs_sbdev_type fsfilt_journal_sbdev(struct obd_device *obd,
 #define FSFILT_OP_JOIN          11
 #define FSFILT_OP_NOOP          15
 
-#define __fsfilt_check_slow(obd, start, msg)                            \
-do {                                                                    \
-        if (time_before(jiffies, start + 15 * HZ))                      \
-                break;                                                  \
-        else if (time_before(jiffies, start + 30 * HZ))                 \
-                CDEBUG(D_VFSTRACE, "%s: slow %s %lus\n", obd->obd_name, \
-                       msg, (jiffies-start) / HZ);                      \
-        else if (time_before(jiffies, start + DISK_TIMEOUT * HZ))       \
-                CWARN("%s: slow %s %lus\n", obd->obd_name, msg,         \
-                      (jiffies - start) / HZ);                          \
-        else                                                            \
-                CERROR("%s: slow %s %lus\n", obd->obd_name, msg,        \
-                       (jiffies - start) / HZ);                         \
+#define __fsfilt_check_slow(obd, start, msg)                                 \
+do {                                                                         \
+        if (time_before(jiffies, start + 15 * HZ))                           \
+                break;                                                       \
+        else if (time_before(jiffies, start + 30 * HZ))                      \
+                CDEBUG(D_VFSTRACE, "%s: slow %s %lus due to heavy IO load\n",\
+                       obd->obd_name, msg, (jiffies-start) / HZ);            \
+        else if (time_before(jiffies, start + DISK_TIMEOUT * HZ))            \
+                LCONSOLE_INFO("%s: slow %s %lus due to heavy IO load\n",     \
+                              obd->obd_name, msg, (jiffies - start) / HZ);   \
+        else                                                                 \
+                LCONSOLE_WARN("%s: slow %s %lus due to heavy IO load\n",     \
+                              obd->obd_name, msg, (jiffies - start) / HZ);   \
 } while (0)
 
 #define fsfilt_check_slow(obd, start, msg)    \
@@ -211,9 +211,6 @@ static inline void *fsfilt_start_log(struct obd_device *obd,
         void *parent_handle = oti ? oti->oti_handle : NULL;
         void *handle;
 
-        if (obd->obd_fail)
-                return ERR_PTR(-EROFS);
-
         handle = obd->obd_fsops->fs_start(inode, op, parent_handle, logs);
         CDEBUG(D_INFO, "started handle %p (%p)\n", handle, parent_handle);
 
@@ -245,9 +242,6 @@ static inline void *fsfilt_brw_start_log(struct obd_device *obd, int objcount,
         void *parent_handle = oti ? oti->oti_handle : NULL;
         void *handle;
 
-        if (obd->obd_fail)
-                return ERR_PTR(-EROFS);
-
         handle = obd->obd_fsops->fs_brw_start(objcount, fso, niocount, nb,
                                               parent_handle, logs);
         CDEBUG(D_INFO, "started handle %p (%p)\n", handle, parent_handle);
@@ -331,11 +325,24 @@ static inline int fsfilt_setattr(struct obd_device *obd, struct dentry *dentry,
         return rc;
 }
 
-static inline int fsfilt_iocontrol(struct obd_device *obd, struct inode *inode,
-                                   struct file *file, unsigned int cmd,
-                                   unsigned long arg)
+static inline int fsfilt_iocontrol(struct obd_device *obd, struct dentry *dentry,
+                                   unsigned int cmd, unsigned long arg)
 {
-        return obd->obd_fsops->fs_iocontrol(inode, file, cmd, arg);
+        struct file *dummy_file = NULL;
+        int ret;
+
+        OBD_ALLOC_PTR(dummy_file);
+        if (!dummy_file)
+                return(-ENOMEM);
+
+        dummy_file->f_dentry = dentry;
+        dummy_file->f_vfsmnt = obd->u.obt.obt_vfsmnt;
+
+        ret = obd->obd_fsops->fs_iocontrol(dentry->d_inode, dummy_file, cmd,
+                                           arg);
+
+        OBD_FREE_PTR(dummy_file);
+        return ret;
 }
 
 static inline int fsfilt_set_md(struct obd_device *obd, struct inode *inode,
