@@ -40,7 +40,6 @@
 #include <lustre_handles.h>
 #include <lustre/lustre_idl.h>
 
-
 /* Adaptive Timeout stuff */
 #define D_ADAPTTO D_OTHER
 #define AT_BINS 4                  /* "bin" means "N seconds of history" */
@@ -69,6 +68,14 @@ enum lustre_imp_state {
         LUSTRE_IMP_EVICTED    = 10,
 };
 
+struct ptlrpc_at_array {
+        struct list_head *paa_reqs_array; /* array to hold requests */
+        __u32             paa_size;       /* the size of array */
+        __u32             paa_count;      /* the total count of reqs */
+        time_t            paa_deadline;   /* the earliest deadline of reqs */
+        __u32            *paa_reqs_count; /* the count of reqs in each entry */
+};
+
 static inline char * ptlrpc_import_state_name(enum lustre_imp_state state)
 {
         static char* import_state_names[] = {
@@ -96,13 +103,20 @@ struct obd_import_conn {
         __u64                     oic_last_attempt; /* jiffies, 64-bit */
 };
 
-#define IMP_AT_MAX_PORTALS 4
+#define IMP_AT_MAX_PORTALS 8
 struct imp_at {
         int                     iat_portal[IMP_AT_MAX_PORTALS];
         struct adaptive_timeout iat_net_latency;
         struct adaptive_timeout iat_service_estimate[IMP_AT_MAX_PORTALS];
 };
 
+/* state history */
+#define IMP_STATE_HIST_LEN 16
+struct import_state_hist {
+        enum lustre_imp_state ish_state;
+        time_t                ish_time;
+};
+
 struct obd_import {
         struct portals_handle     imp_handle;
         atomic_t                  imp_refcount;
@@ -125,8 +139,11 @@ struct obd_import {
         atomic_t                  imp_inflight;
         atomic_t                  imp_unregistering;
         atomic_t                  imp_replay_inflight;
-        atomic_t                  imp_inval_count;
+        atomic_t                  imp_inval_count;  /* in-progress invalidations */
+        atomic_t                  imp_timeouts;
         enum lustre_imp_state     imp_state;
+        struct import_state_hist  imp_state_hist[IMP_STATE_HIST_LEN];
+        int                       imp_state_hist_idx;
         int                       imp_generation;
         __u32                     imp_conn_cnt;
         int                       imp_last_generation_checked;
@@ -150,13 +167,17 @@ struct obd_import {
                                   imp_replayable:1,       /* try to recover the import */
                                   imp_dlm_fake:1,         /* don't run recovery (timeout instead) */
                                   imp_server_timeout:1,   /* use 1/2 timeout on MDS' OSCs */
-                                  imp_initial_recov:1,    /* retry the initial connection */  
+                                  imp_initial_recov:1,    /* retry the initial connection */
                                   imp_initial_recov_bk:1, /* turn off init_recov after trying all failover nids */
+                                  imp_delayed_recovery:1, /* VBR: imp in delayed recovery */
+                                  imp_no_lock_replay:1,   /* VBR: if gap was found then no lock replays */
+                                  imp_vbr_failed:1,       /* recovery by versions was failed */
                                   imp_force_verify:1,     /* force an immidiate ping */
                                   imp_pingable:1,         /* pingable */
                                   imp_resend_replay:1,    /* resend for replay */
                                   imp_recon_bk:1,         /* turn off reconnect if all failovers fail */
-                                  imp_last_recon:1;       /* internally used by above */
+                                  imp_last_recon:1,       /* internally used by above */
+                                  imp_force_reconnect:1;  /* import must be reconnected instead of chouse new connection */
         __u32                     imp_connect_op;
         struct obd_connect_data   imp_connect_data;
         __u64                     imp_connect_flags_orig;
@@ -194,7 +215,7 @@ static inline void at_init(struct adaptive_timeout *at, int val, int flags) {
 static inline int at_get(struct adaptive_timeout *at) {
         return at->at_current;
 }
-int at_add(struct adaptive_timeout *at, unsigned int val);
+int at_measured(struct adaptive_timeout *at, unsigned int val);
 int import_at_get_index(struct obd_import *imp, int portal);
 extern unsigned int at_max;
 #define AT_OFF (at_max == 0)
