@@ -63,6 +63,7 @@ extern cfs_mem_cache_t *obd_lvfs_ctxt_cache;
 /* OBD Operations Declarations */
 extern struct obd_device *class_conn2obd(struct lustre_handle *);
 extern struct obd_device *class_exp2obd(struct obd_export *);
+extern int class_handle_ioctl(unsigned int cmd, unsigned long arg);
 
 /* genops.c */
 struct obd_export *class_conn2export(struct lustre_handle *);
@@ -102,6 +103,7 @@ int obd_export_evict_by_uuid(struct obd_device *obd, char *uuid);
 int obd_zombie_impexp_init(void);
 void obd_zombie_impexp_stop(void);
 void obd_zombie_impexp_cull(void);
+void obd_zombie_barrier(void);
 
 /* obd_config.c */
 int class_process_config(struct lustre_cfg *lcfg);
@@ -153,6 +155,7 @@ struct config_llog_data {
         struct obd_export  *cld_mgcexp;
         unsigned int        cld_stopping:1; /* we were told to stop watching */
         unsigned int        cld_lostlock:1; /* lock not requeued */
+        struct semaphore    cld_sem; /* for exclusive processing of the log */
 };
 
 struct lustre_profile {
@@ -192,6 +195,7 @@ void class_del_profiles(void);
         exp_;                                                                  \
 })
 
+/* class_export_put() is non-blocking */
 #define class_export_put(exp)                                                  \
 do {                                                                           \
         LASSERT((exp) != NULL);                                                \
@@ -219,9 +223,14 @@ int class_connect(struct lustre_handle *conn, struct obd_device *obd,
                   struct obd_uuid *cluuid);
 int class_disconnect(struct obd_export *exp);
 void class_fail_export(struct obd_export *exp);
+int class_connected_export(struct obd_export *exp);
 void class_disconnect_exports(struct obd_device *obddev);
+void class_set_export_delayed(struct obd_export *exp);
+void class_handle_stale_exports(struct obd_device *obddev);
+void class_disconnect_expired_exports(struct obd_device *obd);
 void class_disconnect_stale_exports(struct obd_device *obddev,
                                     enum obd_option flags);
+int class_stale_export_list(struct obd_device *obd, struct obd_ioctl_data *data);
 int class_manual_cleanup(struct obd_device *obd);
 
 static inline enum obd_option exp_flags_from_obd(struct obd_device *obd)
@@ -297,6 +306,26 @@ do {                                                            \
 #define EXP_COUNTER_INCREMENT(exp, op);
 #endif
 
+static inline int lprocfs_nid_ldlm_stats_init(struct nid_stat* tmp) {
+        int rc;
+
+        rc = lprocfs_register_stats(tmp->nid_proc, "stats",
+                                    tmp->nid_stats);
+        if (rc)
+                return rc;
+
+        /* Always add in ldlm_stats */
+        tmp->nid_ldlm_stats = lprocfs_alloc_stats(LDLM_LAST_OPC - LDLM_FIRST_OPC
+                                                  ,LPROCFS_STATS_FLAG_NOPERCPU);
+        if (tmp->nid_ldlm_stats == NULL)
+                return -ENOMEM;
+
+        lprocfs_init_ldlm_stats(tmp->nid_ldlm_stats);
+
+        return lprocfs_register_stats(tmp->nid_proc, "ldlm_stats",
+                                      tmp->nid_ldlm_stats);
+}
+
 #define OBD_CHECK_OP(obd, op, err)                              \
 do {                                                            \
         if (!OBT(obd) || !OBP((obd), op)) {\
@@ -528,6 +557,21 @@ static inline int obd_precreate(struct obd_export *exp)
         RETURN(rc);
 }
 
+static inline int obd_create_async(struct obd_export *exp,
+                                   struct obd_info *oinfo,
+                                   struct lov_stripe_md **ea,
+                                   struct obd_trans_info *oti)
+{
+        int rc;
+        ENTRY;
+
+        EXP_CHECK_OP(exp, create_async);
+        EXP_COUNTER_INCREMENT(exp, create_async);
+
+        rc = OBP(exp->exp_obd, create_async)(exp, oinfo, ea, oti);
+        RETURN(rc);
+}
+
 static inline int obd_create(struct obd_export *exp, struct obdo *obdo,
                              struct lov_stripe_md **ea,
                              struct obd_trans_info *oti)
@@ -718,6 +762,30 @@ static inline int obd_disconnect(struct obd_export *exp)
         RETURN(rc);
 }
 
+static inline int obd_fid_init(struct obd_export *exp)
+{
+        int rc;
+        ENTRY;
+
+        OBD_CHECK_OP(exp->exp_obd, fid_init, 0);
+        EXP_COUNTER_INCREMENT(exp, fid_init);
+
+        rc = OBP(exp->exp_obd, fid_init)(exp);
+        RETURN(rc);
+}
+
+static inline int obd_fid_fini(struct obd_export *exp)
+{
+        int rc;
+        ENTRY;
+
+        OBD_CHECK_OP(exp->exp_obd, fid_fini, 0);
+        EXP_COUNTER_INCREMENT(exp, fid_fini);
+
+        rc = OBP(exp->exp_obd, fid_fini)(exp);
+        RETURN(rc);
+}
+
 static inline int obd_ping(struct obd_export *exp)
 {
         int rc;
@@ -730,6 +798,74 @@ static inline int obd_ping(struct obd_export *exp)
         RETURN(rc);
 }
 
+static inline int obd_pool_new(struct obd_device *obd, char *poolname)
+{
+        int rc;
+        ENTRY;
+
+        OBD_CHECK_OP(obd, pool_new, -EOPNOTSUPP);
+        OBD_COUNTER_INCREMENT(obd, pool_new);
+
+        rc = OBP(obd, pool_new)(obd, poolname);
+        RETURN(rc);
+}
+
+static inline int obd_pool_del(struct obd_device *obd, char *poolname)
+{
+        int rc;
+        ENTRY;
+
+        OBD_CHECK_OP(obd, pool_del, -EOPNOTSUPP);
+        OBD_COUNTER_INCREMENT(obd, pool_del);
+
+        rc = OBP(obd, pool_del)(obd, poolname);
+        RETURN(rc);
+}
+
+static inline int obd_pool_add(struct obd_device *obd, char *poolname, char *ostname)
+{
+        int rc;
+        ENTRY;
+
+        OBD_CHECK_OP(obd, pool_add, -EOPNOTSUPP);
+        OBD_COUNTER_INCREMENT(obd, pool_add);
+
+        rc = OBP(obd, pool_add)(obd, poolname, ostname);
+        RETURN(rc);
+}
+
+static inline int obd_pool_rem(struct obd_device *obd, char *poolname, char *ostname)
+{
+        int rc;
+        ENTRY;
+
+        OBD_CHECK_OP(obd, pool_rem, -EOPNOTSUPP);
+        OBD_COUNTER_INCREMENT(obd, pool_rem);
+
+        rc = OBP(obd, pool_rem)(obd, poolname, ostname);
+        RETURN(rc);
+}
+
+static inline void obd_getref(struct obd_device *obd)
+{
+        ENTRY;
+        if (OBT(obd) && OBP(obd, getref)) {
+                OBD_COUNTER_INCREMENT(obd, getref);
+                OBP(obd, getref)(obd);
+        }
+        EXIT;
+}
+
+static inline void obd_putref(struct obd_device *obd)
+{
+        ENTRY;
+        if (OBT(obd) && OBP(obd, putref)) {
+                OBD_COUNTER_INCREMENT(obd, putref);
+                OBP(obd, putref)(obd);
+        }
+        EXIT;
+}
+
 static inline int obd_init_export(struct obd_export *exp)
 {
         int rc = 0;
@@ -871,9 +1007,30 @@ static inline int obd_statfs(struct obd_device *obd, struct obd_statfs *osfs,
         RETURN(rc);
 }
 
-static inline int obd_sync(struct obd_export *exp, struct obdo *oa,
-                           struct lov_stripe_md *ea, obd_size start,
-                           obd_size end)
+static inline int obd_sync_rqset(struct obd_export *exp, struct obd_info *oinfo,
+                                 obd_size start, obd_size end)
+{
+        struct ptlrpc_request_set *set = NULL;
+        int rc;
+        ENTRY;
+
+        OBD_CHECK_OP(exp->exp_obd, sync, -EOPNOTSUPP);
+        EXP_COUNTER_INCREMENT(exp, sync);
+
+        set =  ptlrpc_prep_set();
+        if (set == NULL)
+                RETURN(-ENOMEM);
+
+        rc = OBP(exp->exp_obd, sync)(exp, oinfo, start, end, set);
+        if (rc == 0)
+                rc = ptlrpc_set_wait(set);
+        ptlrpc_set_destroy(set);
+        RETURN(rc);
+}
+
+static inline int obd_sync(struct obd_export *exp, struct obd_info *oinfo,
+                           obd_size start, obd_size end,
+                           struct ptlrpc_request_set *set)
 {
         int rc;
         ENTRY;
@@ -881,7 +1038,7 @@ static inline int obd_sync(struct obd_export *exp, struct obdo *oa,
         OBD_CHECK_OP(exp->exp_obd, sync, -EOPNOTSUPP);
         EXP_COUNTER_INCREMENT(exp, sync);
 
-        rc = OBP(exp->exp_obd, sync)(exp, oa, ea, start, end);
+        rc = OBP(exp->exp_obd, sync)(exp, oinfo, start, end, set);
         RETURN(rc);
 }
 
@@ -944,7 +1101,7 @@ static inline int obd_brw(int cmd, struct obd_export *exp,
 static inline int obd_brw_async(int cmd, struct obd_export *exp,
                                 struct obd_info *oinfo, obd_count oa_bufs,
                                 struct brw_page *pg, struct obd_trans_info *oti,
-                                struct ptlrpc_request_set *set)
+                                struct ptlrpc_request_set *set, int pshift)
 {
         int rc;
         ENTRY;
@@ -957,7 +1114,8 @@ static inline int obd_brw_async(int cmd, struct obd_export *exp,
                 LBUG();
         }
 
-        rc = OBP(exp->exp_obd, brw_async)(cmd, exp, oinfo, oa_bufs, pg,oti,set);
+        rc = OBP(exp->exp_obd, brw_async)(cmd, exp, oinfo, oa_bufs,
+                                          pg, oti,set, pshift);
         RETURN(rc);
 }
 
@@ -977,7 +1135,7 @@ static inline int obd_brw_rqset(int cmd, struct obd_export *exp,
 
         oinfo.oi_oa = oa;
         oinfo.oi_md = lsm;
-        rc = obd_brw_async(cmd, exp, &oinfo, oa_bufs, pg, oti, set);
+        rc = obd_brw_async(cmd, exp, &oinfo, oa_bufs, pg, oti, set, 0);
         if (rc == 0) {
                 rc = ptlrpc_set_wait(set);
                 if (rc)
@@ -990,12 +1148,16 @@ static inline int obd_brw_rqset(int cmd, struct obd_export *exp,
         RETURN(rc);
 }
 
+/* flags used by obd_prep_async_page */
+#define OBD_PAGE_NO_CACHE  0x00000001 /* don't add to cache */
+#define OBD_FAST_LOCK 0x00000002 /* lockh refers to a "fast lock" */
+
 static inline  int obd_prep_async_page(struct obd_export *exp,
                                        struct lov_stripe_md *lsm,
                                        struct lov_oinfo *loi,
                                        cfs_page_t *page, obd_off offset,
                                        struct obd_async_page_ops *ops,
-                                       void *data, void **res, int nocache,
+                                       void *data, void **res, int flags,
                                        struct lustre_handle *lockh)
 {
         int ret;
@@ -1005,37 +1167,23 @@ static inline  int obd_prep_async_page(struct obd_export *exp,
         EXP_COUNTER_INCREMENT(exp, prep_async_page);
 
         ret = OBP(exp->exp_obd, prep_async_page)(exp, lsm, loi, page, offset,
-                                                 ops, data, res, nocache,
+                                                 ops, data, res, flags,
                                                  lockh);
         RETURN(ret);
 }
 
-static inline int obd_reget_short_lock(struct obd_export *exp,
-                                       struct lov_stripe_md *lsm,
-                                       void **res, int rw,
-                                       obd_off start, obd_off end,
-                                       void **cookie)
+static inline int obd_get_lock(struct obd_export *exp,
+                               struct lov_stripe_md *lsm, void **res, int rw,
+                               obd_off start, obd_off end,
+                               struct lustre_handle *lockh, int flags)
 {
         ENTRY;
 
-        OBD_CHECK_OP(exp->exp_obd, reget_short_lock, -EOPNOTSUPP);
-        EXP_COUNTER_INCREMENT(exp, reget_short_lock);
+        OBD_CHECK_OP(exp->exp_obd, get_lock, -EOPNOTSUPP);
+        EXP_COUNTER_INCREMENT(exp, get_lock);
 
-        RETURN(OBP(exp->exp_obd, reget_short_lock)(exp, lsm, res, rw,
-                                                   start, end, cookie));
-}
-
-static inline int obd_release_short_lock(struct obd_export *exp,
-                                         struct lov_stripe_md *lsm, obd_off end,
-                                         void *cookie, int rw)
-{
-        ENTRY;
-
-        OBD_CHECK_OP(exp->exp_obd, release_short_lock, -EOPNOTSUPP);
-        EXP_COUNTER_INCREMENT(exp, release_short_lock);
-
-        RETURN(OBP(exp->exp_obd, release_short_lock)(exp, lsm, end,
-                                                     cookie, rw));
+        RETURN(OBP(exp->exp_obd, get_lock)(exp, lsm, res, rw, start, end,
+                                           lockh, flags));
 }
 
 static inline int obd_queue_async_io(struct obd_export *exp,
@@ -1124,7 +1272,7 @@ static inline int obd_teardown_async_page(struct obd_export *exp,
 
 static inline int obd_preprw(int cmd, struct obd_export *exp, struct obdo *oa,
                              int objcount, struct obd_ioobj *obj,
-                             int niocount, struct niobuf_remote *remote,
+                             struct niobuf_remote *remote, int *pages,
                              struct niobuf_local *local,
                              struct obd_trans_info *oti)
 {
@@ -1134,14 +1282,15 @@ static inline int obd_preprw(int cmd, struct obd_export *exp, struct obdo *oa,
         OBD_CHECK_OP(exp->exp_obd, preprw, -EOPNOTSUPP);
         EXP_COUNTER_INCREMENT(exp, preprw);
 
-        rc = OBP(exp->exp_obd, preprw)(cmd, exp, oa, objcount, obj, niocount,
-                                       remote, local, oti);
+        rc = OBP(exp->exp_obd, preprw)(cmd, exp, oa, objcount, obj, remote,
+                                       pages, local, oti);
         RETURN(rc);
 }
 
 static inline int obd_commitrw(int cmd, struct obd_export *exp, struct obdo *oa,
                                int objcount, struct obd_ioobj *obj,
-                               int niocount, struct niobuf_local *local,
+                               struct niobuf_remote *rnb, int pages,
+                               struct niobuf_local *local,
                                struct obd_trans_info *oti, int rc)
 {
         ENTRY;
@@ -1149,8 +1298,8 @@ static inline int obd_commitrw(int cmd, struct obd_export *exp, struct obdo *oa,
         OBD_CHECK_OP(exp->exp_obd, commitrw, -EOPNOTSUPP);
         EXP_COUNTER_INCREMENT(exp, commitrw);
 
-        rc = OBP(exp->exp_obd, commitrw)(cmd, exp, oa, objcount, obj, niocount,
-                                         local, oti, rc);
+        rc = OBP(exp->exp_obd, commitrw)(cmd, exp, oa, objcount, obj,
+                                         rnb, pages, local, oti, rc);
         RETURN(rc);
 }
 
@@ -1168,6 +1317,20 @@ static inline int obd_merge_lvb(struct obd_export *exp,
         RETURN(rc);
 }
 
+static inline int obd_update_lvb(struct obd_export *exp,
+                                 struct lov_stripe_md *lsm,
+                                 struct ost_lvb *lvb, obd_flag valid)
+{
+        int rc;
+        ENTRY;
+
+        OBD_CHECK_OP(exp->exp_obd, update_lvb, -EOPNOTSUPP);
+        EXP_COUNTER_INCREMENT(exp, update_lvb);
+
+        rc = OBP(exp->exp_obd, update_lvb)(exp, lsm, lvb, valid);
+        RETURN(rc);
+}
+
 static inline int obd_adjust_kms(struct obd_export *exp,
                                  struct lov_stripe_md *lsm, obd_off size,
                                  int shrink)
@@ -1234,7 +1397,8 @@ static inline int obd_enqueue(struct obd_export *exp,
 
 static inline int obd_match(struct obd_export *exp, struct lov_stripe_md *ea,
                             __u32 type, ldlm_policy_data_t *policy, __u32 mode,
-                            int *flags, void *data, struct lustre_handle *lockh)
+                            int *flags, void *data, struct lustre_handle *lockh,
+                            int *n_matches)
 {
         int rc;
         ENTRY;
@@ -1243,7 +1407,7 @@ static inline int obd_match(struct obd_export *exp, struct lov_stripe_md *ea,
         EXP_COUNTER_INCREMENT(exp, match);
 
         rc = OBP(exp->exp_obd, match)(exp, ea, type, policy, mode, flags, data,
-                                      lockh);
+                                      lockh, n_matches);
         RETURN(rc);
 }
 
@@ -1261,9 +1425,9 @@ static inline int obd_change_cbdata(struct obd_export *exp,
         RETURN(rc);
 }
 
-static inline int obd_cancel(struct obd_export *exp,
-                             struct lov_stripe_md *ea, __u32 mode,
-                             struct lustre_handle *lockh)
+static inline int obd_cancel(struct obd_export *exp, struct lov_stripe_md *ea,
+                             __u32 mode, struct lustre_handle *lockh, int flags,
+                             obd_off end)
 {
         int rc;
         ENTRY;
@@ -1271,7 +1435,7 @@ static inline int obd_cancel(struct obd_export *exp,
         EXP_CHECK_OP(exp, cancel);
         EXP_COUNTER_INCREMENT(exp, cancel);
 
-        rc = OBP(exp->exp_obd, cancel)(exp, ea, mode, lockh);
+        rc = OBP(exp->exp_obd, cancel)(exp, ea, mode, lockh, flags, end);
         RETURN(rc);
 }
 
@@ -1302,8 +1466,8 @@ static inline int obd_join_lru(struct obd_export *exp,
         RETURN(rc);
 }
 
-static inline int obd_pin(struct obd_export *exp, obd_id ino, __u32 gen,
-                          int type, struct obd_client_handle *handle, int flag)
+static inline int obd_pin(struct obd_export *exp, struct ll_fid *fid,
+                          struct obd_client_handle *handle, int flag)
 {
         int rc;
         ENTRY;
@@ -1311,7 +1475,7 @@ static inline int obd_pin(struct obd_export *exp, obd_id ino, __u32 gen,
         EXP_CHECK_OP(exp, pin);
         EXP_COUNTER_INCREMENT(exp, pin);
 
-        rc = OBP(exp->exp_obd, pin)(exp, ino, gen, type, handle, flag);
+        rc = OBP(exp->exp_obd, pin)(exp, fid, handle, flag);
         RETURN(rc);
 }
 
@@ -1483,9 +1647,34 @@ static inline int obd_register_observer(struct obd_device *obd,
 {
         ENTRY;
         OBD_CHECK_DEV(obd);
-        if (obd->obd_observer && observer)
+        down_write(&obd->obd_observer_link_sem);
+        if (obd->obd_observer && observer) {
+                up_write(&obd->obd_observer_link_sem);
                 RETURN(-EALREADY);
+        }
         obd->obd_observer = observer;
+        up_write(&obd->obd_observer_link_sem);
+        RETURN(0);
+}
+
+static inline int obd_pin_observer(struct obd_device *obd,
+                                   struct obd_device **observer)
+{
+        ENTRY;
+        down_read(&obd->obd_observer_link_sem);
+        if (!obd->obd_observer) {
+                *observer = NULL;
+                up_read(&obd->obd_observer_link_sem);
+                RETURN(-ENOENT);
+        }
+        *observer = obd->obd_observer;
+        RETURN(0);
+}
+
+static inline int obd_unpin_observer(struct obd_device *obd)
+{
+        ENTRY;
+        up_read(&obd->obd_observer_link_sem);
         RETURN(0);
 }
 
