@@ -154,19 +154,22 @@ void llu_update_inode(struct inode *inode, struct mds_body *body,
 
         if (body->valid & OBD_MD_FLID)
                 st->st_ino = body->ino;
-        if (body->valid & OBD_MD_FLATIME &&
-            body->atime > LTIME_S(st->st_atime))
-                LTIME_S(st->st_atime) = body->atime;
-        
-        /* mtime is always updated with ctime, but can be set in past.
-           As write and utime(2) may happen within 1 second, and utime's
-           mtime has a priority over write's one, so take mtime from mds 
-           for the same ctimes. */
-        if (body->valid & OBD_MD_FLCTIME &&
-            body->ctime >= LTIME_S(st->st_ctime)) {
-                LTIME_S(st->st_ctime) = body->ctime;
-                if (body->valid & OBD_MD_FLMTIME)
+        if (body->valid & OBD_MD_FLGENER)
+                lli->lli_st_generation = body->generation;
+        if (body->valid & OBD_MD_FLMTIME) {
+                if (body->mtime > LTIME_S(st->st_mtime))
                         LTIME_S(st->st_mtime) = body->mtime;
+                lli->lli_lvb.lvb_mtime = body->mtime;
+        }
+        if (body->valid & OBD_MD_FLATIME) {
+                if (body->atime > LTIME_S(st->st_atime))
+                        LTIME_S(st->st_atime) = body->atime;
+                lli->lli_lvb.lvb_atime = body->atime;
+        }
+        if (body->valid & OBD_MD_FLCTIME) {
+                if (body->ctime > LTIME_S(st->st_ctime))
+                        LTIME_S(st->st_ctime) = body->ctime;
+                lli->lli_lvb.lvb_ctime = body->ctime;
         }
         if (body->valid & OBD_MD_FLMODE)
                 st->st_mode = (st->st_mode & S_IFMT)|(body->mode & ~S_IFMT);
@@ -190,16 +193,8 @@ void llu_update_inode(struct inode *inode, struct mds_body *body,
                 st->st_blocks = body->blocks;
         if (body->valid & OBD_MD_FLFLAGS)
                 lli->lli_st_flags = body->flags;
-        if (body->valid & OBD_MD_FLGENER)
-                lli->lli_st_generation = body->generation;
 
-        /* fillin fid */
-        if (body->valid & OBD_MD_FLID)
-                lli->lli_fid.id = body->ino;
-        if (body->valid & OBD_MD_FLGENER)
-                lli->lli_fid.generation = body->generation;
-        if (body->valid & OBD_MD_FLTYPE)
-                lli->lli_fid.f_type = body->mode & S_IFMT;
+        lli->lli_fid = body->fid1;
 }
 
 void obdo_to_inode(struct inode *dst, struct obdo *src, obd_flag valid)
@@ -429,6 +424,8 @@ static int llu_have_md_lock(struct inode *inode, __u64 lockpart)
 static int llu_inode_revalidate(struct inode *inode)
 {
         struct lov_stripe_md *lsm = NULL;
+        struct llu_inode_info *lli = llu_i2info(inode);
+        struct intnl_stat *st = llu_i2stat(inode);
         ENTRY;
 
         if (!inode) {
@@ -446,15 +443,15 @@ static int llu_inode_revalidate(struct inode *inode)
 
                 /* Why don't we update all valid MDS fields here, if we're
                  * doing an RPC anyways?  -phil */
-                if (S_ISREG(llu_i2stat(inode)->st_mode)) {
+                if (S_ISREG(st->st_mode)) {
                         ealen = obd_size_diskmd(sbi->ll_osc_exp, NULL);
                         valid |= OBD_MD_FLEASIZE;
                 }
-                ll_inode2fid(&fid, inode);
+                llu_inode2fid(&fid, inode);
                 rc = mdc_getattr(sbi->ll_mdc_exp, &fid, valid, ealen, &req);
                 if (rc) {
                         CERROR("failure %d inode %llu\n", rc,
-                               (long long)llu_i2stat(inode)->st_ino);
+                               (long long)st->st_ino);
                         RETURN(-abs(rc));
                 }
                 rc = mdc_req2lustre_md(req, REPLY_REC_OFF, sbi->ll_osc_exp,&md);
@@ -475,18 +472,23 @@ static int llu_inode_revalidate(struct inode *inode)
 
 
                 llu_update_inode(inode, md.body, md.lsm);
-                if (md.lsm != NULL && llu_i2info(inode)->lli_smd != md.lsm)
+                if (md.lsm != NULL && lli->lli_smd != md.lsm)
                         obd_free_memmd(sbi->ll_osc_exp, &md.lsm);
 
                 if (md.body->valid & OBD_MD_FLSIZE)
                         set_bit(LLI_F_HAVE_MDS_SIZE_LOCK,
-                                &llu_i2info(inode)->lli_flags);
+                                &lli->lli_flags);
                 ptlrpc_req_finished(req);
         }
 
-        lsm = llu_i2info(inode)->lli_smd;
-        if (!lsm)       /* object not yet allocated, don't validate size */
+        lsm = lli->lli_smd;
+        if (!lsm) {
+                /* object not yet allocated, don't validate size */
+                st->st_atime = lli->lli_lvb.lvb_atime;
+                st->st_mtime = lli->lli_lvb.lvb_mtime;
+                st->st_ctime = lli->lli_lvb.lvb_ctime;
                 RETURN(0);
+        }
 
         /* ll_glimpse_size will prefer locally cached writes if they extend
          * the file */
@@ -551,7 +553,7 @@ void llu_clear_inode(struct inode *inode)
                (long long)llu_i2stat(inode)->st_ino, lli->lli_st_generation,
                inode);
 
-        ll_inode2fid(&fid, inode);
+        llu_inode2fid(&fid, inode);
         clear_bit(LLI_F_HAVE_MDS_SIZE_LOCK, &(lli->lli_flags));
         mdc_change_cbdata(sbi->ll_mdc_exp, &fid, null_if_equal, inode);
 
@@ -668,16 +670,6 @@ int llu_setattr_raw(struct inode *inode, struct iattr *attr)
                 attr->ia_mtime = CURRENT_TIME;
                 attr->ia_valid |= ATTR_MTIME_SET;
         }
-        if ((attr->ia_valid & ATTR_CTIME) && !(attr->ia_valid & ATTR_MTIME)) {
-                /* To avoid stale mtime on mds, obtain it from ost and send 
-                   to mds. */
-                rc = llu_glimpse_size(inode);
-                if (rc) 
-                        RETURN(rc);
-                
-                attr->ia_valid |= ATTR_MTIME_SET | ATTR_MTIME;
-                attr->ia_mtime = inode->i_stbuf.st_mtime;
-        }
 
         if (attr->ia_valid & (ATTR_MTIME | ATTR_CTIME))
                 CDEBUG(D_INODE, "setting mtime %lu, ctime %lu, now = %lu\n",
@@ -760,7 +752,8 @@ int llu_setattr_raw(struct inode *inode, struct iattr *attr)
 
                 /* check that there are no matching locks */
                 LASSERT(obd_match(sbi->ll_osc_exp, lsm, LDLM_EXTENT, &policy,
-                                  LCK_PW, &flags, inode, &match_lockh) <= 0);
+                                  LCK_PW, &flags, inode, &match_lockh, NULL)
+                                  <= 0);
 
                 /* XXX when we fix the AST intents to pass the discard-range
                  * XXX extent, make ast_flags always LDLM_AST_DISCARD_DATA
@@ -797,15 +790,87 @@ int llu_setattr_raw(struct inode *inode, struct iattr *attr)
                 }
         } else if (ia_valid & (ATTR_MTIME | ATTR_MTIME_SET)) {
                 struct obd_info oinfo = { { { 0 } } };
-                struct obdo oa;
+                struct obdo oa = { 0 };
+                struct lustre_handle lockh = { 0 };
+                obd_valid valid;
 
                 CDEBUG(D_INODE, "set mtime on OST inode %llu to %lu\n",
                        (long long)st->st_ino, LTIME_S(attr->ia_mtime));
+
                 oa.o_id = lsm->lsm_object_id;
                 oa.o_valid = OBD_MD_FLID;
 
-                obdo_from_inode(&oa, inode, OBD_MD_FLTYPE | OBD_MD_FLATIME |
-                                            OBD_MD_FLMTIME | OBD_MD_FLCTIME);
+                valid = OBD_MD_FLTYPE;
+
+                if (LTIME_S(attr->ia_mtime) < LTIME_S(attr->ia_ctime)){
+                        struct ost_lvb xtimes;
+
+                        /* setting mtime to past is performed under PW
+                         * EOF extent lock */
+                        oinfo.oi_policy.l_extent.start = 0;
+                        oinfo.oi_policy.l_extent.end = OBD_OBJECT_EOF;
+                        rc = llu_extent_lock(NULL, inode, lsm, LCK_PW,
+                                             &oinfo.oi_policy,
+                                             &lockh, 0);
+                        if (rc)
+                                RETURN(rc);
+
+                        /* setattr under locks
+                         *
+                         * 1. restore inode's timestamps which are
+                         * about to be set as long as concurrent stat
+                         * (via llu_glimpse_size) might bring
+                         * out-of-date ones
+                         *
+                         * 2. update lsm so that next stat (via
+                         * llu_glimpse_size) could get correct values
+                         * in lsm */
+                        lov_stripe_lock(lsm);
+                        if (ia_valid & ATTR_ATIME) {
+                                st->st_atime = xtimes.lvb_atime =
+                                        attr->ia_atime;
+                                valid |= OBD_MD_FLATIME;
+                        }
+                        if (ia_valid & ATTR_MTIME) {
+                                st->st_mtime = xtimes.lvb_mtime =
+                                        attr->ia_mtime;
+                                valid |= OBD_MD_FLMTIME;
+                        }
+                        if (ia_valid & ATTR_CTIME) {
+                                st->st_ctime = xtimes.lvb_ctime =
+                                        attr->ia_mtime;
+                                valid |= OBD_MD_FLCTIME;
+                        }
+
+                        obd_update_lvb(sbi->ll_osc_exp, lsm,
+                                       &xtimes, valid);
+                        lov_stripe_unlock(lsm);
+                } else {
+                        /* lockless setattr
+                         *
+                         * 1. do not use inode's timestamps because
+                         * concurrent stat might fill the inode with
+                         * out-of-date times, send values from attr
+                         * instead
+                         *
+                         * 2.do no update lsm, as long as stat (via
+                         * ll_glimpse_size) will bring attributes from
+                         * osts anyway */
+                        if (ia_valid & ATTR_ATIME) {
+                                oa.o_atime = attr->ia_atime;
+                                oa.o_valid |= OBD_MD_FLATIME;
+                        }
+                        if (ia_valid & ATTR_MTIME) {
+                                oa.o_mtime = attr->ia_mtime;
+                                oa.o_valid |= OBD_MD_FLMTIME;
+                        }
+                        if (ia_valid & ATTR_CTIME) {
+                                oa.o_ctime = attr->ia_ctime;
+                                oa.o_valid |= OBD_MD_FLCTIME;
+                        }
+                }
+
+                obdo_from_inode(&oa, inode, valid);
 
                 oinfo.oi_oa = &oa;
                 oinfo.oi_md = lsm;
@@ -813,6 +878,19 @@ int llu_setattr_raw(struct inode *inode, struct iattr *attr)
                 rc = obd_setattr_rqset(sbi->ll_osc_exp, &oinfo, NULL);
                 if (rc)
                         CERROR("obd_setattr_async fails: rc=%d\n", rc);
+
+                if (LTIME_S(attr->ia_mtime) < LTIME_S(attr->ia_ctime)){
+                        int err;
+
+                        err = llu_extent_unlock(NULL, inode, lsm,
+                                               LCK_PW, &lockh);
+                        if (unlikely(err != 0)) {
+                                CERROR("extent unlock failed: "
+                                       "err=%d\n", err);
+                                if (rc == 0)
+                                        rc = err;
+                        }
+                }
         }
         RETURN(rc);
 }
@@ -909,6 +987,7 @@ static int llu_readlink_internal(struct inode *inode,
         ENTRY;
 
         *request = NULL;
+        *symname = NULL;
 
         if (lli->lli_symlink_name) {
                 *symname = lli->lli_symlink_name;
@@ -916,7 +995,7 @@ static int llu_readlink_internal(struct inode *inode,
                 RETURN(0);
         }
 
-        ll_inode2fid(&fid, inode);
+        llu_inode2fid(&fid, inode);
         rc = mdc_getattr(sbi->ll_mdc_exp, &fid,
                          OBD_MD_LINKNAME, symlen, request);
         if (rc) {
@@ -1169,7 +1248,8 @@ static int llu_statfs(struct llu_sb_info *sbi, struct statfs *sfs)
         /* For now we will always get up-to-date statfs values, but in the
          * future we may allow some amount of caching on the client (e.g.
          * from QOS or lprocfs updates). */
-        rc = llu_statfs_internal(sbi, &osfs, cfs_time_current_64() - HZ);
+        rc = llu_statfs_internal(sbi, &osfs,
+                                 cfs_time_shift_64(-OBD_STATFS_CACHE_SECONDS));
         if (rc)
                 return rc;
 
@@ -1606,27 +1686,35 @@ static int llu_lov_dir_setstripe(struct inode *ino, unsigned long arg)
         struct ptlrpc_request *request = NULL;
         struct mdc_op_data op_data;
         struct iattr attr = { 0 };
-        struct lov_user_md lum, *lump = (struct lov_user_md *)arg;
-        int rc = 0;
+        struct lov_user_md_v3 lum;
+        struct lov_user_md *lump = (struct lov_user_md *)arg;
+        int rc = 0, lum_size = 0;
 
         llu_prepare_mdc_op_data(&op_data, ino, NULL, NULL, 0, 0);
 
-        LASSERT(sizeof(lum) == sizeof(*lump));
         LASSERT(sizeof(lum.lmm_objects[0]) ==
                 sizeof(lump->lmm_objects[0]));
-        rc = copy_from_user(&lum, lump, sizeof(lum));
+        rc = copy_from_user(&lum, lump, sizeof(*lump));
         if (rc)
                 return(-EFAULT);
+        lum_size = sizeof(struct lov_user_md_v1);
+        if (lum.lmm_magic == LOV_USER_MAGIC_V3) {
+                rc = copy_from_user(&lum, lump, sizeof(lum));
+                if (rc)
+                        return(-EFAULT);
+                lum_size = sizeof(struct lov_user_md_v3);
+        }
 
-        if (lum.lmm_magic != LOV_USER_MAGIC)
-                RETURN(-EINVAL);
-
-        if (lum.lmm_magic != cpu_to_le32(LOV_USER_MAGIC))
-                lustre_swab_lov_user_md(&lum);
+        if ((lum.lmm_magic != cpu_to_le32(LOV_USER_MAGIC_V1)) &&
+            (lum.lmm_magic != cpu_to_le32(LOV_USER_MAGIC_V3))) {
+                rc = lustre_swab_lov_user_md((struct lov_user_md_v1 *)&lum);
+                if (rc) 
+                        RETURN(rc);
+        }
 
         /* swabbing is done in lov_setstripe() on server side */
         rc = mdc_setattr(sbi->ll_mdc_exp, &op_data,
-                         &attr, &lum, sizeof(lum), NULL, 0, &request);
+                         &attr, &lum, lum_size, NULL, 0, &request);
         if (rc) {
                 ptlrpc_req_finished(request);
                 if (rc != -EPERM && rc != -EACCES)
@@ -1832,9 +1920,7 @@ struct inode *llu_iget(struct filesys *fs, struct lustre_md *md)
         }
 
         /* try to find existing inode */
-        fid.id = md->body->ino;
-        fid.generation = md->body->generation;
-        fid.f_type = md->body->mode & S_IFMT;
+        fid = md->body->fid1;
 
         inode = _sysio_i_find(fs, &fileid);
         if (inode) {
@@ -1956,7 +2042,7 @@ llu_fsswop_mount(const char *source,
                            sizeof(async), &async, NULL);
 
         ocd.ocd_connect_flags = OBD_CONNECT_IBITS | OBD_CONNECT_VERSION |
-                OBD_CONNECT_AT;
+                                OBD_CONNECT_AT | OBD_CONNECT_VBR;
 #ifdef LIBLUSTRE_POSIX_ACL
         ocd.ocd_connect_flags |= OBD_CONNECT_ACL;
 #endif
@@ -1990,6 +2076,10 @@ llu_fsswop_mount(const char *source,
         obd->obd_upcall.onu_owner = &sbi->ll_lco;
         obd->obd_upcall.onu_upcall = ll_ocd_update;
 
+        /* ask lov to generate OBD_NOTIFY_CREATE events for already registered
+         * targets */
+        obd_notify(obd, NULL, OBD_NOTIFY_CREATE, NULL);
+
         obd_register_lock_cancel_cb(obd, llu_extent_lock_cancel_cb);
 
         ocd.ocd_connect_flags = OBD_CONNECT_SRVLOCK | OBD_CONNECT_REQPORTAL |
