@@ -90,12 +90,15 @@ int it_open_error(int phase, struct lookup_intent *it)
 EXPORT_SYMBOL(it_open_error);
 
 /* this must be called on a lockh that is known to have a referenced lock */
-void mdc_set_lock_data(__u64 *l, void *data)
+void mdc_set_lock_data(__u64 *l, void *data, __u32 *bits)
 {
         struct ldlm_lock *lock;
         struct lustre_handle *lockh = (struct lustre_handle *)l;
         ENTRY;
 
+        if(bits)
+                *bits = 0;
+
         if (!*l) {
                 EXIT;
                 return;
@@ -118,6 +121,8 @@ void mdc_set_lock_data(__u64 *l, void *data)
         }
 #endif
         lock->l_ast_data = data;
+        if (bits)
+                *bits = lock->l_policy_data.l_inodebits.bits;
         unlock_res_and_lock(lock);
         LDLM_LOCK_PUT(lock);
 
@@ -128,12 +133,10 @@ EXPORT_SYMBOL(mdc_set_lock_data);
 int mdc_change_cbdata(struct obd_export *exp, struct ll_fid *fid, 
                       ldlm_iterator_t it, void *data)
 {
-        struct ldlm_res_id res_id = { .name = {0} };
+        struct ldlm_res_id res_id;
         ENTRY;
 
-        res_id.name[0] = fid->id;
-        res_id.name[1] = fid->generation;
-
+        fid_build_reg_res_name((struct lu_fid*)fid, &res_id);
         ldlm_resource_iterate(class_exp2obd(exp)->obd_namespace, &res_id,
                               it, data);
 
@@ -182,16 +185,26 @@ static void mdc_realloc_openmsg(struct ptlrpc_request *req,
         int old_len, new_size, old_size;
         struct lustre_msg *old_msg = req->rq_reqmsg;
         struct lustre_msg *new_msg;
+        int offset;
 
-        old_len = lustre_msg_buflen(old_msg, DLM_INTENT_REC_OFF + 2);
+        if (mdc_req_is_2_0_server(req))
+                offset = 4;
+        else
+                offset = 2;
+
+        old_len = lustre_msg_buflen(old_msg, DLM_INTENT_REC_OFF + offset);
         old_size = lustre_packed_msg_size(old_msg);
-        lustre_msg_set_buflen(old_msg, DLM_INTENT_REC_OFF + 2,
+        lustre_msg_set_buflen(old_msg, DLM_INTENT_REC_OFF + offset,
                               body->eadatasize);
+        /* old buffer is more then need */
+        if (old_len > body->eadatasize)
+                return;
+
         new_size = lustre_packed_msg_size(old_msg);
 
         OBD_ALLOC(new_msg, new_size);
         if (new_msg != NULL) {
-                DEBUG_REQ(D_INFO, req, "replace reqmsg for larger EA %u\n",
+                DEBUG_REQ(D_INFO, req, "replace reqmsg for larger EA %u",
                           body->eadatasize);
                 memcpy(new_msg, old_msg, old_size);
 
@@ -202,7 +215,8 @@ static void mdc_realloc_openmsg(struct ptlrpc_request *req,
 
                 OBD_FREE(old_msg, old_size);
         } else {
-                lustre_msg_set_buflen(old_msg, DLM_INTENT_REC_OFF + 2, old_len);
+                lustre_msg_set_buflen(old_msg,
+                                      DLM_INTENT_REC_OFF + offset, old_len);
                 body->valid &= ~OBD_MD_FLEASIZE;
                 body->eadatasize = 0;
         }
@@ -211,12 +225,12 @@ static void mdc_realloc_openmsg(struct ptlrpc_request *req,
 static struct ptlrpc_request *mdc_intent_open_pack(struct obd_export *exp,
                                                    struct lookup_intent *it,
                                                    struct mdc_op_data *data,
-                                                   void *lmm, unsigned lmmsize)
+                                                   void *lmm, __u32 lmmsize)
 {
         struct ptlrpc_request *req;
         struct ldlm_intent *lit;
         struct obd_device *obddev = class_exp2obd(exp);
-        __u32 size[7] = { [MSG_PTLRPC_BODY_OFF] = sizeof(struct ptlrpc_body),
+        __u32 size[9] = { [MSG_PTLRPC_BODY_OFF] = sizeof(struct ptlrpc_body),
                         [DLM_LOCKREQ_OFF]     = sizeof(struct ldlm_request),
                         [DLM_INTENT_IT_OFF]   = sizeof(*lit),
                         [DLM_INTENT_REC_OFF]  = sizeof(struct mds_rec_create),
@@ -230,27 +244,40 @@ static struct ptlrpc_request *mdc_intent_open_pack(struct obd_export *exp,
                          * default-sized LOV EA for open replay. */
                         [DLM_INTENT_REC_OFF+2]= max(lmmsize,
                                          obddev->u.cli.cl_default_mds_easize) };
-        __u32 repsize[5] = { [MSG_PTLRPC_BODY_OFF] = sizeof(struct ptlrpc_body),
+        __u32 repsize[7] = { [MSG_PTLRPC_BODY_OFF] = sizeof(struct ptlrpc_body),
                            [DLM_LOCKREPLY_OFF]   = sizeof(struct ldlm_reply),
-                           [DLM_REPLY_REC_OFF]   = sizeof(struct mds_body),
+                           [DLM_REPLY_REC_OFF]   = sizeof(struct mdt_body),
                            [DLM_REPLY_REC_OFF+1] = obddev->u.cli.
                                                         cl_max_mds_easize,
                            [DLM_REPLY_REC_OFF+2] = LUSTRE_POSIX_ACL_MAX_SIZE };
         CFS_LIST_HEAD(cancels);
-        int do_join = (it->it_flags & O_JOIN_FILE) && data->data;
+        int do_join = (it->it_create_mode & M_JOIN_FILE) && data->data;
         int count = 0;
+        int bufcount = 6;
+        int repbufcount = 5;
         int mode;
         int rc;
+        ENTRY;
 
-        it->it_create_mode |= S_IFREG;
-
-        rc = lustre_msg_size(class_exp2cliimp(exp)->imp_msg_magic, 6, size);
+        it->it_create_mode = (it->it_create_mode & ~S_IFMT) | S_IFREG;
+        if (mdc_exp_is_2_0_server(exp)) {
+                size[DLM_INTENT_REC_OFF] = sizeof(struct mdt_rec_create);
+                size[DLM_INTENT_REC_OFF+4] = size[DLM_INTENT_REC_OFF+2];
+                size[DLM_INTENT_REC_OFF+3] = size[DLM_INTENT_REC_OFF+1];
+                size[DLM_INTENT_REC_OFF+2] = 0; /* capa */
+                size[DLM_INTENT_REC_OFF+1] = 0; /* capa */
+                bufcount = 8;
+                repsize[DLM_REPLY_REC_OFF+3]=sizeof(struct lustre_capa);
+                repsize[DLM_REPLY_REC_OFF+4]=sizeof(struct lustre_capa);
+                repbufcount = 7;
+        }
+        rc = lustre_msg_size(class_exp2cliimp(exp)->imp_msg_magic,
+                             bufcount, size);
         if (rc & (rc - 1))
-                size[DLM_INTENT_REC_OFF + 2] =
-                         min(size[DLM_INTENT_REC_OFF + 2] + round_up(rc) - rc,
-                                     obddev->u.cli.cl_max_mds_easize);
+                size[bufcount - 1] = min(size[bufcount - 1] + round_up(rc) - rc,
+                                         (__u32)obddev->u.cli.cl_max_mds_easize);
 
-                /* If inode is known, cancel conflicting OPEN locks. */
+        /* If inode is known, cancel conflicting OPEN locks. */
         if (data->fid2.id) {
                 if (it->it_flags & (FMODE_WRITE|MDS_OPEN_TRUNC))
                         mode = LCK_CW;
@@ -273,20 +300,25 @@ static struct ptlrpc_request *mdc_intent_open_pack(struct obd_export *exp,
                                          MDS_INODELOCK_UPDATE);
         if (do_join) {
                 __u64 head_size = (*(__u64 *)data->data);
-                        /* join is like an unlink of the tail */
-                size[DLM_INTENT_REC_OFF + 3] = sizeof(struct mds_rec_join);
-                req = ldlm_prep_enqueue_req(exp, 7, size, &cancels, count);
+                /* join is like an unlink of the tail */
+                if (mdc_exp_is_2_0_server(exp)) {
+                        size[DLM_INTENT_REC_OFF+5]=sizeof(struct mdt_rec_join);
+                } else {
+                        size[DLM_INTENT_REC_OFF+3]=sizeof(struct mds_rec_join);
+                }
+                bufcount++;
+
+                req = ldlm_prep_enqueue_req(exp, bufcount, size,&cancels,count);
                 if (req)
-                        mdc_join_pack(req, DLM_INTENT_REC_OFF + 3, data, 
-                                      head_size);
+                        mdc_join_pack(req, bufcount - 1, data, head_size);
         } else {
-                req = ldlm_prep_enqueue_req(exp, 6, size, &cancels, count);
-                it->it_flags &= ~O_JOIN_FILE;
+                req = ldlm_prep_enqueue_req(exp, bufcount, size,&cancels,count);
+                it->it_create_mode &= ~M_JOIN_FILE;
         }
 
         if (req) {
                 spin_lock(&req->rq_lock);
-                req->rq_replay = 1;
+                req->rq_replay = req->rq_import->imp_replayable;
                 spin_unlock(&req->rq_lock);
 
                 /* pack the intent */
@@ -299,9 +331,9 @@ static struct ptlrpc_request *mdc_intent_open_pack(struct obd_export *exp,
                               it->it_create_mode, 0, it->it_flags,
                               lmm, lmmsize);
 
-                ptlrpc_req_set_repsize(req, 5, repsize);
+                ptlrpc_req_set_repsize(req, repbufcount, repsize);
         }
-        return req;
+        RETURN(req);
 }
 
 static struct ptlrpc_request *mdc_intent_unlink_pack(struct obd_export *exp,
@@ -314,15 +346,18 @@ static struct ptlrpc_request *mdc_intent_unlink_pack(struct obd_export *exp,
         __u32 size[5] = { [MSG_PTLRPC_BODY_OFF] = sizeof(struct ptlrpc_body),
                         [DLM_LOCKREQ_OFF]     = sizeof(struct ldlm_request),
                         [DLM_INTENT_IT_OFF]   = sizeof(*lit),
-                        [DLM_INTENT_REC_OFF]  = sizeof(struct mds_rec_unlink),
+                        [DLM_INTENT_REC_OFF]  = mdc_exp_is_2_0_server(exp) ?
+                                                sizeof(struct mdt_rec_unlink) :
+                                                sizeof(struct mds_rec_unlink),
                         [DLM_INTENT_REC_OFF+1]= data->namelen + 1 };
         __u32 repsize[5] = { [MSG_PTLRPC_BODY_OFF] = sizeof(struct ptlrpc_body),
                            [DLM_LOCKREPLY_OFF]   = sizeof(struct ldlm_reply),
-                           [DLM_REPLY_REC_OFF]   = sizeof(struct mds_body),
+                           [DLM_REPLY_REC_OFF]   = sizeof(struct mdt_body),
                            [DLM_REPLY_REC_OFF+1] = obddev->u.cli.
                                                         cl_max_mds_easize,
                            [DLM_REPLY_REC_OFF+2] = obddev->u.cli.
                                                         cl_max_mds_cookiesize };
+        ENTRY;
 
         req = ldlm_prep_enqueue_req(exp, 5, size, NULL, 0);
         if (req) {
@@ -336,7 +371,7 @@ static struct ptlrpc_request *mdc_intent_unlink_pack(struct obd_export *exp,
 
                 ptlrpc_req_set_repsize(req, 5, repsize);
         }
-        return req;
+        RETURN(req);
 }
 
 static struct ptlrpc_request *mdc_intent_lookup_pack(struct obd_export *exp,
@@ -346,21 +381,30 @@ static struct ptlrpc_request *mdc_intent_lookup_pack(struct obd_export *exp,
         struct ptlrpc_request *req;
         struct ldlm_intent *lit;
         struct obd_device *obddev = class_exp2obd(exp);
-        __u32 size[5] = { [MSG_PTLRPC_BODY_OFF] = sizeof(struct ptlrpc_body),
+        __u32 size[6] = { [MSG_PTLRPC_BODY_OFF] = sizeof(struct ptlrpc_body),
                         [DLM_LOCKREQ_OFF]     = sizeof(struct ldlm_request),
                         [DLM_INTENT_IT_OFF]   = sizeof(*lit),
-                        [DLM_INTENT_REC_OFF]  = sizeof(struct mds_body),
-                        [DLM_INTENT_REC_OFF+1]= data->namelen + 1 };
-        __u32 repsize[5] = { [MSG_PTLRPC_BODY_OFF] = sizeof(struct ptlrpc_body),
+                        [DLM_INTENT_REC_OFF]  = sizeof(struct mdt_body),
+                        [DLM_INTENT_REC_OFF+1]= data->namelen + 1,
+                        [DLM_INTENT_REC_OFF+2]= 0 };
+        __u32 repsize[6] = { [MSG_PTLRPC_BODY_OFF] = sizeof(struct ptlrpc_body),
                            [DLM_LOCKREPLY_OFF]   = sizeof(struct ldlm_reply),
-                           [DLM_REPLY_REC_OFF]   = sizeof(struct mds_body),
+                           [DLM_REPLY_REC_OFF]   = sizeof(struct mdt_body),
                            [DLM_REPLY_REC_OFF+1] = obddev->u.cli.
                                                         cl_max_mds_easize,
-                           [DLM_REPLY_REC_OFF+2] = LUSTRE_POSIX_ACL_MAX_SIZE };
+                           [DLM_REPLY_REC_OFF+2] = LUSTRE_POSIX_ACL_MAX_SIZE,
+                           [DLM_REPLY_REC_OFF+3] = 0 };
         obd_valid valid = OBD_MD_FLGETATTR | OBD_MD_FLEASIZE | OBD_MD_FLACL |
                           OBD_MD_FLMODEASIZE | OBD_MD_FLDIREA;
+        int bufcount = 5;
+        ENTRY;
 
-                req = ldlm_prep_enqueue_req(exp, 5, size, NULL, 0);
+        if (mdc_exp_is_2_0_server(exp)) {
+                size[DLM_INTENT_REC_OFF+1] = 0; /* capa */
+                size[DLM_INTENT_REC_OFF+2] = data->namelen + 1;
+                bufcount = 6;
+        }
+        req = ldlm_prep_enqueue_req(exp, bufcount, size, NULL, 0);
         if (req) {
                 /* pack the intent */
                 lit = lustre_msg_buf(req->rq_reqmsg, DLM_INTENT_IT_OFF,
@@ -370,9 +414,9 @@ static struct ptlrpc_request *mdc_intent_lookup_pack(struct obd_export *exp,
                 /* pack the intended request */
                 mdc_getattr_pack(req, DLM_INTENT_REC_OFF, valid, it->it_flags,
                                  data);
-                ptlrpc_req_set_repsize(req, 5, repsize);
+                ptlrpc_req_set_repsize(req, bufcount, repsize);
         }
-        return req;
+        RETURN(req);
 }
 
 static struct ptlrpc_request *mdc_intent_readdir_pack(struct obd_export *exp)
@@ -380,13 +424,15 @@ static struct ptlrpc_request *mdc_intent_readdir_pack(struct obd_export *exp)
         struct ptlrpc_request *req;
         __u32 size[2] = { [MSG_PTLRPC_BODY_OFF] = sizeof(struct ptlrpc_body),
                         [DLM_LOCKREQ_OFF]     = sizeof(struct ldlm_request) };
-        __u32 repsize[2] = { [MSG_PTLRPC_BODY_OFF] = sizeof(struct ptlrpc_body),
-                           [DLM_LOCKREPLY_OFF]   = sizeof(struct ldlm_reply) };
+        __u32 repsize[3] = { [MSG_PTLRPC_BODY_OFF] = sizeof(struct ptlrpc_body),
+                           [DLM_LOCKREPLY_OFF]   = sizeof(struct ldlm_reply),
+                           [DLM_REPLY_REC_OFF] = sizeof(struct ost_lvb) };
+        ENTRY;
 
         req = ldlm_prep_enqueue_req(exp, 2, size, NULL, 0);
         if (req)
-                ptlrpc_req_set_repsize(req, 2, repsize);
-        return req;
+                ptlrpc_req_set_repsize(req, 3, repsize);
+        RETURN(req);
 }
 
 static int mdc_finish_enqueue(struct obd_export *exp,
@@ -429,13 +475,14 @@ static int mdc_finish_enqueue(struct obd_export *exp,
 
         lockrep = lustre_msg_buf(req->rq_repmsg, DLM_LOCKREPLY_OFF,
                                  sizeof(*lockrep));
-        LASSERT(lockrep != NULL);                 /* checked by ldlm_cli_enqueue() */
+        LASSERT(lockrep != NULL);  /* checked by ldlm_cli_enqueue() */
         /* swabbed by ldlm_cli_enqueue() */
         LASSERT(lustre_rep_swabbed(req, DLM_LOCKREPLY_OFF));
 
         it->d.lustre.it_disposition = (int)lockrep->lock_policy_res1;
         it->d.lustre.it_status = (int)lockrep->lock_policy_res2;
         it->d.lustre.it_lock_mode = einfo->ei_mode;
+        it->d.lustre.it_lock_handle = lockh->cookie;
         it->d.lustre.it_data = req;
 
         if (it->d.lustre.it_status < 0 && req->rq_replay)
@@ -479,6 +526,8 @@ static int mdc_finish_enqueue(struct obd_export *exp,
                 if ((body->valid & OBD_MD_FLEASIZE) != 0) {
                         void *eadata;
 
+                        mdc_update_max_ea_from_body(exp, body);
+
                         /* The eadata is opaque; just check that it is there.
                          * Eventually, obd_unpackmd() will check the contents */
                         eadata = lustre_swab_repbuf(req, DLM_REPLY_REC_OFF + 1,
@@ -492,10 +541,15 @@ static int mdc_finish_enqueue(struct obd_export *exp,
                          * large enough request buffer above we need to
                          * reallocate it here to hold the actual LOV EA. */
                         if (it->it_op & IT_OPEN) {
-                                int offset = DLM_INTENT_REC_OFF + 2;
+                                int offset = DLM_INTENT_REC_OFF;
                                 void *lmm;
 
-                                if (lustre_msg_buflen(req->rq_reqmsg, offset) <
+                                if (mdc_req_is_2_0_server(req))
+                                        offset += 4;
+                                else
+                                        offset += 2;
+
+                                if (lustre_msg_buflen(req->rq_reqmsg, offset) !=
                                     body->eadatasize)
                                         mdc_realloc_openmsg(req, body);
 
@@ -519,20 +573,29 @@ int mdc_enqueue(struct obd_export *exp, struct ldlm_enqueue_info *einfo,
 {
         struct ptlrpc_request *req;
         struct obd_device *obddev = class_exp2obd(exp);
-        struct ldlm_res_id res_id =
-                { .name = {data->fid1.id, data->fid1.generation} };
+        struct ldlm_res_id res_id;
         ldlm_policy_data_t policy = { .l_inodebits = { MDS_INODELOCK_LOOKUP } };
         int flags = extra_lock_flags | LDLM_FL_HAS_INTENT;
         int rc;
         ENTRY;
 
+        fid_build_reg_res_name((void *)&data->fid1, &res_id);
         LASSERTF(einfo->ei_type == LDLM_IBITS,"lock type %d\n", einfo->ei_type);
         if (it->it_op & (IT_UNLINK | IT_GETATTR | IT_READDIR))
                 policy.l_inodebits.bits = MDS_INODELOCK_UPDATE;
 
         if (it->it_op & IT_OPEN) {
+                if ((it->it_op & IT_CREAT) && mdc_exp_is_2_0_server(exp)) {
+                        struct client_obd *cli = &obddev->u.cli;
+                        data->fid3 = data->fid2;
+                        rc = mdc_fid_alloc(cli->cl_seq, (void *)&data->fid2);
+                        if (rc) {
+                                CERROR("fid allocation result: %d\n", rc);
+                                RETURN(rc);
+                        }
+                }
                 req = mdc_intent_open_pack(exp, it, data, lmm, lmmsize);
-                if (it->it_flags & O_JOIN_FILE) {
+                if (it->it_create_mode & M_JOIN_FILE) {
                         policy.l_inodebits.bits = MDS_INODELOCK_UPDATE;
                 }
         } else if (it->it_op & IT_UNLINK) {
@@ -576,11 +639,13 @@ int mdc_revalidate_lock(struct obd_export *exp, struct lookup_intent *it,
                 /* We could just return 1 immediately, but since we should only
                  * be called in revalidate_it if we already have a lock, let's
                  * verify that. */
-        struct ldlm_res_id res_id = {.name ={fid->id, fid->generation}};
+        struct ldlm_res_id res_id;
         struct lustre_handle lockh;
         ldlm_policy_data_t policy;
         ldlm_mode_t mode;
+        ENTRY;
 
+        fid_build_reg_res_name((struct lu_fid*)fid, &res_id);
         /* As not all attributes are kept under update lock, e.g. 
            owner/group/acls are under lookup lock, we need both 
            ibits for GETATTR. */
@@ -596,7 +661,7 @@ int mdc_revalidate_lock(struct obd_export *exp, struct lookup_intent *it,
                 it->d.lustre.it_lock_mode = mode;
         }
 
-        return !!mode;
+        RETURN(!!mode);
 }
 EXPORT_SYMBOL(mdc_revalidate_lock);
 
@@ -635,10 +700,15 @@ static int mdc_finish_intent_lock(struct obd_export *exp,
 
         /* If we were revalidating a fid/name pair, mark the intent in
          * case we fail and get called again from lookup */
-        if (data->fid2.id && (it->it_op != IT_GETATTR)) {
+
+        if (data->fid2.id && (it->it_op != IT_GETATTR) &&
+           ( !mdc_exp_is_2_0_server(exp) ||
+             (mdc_exp_is_2_0_server(exp) && (it->it_create_mode & M_CHECK_STALE)))) {
                 it_set_disposition(it, DISP_ENQ_COMPLETE);
+
                 /* Also: did we find the same inode? */
-                if (memcmp(&data->fid2, &mds_body->fid1, sizeof(data->fid2))) 
+                if (memcmp(&data->fid2, &mds_body->fid1, sizeof(data->fid2)) &&
+                    memcmp(&data->fid3, &mds_body->fid1, sizeof(data->fid3)))
                         RETURN(-ESTALE);
         }
 
@@ -738,8 +808,11 @@ int mdc_intent_lock(struct obd_export *exp, struct mdc_op_data *op_data,
 
         LASSERT(it);
 
-        CDEBUG(D_DLMTRACE,"name: %.*s in inode "LPU64", intent: %s flags %#o\n",
-               op_data->namelen, op_data->name, op_data->fid1.id,
+        CDEBUG(D_DLMTRACE,"name: %.*s("DFID") in inode ("DFID"), "
+               "intent: %s flags %#o\n",
+               op_data->namelen, op_data->name,
+               PFID(((void *)&op_data->fid2)),
+               PFID(((void *)&op_data->fid1)),
                ldlm_it2str(it->it_op), it->it_flags);
 
         lockh.cookie = 0;
@@ -769,7 +842,6 @@ int mdc_intent_lock(struct obd_export *exp, struct mdc_op_data *op_data,
                                  lmm, lmmsize, extra_lock_flags);
                 if (rc < 0)
                         RETURN(rc);
-                memcpy(&it->d.lustre.it_lock_handle, &lockh, sizeof(lockh));
         } else if (!op_data->fid2.id) {
                 /* DISP_ENQ_COMPLETE set means there is extra reference on
                  * request referenced from this intent, saved for subsequent
@@ -818,8 +890,6 @@ static int mdc_intent_getattr_async_interpret(struct ptlrpc_request *req,
         if (rc)
                 GOTO(out, rc);
 
-        memcpy(&it->d.lustre.it_lock_handle, lockh, sizeof(*lockh));
-
         rc = mdc_finish_intent_lock(exp, req, &minfo->mi_data, it, lockh);
         GOTO(out, rc);
 out:
@@ -837,10 +907,7 @@ int mdc_intent_getattr_async(struct obd_export *exp,
         struct lookup_intent    *it = &minfo->mi_it;
         struct ptlrpc_request   *req;
         struct obd_device       *obddev = class_exp2obd(exp);
-        struct ldlm_res_id       res_id = {
-                                        .name = {op_data->fid1.id,
-                                                 op_data->fid1.generation}
-                                 };
+        struct ldlm_res_id res_id;
         ldlm_policy_data_t       policy = {
                                         .l_inodebits = { MDS_INODELOCK_LOOKUP }
                                  };
@@ -852,6 +919,7 @@ int mdc_intent_getattr_async(struct obd_export *exp,
                op_data->namelen, op_data->name, op_data->fid1.id,
                ldlm_it2str(it->it_op), it->it_flags);
 
+        fid_build_reg_res_name((void *)&op_data->fid1, &res_id);
         req = mdc_intent_lookup_pack(exp, it, op_data);
         if (!req)
                 RETURN(-ENOMEM);
