@@ -66,22 +66,25 @@ static int mdc_cleanup(struct obd_device *obd);
 extern int mds_queue_req(struct ptlrpc_request *);
 /* Helper that implements most of mdc_getstatus and signal_completed_replay. */
 /* XXX this should become mdc_get_info("key"), sending MDS_GET_INFO RPC */
-static int send_getstatus(struct obd_import *imp, struct ll_fid *rootfid,
+static int send_getstatus(struct obd_export *exp, struct ll_fid *rootfid,
                           int level, int msg_flags)
 {
         struct ptlrpc_request *req;
         struct mds_body *body;
-        __u32 size[2] = { sizeof(struct ptlrpc_body), sizeof(*body) };
+        __u32 size[3] = { sizeof(struct ptlrpc_body),
+                          sizeof(struct mdt_body),
+                          sizeof(struct lustre_capa) };
         int rc;
         ENTRY;
 
-        req = ptlrpc_prep_req(imp, LUSTRE_MDS_VERSION, MDS_GETSTATUS, 2, size,
+        req = ptlrpc_prep_req(class_exp2cliimp(exp), LUSTRE_MDS_VERSION, MDS_GETSTATUS, 2, size,
                               NULL);
         if (!req)
                 GOTO(out, rc = -ENOMEM);
 
+        req->rq_export = class_export_get(exp);
         req->rq_send_state = level;
-        ptlrpc_req_set_repsize(req, 2, size);
+        ptlrpc_req_set_repsize(req, 3, size);
 
         mdc_pack_req_body(req, REQ_REC_OFF, 0, NULL, 0, 0);
         lustre_msg_add_flags(req->rq_reqmsg, msg_flags);
@@ -110,11 +113,10 @@ static int send_getstatus(struct obd_import *imp, struct ll_fid *rootfid,
         return rc;
 }
 
-/* This should be mdc_get_info("rootfid") */
+/* This should be mdc_get_info("ROOT") */
 int mdc_getstatus(struct obd_export *exp, struct ll_fid *rootfid)
 {
-        return send_getstatus(class_exp2cliimp(exp), rootfid, LUSTRE_IMP_FULL,
-                              0);
+        return send_getstatus(exp, rootfid, LUSTRE_IMP_FULL, 0);
 }
 
 static
@@ -124,12 +126,12 @@ int mdc_getattr_common(struct obd_export *exp, unsigned int ea_size,
         struct obd_device *obddev = class_exp2obd(exp);
         struct mds_body *body;
         void *eadata;
-        __u32 size[4] = { sizeof(struct ptlrpc_body), sizeof(*body) };
+        __u32 size[6] = { sizeof(struct ptlrpc_body),
+                          sizeof(struct mdt_body) };
         int bufcount = 2, rc;
         ENTRY;
 
         /* request message already built */
-
         if (ea_size != 0) {
                 size[bufcount++] = ea_size;
                 CDEBUG(D_INODE, "reserved %u bytes for MD/symlink in packet\n",
@@ -140,6 +142,10 @@ int mdc_getattr_common(struct obd_export *exp, unsigned int ea_size,
                 CDEBUG(D_INODE, "reserved %u bytes for ACL\n", acl_size);
         }
 
+        if (mdc_exp_is_2_0_server(exp)) {
+                bufcount = 6;
+        }
+
         ptlrpc_req_set_repsize(req, bufcount, size);
 
         mdc_enter_request(&obddev->u.cli);
@@ -158,6 +164,8 @@ int mdc_getattr_common(struct obd_export *exp, unsigned int ea_size,
         CDEBUG(D_NET, "mode: %o\n", body->mode);
 
         lustre_set_rep_swabbed(req, REPLY_REC_OFF + 1);
+        mdc_update_max_ea_from_body(exp, body);
+
         if (body->eadatasize != 0) {
                 /* reply indicates presence of eadata; check it's there... */
                 eadata = lustre_msg_buf(req->rq_repmsg, REPLY_REC_OFF + 1,
@@ -176,7 +184,8 @@ int mdc_getattr(struct obd_export *exp, struct ll_fid *fid,
                 struct ptlrpc_request **request)
 {
         struct ptlrpc_request *req;
-        __u32 size[2] = { sizeof(struct ptlrpc_body), sizeof(struct mds_body) };
+        __u32 size[2] = { sizeof(struct ptlrpc_body),
+                          sizeof(struct mdt_body) };
         int acl_size = 0, rc;
         ENTRY;
 
@@ -188,6 +197,7 @@ int mdc_getattr(struct obd_export *exp, struct ll_fid *fid,
         if (!req)
                 GOTO(out, rc = -ENOMEM);
 
+        req->rq_export = class_export_get(exp);
         mdc_pack_req_body(req, REQ_REC_OFF, valid, fid, ea_size,
                           MDS_BFLAG_EXT_FLAGS/*request "new" flags(bug 9486)*/);
 
@@ -210,21 +220,32 @@ int mdc_getattr_name(struct obd_export *exp, struct ll_fid *fid,
                      unsigned int ea_size, struct ptlrpc_request **request)
 {
         struct ptlrpc_request *req;
-        struct mds_body *body;
-        __u32 size[3] = { sizeof(struct ptlrpc_body), sizeof(*body), namelen};
+        __u32 size[4] = { [MSG_PTLRPC_BODY_OFF] = sizeof(struct ptlrpc_body),
+                          [REQ_REC_OFF] = sizeof(struct mdt_body),
+                          [REQ_REC_OFF + 1] = namelen };
         int rc;
+        int bufcount = 3;
+        int nameoffset = REQ_REC_OFF + 1;
         ENTRY;
 
+        if (mdc_exp_is_2_0_server(exp)) {
+                size[REQ_REC_OFF + 1] = 0;
+                size[REQ_REC_OFF + 2] = namelen;
+                bufcount ++;
+                nameoffset ++;
+        }
+
         req = ptlrpc_prep_req(class_exp2cliimp(exp), LUSTRE_MDS_VERSION,
-                              MDS_GETATTR_NAME, 3, size, NULL);
+                              MDS_GETATTR_NAME, bufcount, size, NULL);
         if (!req)
                 GOTO(out, rc = -ENOMEM);
 
+        req->rq_export = class_export_get(exp);
         mdc_pack_req_body(req, REQ_REC_OFF, valid, fid, ea_size,
                           MDS_BFLAG_EXT_FLAGS/*request "new" flags(bug 9486)*/);
 
         LASSERT(strnlen(filename, namelen) == namelen - 1);
-        memcpy(lustre_msg_buf(req->rq_reqmsg, REQ_REC_OFF + 1, namelen),
+        memcpy(lustre_msg_buf(req->rq_reqmsg, nameoffset, namelen),
                filename, namelen);
 
         rc = mdc_getattr_common(exp, ea_size, 0, req);
@@ -245,12 +266,24 @@ int mdc_xattr_common(struct obd_export *exp, struct ll_fid *fid,
 {
         struct obd_device *obddev = class_exp2obd(exp);
         struct ptlrpc_request *req;
-        __u32 size[4] = { sizeof(struct ptlrpc_body), sizeof(struct mds_body) };
-        // int size[3] = {sizeof(struct mds_body)}, bufcnt = 1;
-        int rc, xattr_namelen = 0, bufcnt = 2, offset;
+        __u32 size[5] = { [MSG_PTLRPC_BODY_OFF] = sizeof(struct ptlrpc_body),
+                        [REQ_REC_OFF] = sizeof(struct mdt_body),
+                        [REQ_REC_OFF + 1] = 0, /* capa */
+                        [REQ_REC_OFF + 2] = 0, /* name */
+                        [REQ_REC_OFF + 3] = 0 };
+        int rc, xattr_namelen = 0, bufcnt = 2, offset = REQ_REC_OFF + 1;
         void *tmp;
         ENTRY;
 
+        if (mdc_exp_is_2_0_server(exp)) {
+                bufcnt++;
+                offset++;
+                if (opcode == MDS_SETXATTR) {
+                        size[REQ_REC_OFF] = sizeof (struct mdt_rec_setxattr);
+                        opcode = MDS_REINT;
+                }
+        }
+
         if (xattr_name) {
                 xattr_namelen = strlen(xattr_name) + 1;
                 size[bufcnt++] = xattr_namelen;
@@ -265,10 +298,26 @@ int mdc_xattr_common(struct obd_export *exp, struct ll_fid *fid,
         if (!req)
                 GOTO(out, rc = -ENOMEM);
 
-        /* request data */
-        mdc_pack_req_body(req, REQ_REC_OFF, valid, fid, output_size, flags);
+        req->rq_export = class_export_get(exp);
 
-        offset = REQ_REC_OFF + 1;
+        if (opcode == MDS_REINT && mdc_exp_is_2_0_server(exp)) {
+                struct mdt_rec_setxattr *rec;
+                rec = lustre_msg_buf(req->rq_reqmsg, REQ_REC_OFF,
+                                     sizeof(struct mdt_rec_setxattr));
+                rec->sx_opcode = REINT_SETXATTR;
+                rec->sx_fsuid  = cfs_curproc_fsuid();
+                rec->sx_fsgid  = cfs_curproc_fsgid();
+                rec->sx_cap    = cfs_curproc_cap_pack();
+                rec->sx_suppgid1 = -1;
+                rec->sx_suppgid2 = -1;
+                rec->sx_fid    = *((struct lu_fid*)fid);
+                rec->sx_valid  = valid;
+                rec->sx_size   = output_size;
+                rec->sx_flags  = flags;
+        } else {
+                /* request data */
+                mdc_pack_req_body(req, REQ_REC_OFF, valid, fid, output_size, flags);
+        }
 
         if (xattr_name) {
                 tmp = lustre_msg_buf(req->rq_reqmsg, offset++, xattr_namelen);
@@ -279,28 +328,32 @@ int mdc_xattr_common(struct obd_export *exp, struct ll_fid *fid,
                 memcpy(tmp, input, input_size);
         }
 
-        /* reply buffers */
-        if (opcode == MDS_GETXATTR) {
-                size[REPLY_REC_OFF] = sizeof(struct mds_body);
+        size[REPLY_REC_OFF] = sizeof(struct mdt_body);
+        if (mdc_exp_is_2_0_server(exp)) {
                 bufcnt = 2;
         } else {
-                bufcnt = 1;
+                /* reply buffers */
+                if (opcode == MDS_GETXATTR) {
+                        bufcnt = 2;
+                } else {
+                        bufcnt = 1;
+                }
+
         }
 
         /* we do this even output_size is 0, because server is doing that */
         size[bufcnt++] = output_size;
-
         ptlrpc_req_set_repsize(req, bufcnt, size);
 
         /* make rpc */
-        if (opcode == MDS_SETXATTR)
+        if (opcode == MDS_SETXATTR || opcode == MDS_REINT)
                 mdc_get_rpc_lock(exp->exp_obd->u.cli.cl_rpc_lock, NULL);
         else
                 mdc_enter_request(&obddev->u.cli);
 
         rc = ptlrpc_queue_wait(req);
 
-        if (opcode == MDS_SETXATTR)
+        if (opcode == MDS_SETXATTR || opcode == MDS_REINT)
                 mdc_put_rpc_lock(exp->exp_obd->u.cli.cl_rpc_lock, NULL);
         else
                 mdc_exit_request(&obddev->u.cli);
@@ -345,9 +398,9 @@ int mdc_getxattr(struct obd_export *exp, struct ll_fid *fid,
                                 input, input_size, output_size, 0, request);
 }
 
-/* This should be called with both the request and the reply still packed. */
-void mdc_store_inode_generation(struct ptlrpc_request *req, int reqoff,
-                                int repoff)
+/* For the fid-less server */
+static void mdc_store_inode_generation_18(struct ptlrpc_request *req,
+                                          int reqoff, int repoff)
 {
         struct mds_rec_create *rec = lustre_msg_buf(req->rq_reqmsg, reqoff,
                                                     sizeof(*rec));
@@ -368,6 +421,41 @@ void mdc_store_inode_generation(struct ptlrpc_request *req, int reqoff,
                   rec->cr_replayfid.generation, rec->cr_replayfid.id);
 }
 
+static void mdc_store_inode_generation_20(struct ptlrpc_request *req,
+                                          int reqoff, int repoff)
+{
+        struct mdt_rec_create *rec = lustre_msg_buf(req->rq_reqmsg, reqoff,
+                                                    sizeof(*rec));
+        struct mdt_body *body = lustre_msg_buf(req->rq_repmsg, repoff,
+                                               sizeof(*body));
+
+        LASSERT (rec != NULL);
+        LASSERT (body != NULL);
+
+        rec->cr_fid2 = body->fid1;
+        rec->cr_ioepoch = body->ioepoch;
+        rec->cr_old_handle.cookie = body->handle.cookie;
+
+        if (!fid_is_sane(&body->fid1)) {
+                DEBUG_REQ(D_ERROR, req, "saving replay request with"
+                          "insane fid");
+                LBUG();
+        }
+
+        DEBUG_REQ(D_INODE, req, "storing generation %u for ino "LPU64,
+                  rec->cr_fid1.f_oid, rec->cr_fid2.f_seq);
+}
+
+/* This should be called with both the request and the reply still packed. */
+void mdc_store_inode_generation(struct ptlrpc_request *req, int reqoff,
+                                int repoff)
+{
+        if (mdc_req_is_2_0_server(req))
+                mdc_store_inode_generation_20(req, reqoff, repoff);
+        else
+                mdc_store_inode_generation_18(req, reqoff, repoff);
+}
+
 #ifdef CONFIG_FS_POSIX_ACL
 static
 int mdc_unpack_acl(struct obd_export *exp, struct ptlrpc_request *req,
@@ -416,6 +504,7 @@ int mdc_req2lustre_md(struct ptlrpc_request *req, int offset,
                       struct lustre_md *md)
 {
         int rc = 0;
+        int iop = mdc_req_is_2_0_server(req);
         ENTRY;
 
         LASSERT(md);
@@ -459,17 +548,19 @@ int mdc_req2lustre_md(struct ptlrpc_request *req, int offset,
                 }
                 rc = 0;
 
-                offset++;
-        }
-
-        if (md->body->valid & OBD_MD_FLDIREA) {
+                if (!iop)
+                        offset++;
+        } else if (md->body->valid & OBD_MD_FLDIREA) {
                 if(!S_ISDIR(md->body->mode)) {
                         CERROR("OBD_MD_FLDIREA set, should be a directory, but "
                                "is not\n");
                         GOTO(err_out, rc = -EPROTO);
                 }
-                offset++;
+                if (!iop)
+                        offset++;
         }
+        if (iop)
+                offset++;
 
         /* for ACL, it's possible that FLACL is set but aclsize is zero.
          * only when aclsize != 0 there's an actual segment for ACL in
@@ -509,13 +600,10 @@ static void mdc_commit_open(struct ptlrpc_request *req)
         if (mod == NULL)
                 return;
 
-        if (mod->mod_close_req != NULL)
-                mod->mod_close_req->rq_cb_data = NULL;
-
         if (mod->mod_och != NULL)
                 mod->mod_och->och_mod = NULL;
 
-        OBD_FREE(mod, sizeof(*mod));
+        OBD_FREE_PTR(mod);
         req->rq_cb_data = NULL;
 }
 
@@ -538,7 +626,7 @@ static void mdc_replay_open(struct ptlrpc_request *req)
                 EXIT;
                 return;
         }
-        DEBUG_REQ(D_ERROR, req, "mdc open data found");
+        DEBUG_REQ(D_HA, req, "mdc open data found");
 
         och = mod->mod_och;
         if (och != NULL) {
@@ -547,28 +635,99 @@ static void mdc_replay_open(struct ptlrpc_request *req)
                 file_fh = &och->och_fh;
                 CDEBUG(D_RPCTRACE, "updating handle from "LPX64" to "LPX64"\n",
                        file_fh->cookie, body->handle.cookie);
-                memcpy(&old, file_fh, sizeof(old));
-                memcpy(file_fh, &body->handle, sizeof(*file_fh));
+                old = *file_fh;
+                *file_fh = body->handle;
         }
 
         close_req = mod->mod_close_req;
+
         if (close_req != NULL) {
-                struct mds_body *close_body;
                 LASSERT(lustre_msg_get_opc(close_req->rq_reqmsg) == MDS_CLOSE);
-                close_body = lustre_msg_buf(close_req->rq_reqmsg, REQ_REC_OFF,
-                                            sizeof(*close_body));
-                if (och != NULL)
-                        LASSERT(!memcmp(&old, &close_body->handle, sizeof old));
-                DEBUG_REQ(D_RPCTRACE, close_req, "updating close with new fh");
-                memcpy(&close_body->handle, &body->handle,
-                       sizeof(close_body->handle));
+                if (mdc_req_is_2_0_server(close_req)) {
+                        struct mdt_epoch *epoch = NULL;
+
+                        epoch = lustre_msg_buf(close_req->rq_reqmsg,
+                                               REQ_REC_OFF, sizeof(*epoch));
+                        LASSERT(epoch);
+                        if (och != NULL)
+                                LASSERT(!memcmp(&old, &epoch->handle,
+                                        sizeof(old)));
+                        DEBUG_REQ(D_RPCTRACE, close_req,
+                                  "updating close with new fh");
+                        epoch->handle = body->handle;
+                 } else {
+                        struct mds_body *close_body = NULL;
+
+                        close_body = lustre_msg_buf(close_req->rq_reqmsg,
+                                                    REQ_REC_OFF,
+                                                    sizeof(*close_body));
+                        if (och != NULL)
+                                LASSERT(!memcmp(&old, &close_body->handle,
+                                        sizeof(old)));
+                        DEBUG_REQ(D_RPCTRACE, close_req,
+                                  "updating close with new fh");
+                        close_body->handle = body->handle;
+                 }
         }
 
         EXIT;
 }
 
-void mdc_set_open_replay_data(struct obd_client_handle *och,
-                              struct ptlrpc_request *open_req)
+static void mdc_set_open_replay_data_20(struct obd_client_handle *och,
+                                        struct ptlrpc_request *open_req)
+{
+        struct mdc_open_data  *mod;
+        struct obd_import     *imp = open_req->rq_import;
+        struct mdt_rec_create *rec = lustre_msg_buf(open_req->rq_reqmsg,
+                                                    DLM_INTENT_REC_OFF,
+                                                    sizeof(*rec));
+        struct mdt_body       *body = lustre_msg_buf(open_req->rq_repmsg,
+                                                     DLM_REPLY_REC_OFF,
+                                                     sizeof(*body));
+
+        /* If request is not eligible for replay, just bail out */
+        if (!open_req->rq_replay)
+                return;
+
+        /* incoming message in my byte order (it's been swabbed) */
+        LASSERT(rec != NULL);
+        LASSERT(lustre_rep_swabbed(open_req, DLM_REPLY_REC_OFF));
+        /* outgoing messages always in my byte order */
+        LASSERT(body != NULL);
+
+        /* Only if the import is replayable, we set replay_open data */
+        if (och && imp->imp_replayable) {
+                OBD_ALLOC_PTR(mod);
+                if (mod == NULL) {
+                        DEBUG_REQ(D_ERROR, open_req,
+                                  "can't allocate mdc_open_data");
+                        return;
+                }
+
+                spin_lock(&open_req->rq_lock);
+                och->och_mod = mod;
+                mod->mod_och = och;
+                mod->mod_open_req = open_req;
+                open_req->rq_cb_data = mod;
+                open_req->rq_commit_cb = mdc_commit_open;
+                spin_unlock(&open_req->rq_lock);
+        }
+
+        rec->cr_fid2 = body->fid1;
+        rec->cr_ioepoch = body->ioepoch;
+        rec->cr_old_handle.cookie = body->handle.cookie;
+        open_req->rq_replay_cb = mdc_replay_open;
+        if (!fid_is_sane(&body->fid1)) {
+                DEBUG_REQ(D_ERROR, open_req, "saving replay request with "
+                          "insane fid");
+                LBUG();
+        }
+
+        DEBUG_REQ(D_RPCTRACE, open_req, "set up replay data");
+}
+
+static void mdc_set_open_replay_data_18(struct obd_client_handle *och,
+                                        struct ptlrpc_request *open_req)
 {
         struct mdc_open_data *mod;
         struct mds_rec_create *rec = lustre_msg_buf(open_req->rq_reqmsg,
@@ -615,6 +774,15 @@ void mdc_set_open_replay_data(struct obd_client_handle *och,
         DEBUG_REQ(D_RPCTRACE, open_req, "set up replay data");
 }
 
+void mdc_set_open_replay_data(struct obd_client_handle *och,
+                              struct ptlrpc_request *open_req)
+{
+        if (mdc_req_is_2_0_server(open_req))
+                mdc_set_open_replay_data_20(och, open_req);
+        else
+                mdc_set_open_replay_data_18(och, open_req);
+}
+
 void mdc_clear_open_replay_data(struct obd_client_handle *och)
 {
         struct mdc_open_data *mod = och->och_mod;
@@ -629,55 +797,34 @@ void mdc_clear_open_replay_data(struct obd_client_handle *och)
         och->och_mod = NULL;
 }
 
-static void mdc_commit_close(struct ptlrpc_request *req)
-{
-        struct mdc_open_data *mod = req->rq_cb_data;
-        struct ptlrpc_request *open_req;
-        struct obd_import *imp = req->rq_import;
-
-        DEBUG_REQ(D_RPCTRACE, req, "close req committed");
-        if (mod == NULL)
-                return;
-
-        mod->mod_close_req = NULL;
-        req->rq_cb_data = NULL;
-        req->rq_commit_cb = NULL;
-
-        open_req = mod->mod_open_req;
-        LASSERT(open_req != NULL);
-        LASSERT(open_req != LP_POISON);
-        LASSERT(open_req->rq_type != LI_POISON);
-
-        DEBUG_REQ(D_RPCTRACE, open_req, "open req balanced");
-        LASSERT(open_req->rq_transno != 0);
-        LASSERT(open_req->rq_import == imp);
-
-        /* We no longer want to preserve this for transno-unconditional
-         * replay. */
-        spin_lock(&open_req->rq_lock);
-        open_req->rq_replay = 0;
-        spin_unlock(&open_req->rq_lock);
-}
-
-int mdc_close(struct obd_export *exp, struct obdo *oa,
+int mdc_close(struct obd_export *exp, struct mdc_op_data *data, struct obdo *oa,
               struct obd_client_handle *och, struct ptlrpc_request **request)
 {
         struct obd_device *obd = class_exp2obd(exp);
-        __u32 reqsize[2] = { sizeof(struct ptlrpc_body),
-                           sizeof(struct mds_body) };
-        __u32 repsize[4] = { sizeof(struct ptlrpc_body),
-                               sizeof(struct mds_body),
-                               obd->u.cli.cl_max_mds_easize,
-                               obd->u.cli.cl_max_mds_cookiesize };
+        __u32 reqsize[4] = { sizeof(struct ptlrpc_body),
+                             sizeof(struct mdt_body) };
+        __u32 repsize[6] = { sizeof(struct ptlrpc_body),
+                             sizeof(struct mdt_body),
+                             obd->u.cli.cl_max_mds_easize,
+                             obd->u.cli.cl_max_mds_cookiesize,
+                             sizeof(struct lustre_capa),
+                             sizeof(struct lustre_capa) };
         int rc;
         struct ptlrpc_request *req;
-        struct mdc_open_data *mod;
+        int bufcount = 2;
         ENTRY;
 
+        if (mdc_exp_is_2_0_server(exp)) {
+                reqsize[1] = sizeof(struct mdt_epoch);
+                reqsize[2] = sizeof(struct mdt_rec_create);
+                reqsize[3] = 0; /* capa */
+                bufcount = 4;
+        }
         req = ptlrpc_prep_req(class_exp2cliimp(exp), LUSTRE_MDS_VERSION,
-                              MDS_CLOSE, 2, reqsize, NULL);
+                              MDS_CLOSE, bufcount, reqsize, NULL);
         if (req == NULL)
                 GOTO(out, rc = -ENOMEM);
+        req->rq_export = class_export_get(exp);
 
         /* To avoid a livelock (bug 7034), we need to send CLOSE RPCs to a
          * portal whose threads are not taking any DLM locks and are therefore
@@ -688,27 +835,32 @@ int mdc_close(struct obd_export *exp, struct obdo *oa,
         /* Ensure that this close's handle is fixed up during replay. */
         LASSERT(och != NULL);
         LASSERT(och->och_magic == OBD_CLIENT_HANDLE_MAGIC);
-        mod = och->och_mod;
-        if (likely(mod != NULL)) {
-                if (mod->mod_open_req->rq_type == LI_POISON) {
-                        CERROR("LBUG POISONED open %p!\n", mod->mod_open_req);
+        if (likely(och->och_mod != NULL)) {
+                struct ptlrpc_request *open_req = och->och_mod->mod_open_req;
+
+                if (open_req->rq_type == LI_POISON) {
+                        CERROR("LBUG POISONED open %p!\n", open_req);
                         LBUG();
                         ptlrpc_req_finished(req);
                         req = NULL;
                         GOTO(out, rc = -EIO);
                 }
-                mod->mod_close_req = req;
-                DEBUG_REQ(D_RPCTRACE, mod->mod_open_req, "matched open");
+                och->och_mod->mod_close_req = req;
+                DEBUG_REQ(D_RPCTRACE, req, "close req");
+                DEBUG_REQ(D_RPCTRACE, open_req, "clear open replay");
+
+                /* We no longer want to preserve this open for replay even
+                 * though the open was committed. b=3632, b=3633 */
+                spin_lock(&open_req->rq_lock);
+                open_req->rq_replay = 0;
+                spin_unlock(&open_req->rq_lock);
         } else {
                 CDEBUG(D_RPCTRACE, "couldn't find open req; expecting error\n");
         }
 
-        mdc_close_pack(req, REQ_REC_OFF, oa, oa->o_valid, och);
+        mdc_close_pack(req, REQ_REC_OFF, data, oa, oa->o_valid, och);
 
-        ptlrpc_req_set_repsize(req, 4, repsize);
-        req->rq_commit_cb = mdc_commit_close;
-        LASSERT(req->rq_cb_data == NULL);
-        req->rq_cb_data = mod;
+        ptlrpc_req_set_repsize(req, 6, repsize);
 
         mdc_get_rpc_lock(obd->u.cli.cl_close_lock, NULL);
         rc = ptlrpc_queue_wait(req);
@@ -726,10 +878,8 @@ int mdc_close(struct obd_export *exp, struct obdo *oa,
                                   "= %d", rc);
                         if (rc > 0)
                                 rc = -rc;
-                } else if (mod == NULL) {
-                        CERROR("Unexpected: can't find mdc_open_data, but the "
-                               "close succeeded.  Please tell <http://bugzilla.lustre.org/>.\n");
                 }
+
                 if (!lustre_swab_repbuf(req, REPLY_REC_OFF,
                                         sizeof(struct mds_body),
                                         lustre_swab_mds_body)) {
@@ -741,17 +891,19 @@ int mdc_close(struct obd_export *exp, struct obdo *oa,
         EXIT;
         *request = req;
  out:
-        if (rc != 0 && req && req->rq_commit_cb)
-                req->rq_commit_cb(req);
+        if (rc != 0 && och->och_mod)
+                 och->och_mod->mod_close_req = NULL;
 
         return rc;
 }
 
-int mdc_done_writing(struct obd_export *exp, struct obdo *obdo)
+int mdc_done_writing(struct obd_export *exp, struct mdc_op_data *data,
+                     struct obdo *obdo)
 {
         struct ptlrpc_request *req;
         struct mds_body *body;
-        __u32 size[2] = { sizeof(struct ptlrpc_body), sizeof(*body) };
+        __u32 size[2] = { sizeof(struct ptlrpc_body),
+                          sizeof(struct mdt_body) };
         int rc;
         ENTRY;
 
@@ -760,8 +912,9 @@ int mdc_done_writing(struct obd_export *exp, struct obdo *obdo)
         if (req == NULL)
                 RETURN(-ENOMEM);
 
+        req->rq_export = class_export_get(exp);
         body = lustre_msg_buf(req->rq_reqmsg, REQ_REC_OFF, sizeof(*body));
-        mdc_pack_fid(&body->fid1, obdo->o_id, 0, obdo->o_mode);
+        body->fid1 = data->fid1;
         body->size = obdo->o_size;
         body->blocks = obdo->o_blocks;
         body->flags = obdo->o_flags;
@@ -782,7 +935,8 @@ int mdc_readpage(struct obd_export *exp, struct ll_fid *fid, __u64 offset,
         struct ptlrpc_request *req = NULL;
         struct ptlrpc_bulk_desc *desc = NULL;
         struct mds_body *body;
-        __u32 size[2] = { sizeof(struct ptlrpc_body), sizeof(*body) };
+        __u32 size[2] = { sizeof(struct ptlrpc_body),
+                          sizeof(struct mdt_body) };
         int rc;
         ENTRY;
 
@@ -793,6 +947,7 @@ int mdc_readpage(struct obd_export *exp, struct ll_fid *fid, __u64 offset,
         if (req == NULL)
                 GOTO(out, rc = -ENOMEM);
 
+        req->rq_export = class_export_get(exp);
         req->rq_request_portal = MDS_READPAGE_PORTAL;
         ptlrpc_at_set_req_timeout(req);
 
@@ -874,6 +1029,9 @@ static int mdc_iocontrol(unsigned int cmd, struct obd_export *exp, int len,
                 rc = lquota_poll_check(quota_interface, exp,
                                        (struct if_quotacheck *)karg);
                 GOTO(out, rc);
+        case OBD_IOC_PING_TARGET:
+                rc = ptlrpc_obd_ping(obd);
+                GOTO(out, rc);
         default:
                 CERROR("mdc_ioctl(): unrecognised ioctl %#x\n", cmd);
                 GOTO(out, rc = -ENOTTY);
@@ -940,6 +1098,7 @@ int mdc_set_info_async(struct obd_export *exp, obd_count keylen,
                 if (req == NULL)
                         RETURN(-ENOMEM);
 
+                req->rq_export = class_export_get(exp);
                 ptlrpc_req_set_repsize(req, 1, NULL);
                 if (set) {
                         rc = 0;
@@ -995,6 +1154,7 @@ static int mdc_statfs(struct obd_device *obd, struct obd_statfs *osfs,
         if (!imp)
                 RETURN(-ENODEV);
 
+
         /* We could possibly pass max_age in the request (as an absolute
          * timestamp or a "seconds.usec ago") so the target can avoid doing
          * extra calls into the filesystem if that isn't necessary (e.g.
@@ -1035,22 +1195,27 @@ output:
         return rc;
 }
 
-static int mdc_pin(struct obd_export *exp, obd_id ino, __u32 gen, int type,
+static int mdc_pin(struct obd_export *exp, struct ll_fid *fid,
                    struct obd_client_handle *handle, int flag)
 {
         struct ptlrpc_request *req;
         struct mds_body *body;
-        __u32 size[2] = { sizeof(struct ptlrpc_body), sizeof(*body) };
+        __u32 size[3] = { sizeof(struct ptlrpc_body),
+                          sizeof(struct mdt_body), 0 };
         int rc;
+        int bufcount = 2;
         ENTRY;
 
+        if (mdc_exp_is_2_0_server(exp))
+                bufcount = 3;
         req = ptlrpc_prep_req(class_exp2cliimp(exp), LUSTRE_MDS_VERSION,
-                              MDS_PIN, 2, size, NULL);
+                              MDS_PIN, bufcount, size, NULL);
         if (req == NULL)
                 RETURN(-ENOMEM);
 
+        req->rq_export = class_export_get(exp);
         body = lustre_msg_buf(req->rq_reqmsg, REQ_REC_OFF, sizeof(*body));
-        mdc_pack_fid(&body->fid1, ino, gen, type);
+        body->fid1 = *fid;
         body->flags = flag;
 
         ptlrpc_req_set_repsize(req, 2, size);
@@ -1089,7 +1254,8 @@ static int mdc_unpin(struct obd_export *exp,
 {
         struct ptlrpc_request *req;
         struct mds_body *body;
-        __u32 size[2] = { sizeof(struct ptlrpc_body), sizeof(*body) };
+        __u32 size[2] = { sizeof(struct ptlrpc_body),
+                          sizeof(struct mdt_body) };
         int rc;
         ENTRY;
 
@@ -1101,6 +1267,7 @@ static int mdc_unpin(struct obd_export *exp,
         if (req == NULL)
                 RETURN(-ENOMEM);
 
+        req->rq_export = class_export_get(exp);
         body = lustre_msg_buf(req->rq_reqmsg, REQ_REC_OFF, sizeof(*body));
         memcpy(&body->handle, &handle->och_fh, sizeof(body->handle));
         body->flags = flag;
@@ -1123,15 +1290,21 @@ int mdc_sync(struct obd_export *exp, struct ll_fid *fid,
              struct ptlrpc_request **request)
 {
         struct ptlrpc_request *req;
-        __u32 size[2] = { sizeof(struct ptlrpc_body), sizeof(struct mds_body) };
+        __u32 size[3] = { sizeof(struct ptlrpc_body),
+                          sizeof(struct mdt_body), 0 };
+        int bufcount = 2;
         int rc;
         ENTRY;
 
+
+        if (mdc_exp_is_2_0_server(exp))
+                bufcount = 3;
         req = ptlrpc_prep_req(class_exp2cliimp(exp), LUSTRE_MDS_VERSION,
-                              MDS_SYNC, 2, size, NULL);
+                              MDS_SYNC, bufcount, size, NULL);
         if (!req)
                 RETURN(rc = -ENOMEM);
 
+        req->rq_export = class_export_get(exp);
         mdc_pack_req_body(req, REQ_REC_OFF, 0, fid, 0, 0);
 
         ptlrpc_req_set_repsize(req, 2, size);
@@ -1183,6 +1356,23 @@ static int mdc_import_event(struct obd_device *obd, struct obd_import *imp,
         RETURN(rc);
 }
 
+/* determine whether the lock can be canceled before replaying it during
+ * recovery, non zero value will be return if the lock can be canceled, 
+ * or zero returned for not */
+static int mdc_cancel_for_recovery(struct ldlm_lock *lock)
+{
+        if (lock->l_resource->lr_type != LDLM_IBITS)
+                RETURN(0);
+
+	/* FIXME: if we ever get into a situation where there are too many
+	 * opened files with open locks on a single node, then we really
+	 * should replay these open locks to reget it */
+        if (lock->l_policy_data.l_inodebits.bits & MDS_INODELOCK_OPEN)
+                RETURN(0);
+
+        RETURN(1);
+}
+
 static int mdc_setup(struct obd_device *obd, obd_count len, void *buf)
 {
         struct client_obd *cli = &obd->u.cli;
@@ -1214,7 +1404,9 @@ static int mdc_setup(struct obd_device *obd, obd_count len, void *buf)
         if (lprocfs_obd_setup(obd, lvars.obd_vars) == 0)
                 ptlrpc_lprocfs_register_obd(obd);
 
-        rc = obd_llog_init(obd, obd, 0, NULL, NULL);
+        ns_register_cancel(obd->obd_namespace, mdc_cancel_for_recovery);
+
+        rc = obd_llog_init(obd, obd, NULL);
         if (rc) {
                 mdc_cleanup(obd);
                 CERROR("failed to setup llogging subsystems\n");
@@ -1240,7 +1432,7 @@ int mdc_init_ea_size(struct obd_export *mdc_exp, struct obd_export *lov_exp)
 {
         struct obd_device *obd = mdc_exp->exp_obd;
         struct client_obd *cli = &obd->u.cli;
-        struct lov_stripe_md lsm = { .lsm_magic = LOV_MAGIC };
+        struct lov_stripe_md lsm = { .lsm_magic = LOV_MAGIC_V3 };
         struct lov_desc desc;
         __u32 valsize = sizeof(desc);
         __u32 stripes;
@@ -1285,16 +1477,17 @@ static int mdc_precleanup(struct obd_device *obd, enum obd_cleanup_stage stage)
         case OBD_CLEANUP_EXPORTS:
                 /* If we set up but never connected, the
                    client import will not have been cleaned. */
+                down_write(&obd->u.cli.cl_sem);
                 if (obd->u.cli.cl_import) {
                         struct obd_import *imp;
-                        down_write(&obd->u.cli.cl_sem);
                         imp = obd->u.cli.cl_import;
                         CERROR("client import never connected\n");
                         ptlrpc_invalidate_import(imp);
                         class_destroy_import(imp);
-                        up_write(&obd->u.cli.cl_sem);
                         obd->u.cli.cl_import = NULL;
                 }
+                up_write(&obd->u.cli.cl_sem);
+
                 rc = obd_llog_finish(obd, 0);
                 if (rc != 0)
                         CERROR("failed to cleanup llogging subsystems\n");
@@ -1323,15 +1516,14 @@ static int mdc_cleanup(struct obd_device *obd)
 }
 
 
-static int mdc_llog_init(struct obd_device *obd, struct obd_device *tgt,
-                         int count, struct llog_catid *logid,
-                         struct obd_uuid *uuid)
+static int mdc_llog_init(struct obd_device *obd, struct obd_device *disk_obd,
+                         int *index)
 {
         struct llog_ctxt *ctxt;
         int rc;
         ENTRY;
 
-        rc = llog_setup(obd, LLOG_CONFIG_REPL_CTXT, tgt, 0, NULL,
+        rc = llog_setup(obd, LLOG_CONFIG_REPL_CTXT, disk_obd, 0, NULL,
                         &llog_client_ops);
         if (rc == 0) {
                 ctxt = llog_get_context(obd, LLOG_CONFIG_REPL_CTXT);
@@ -1339,7 +1531,7 @@ static int mdc_llog_init(struct obd_device *obd, struct obd_device *tgt,
                 llog_ctxt_put(ctxt);
         }
 
-        rc = llog_setup(obd, LLOG_LOVEA_REPL_CTXT, tgt, 0, NULL,
+        rc = llog_setup(obd, LLOG_LOVEA_REPL_CTXT, disk_obd, 0, NULL,
                        &llog_client_ops);
         if (rc == 0) {
                 ctxt = llog_get_context(obd, LLOG_LOVEA_REPL_CTXT);
@@ -1392,6 +1584,57 @@ static int mdc_process_config(struct obd_device *obd, obd_count len, void *buf)
         return(rc);
 }
 
+static int mdc_fid_init(struct obd_export *exp)
+{
+        struct client_obd *cli;
+        char              *prefix;
+        int                rc;
+        ENTRY;
+
+        cli = &exp->exp_obd->u.cli;
+
+        OBD_ALLOC_PTR(cli->cl_seq);
+        if (cli->cl_seq == NULL)
+                RETURN(-ENOMEM);
+
+        OBD_ALLOC(prefix, MAX_OBD_NAME + 5);
+        if (prefix == NULL)
+                GOTO(out_free_seq, rc = -ENOMEM);
+
+        snprintf(prefix, MAX_OBD_NAME + 5, "srv-%s", exp->exp_obd->obd_name);
+
+        /* Init client side sequence-manager */
+        rc = seq_client_init(cli->cl_seq, exp,
+                             LUSTRE_SEQ_METADATA,
+                             LUSTRE_SEQ_MAX_WIDTH,
+                             prefix);
+        OBD_FREE(prefix, MAX_OBD_NAME + 5);
+        if (rc)
+                GOTO(out_free_seq, rc);
+
+        RETURN(rc);
+
+out_free_seq:
+        OBD_FREE_PTR(cli->cl_seq);
+        cli->cl_seq = NULL;
+        return rc;
+}
+
+static int mdc_fid_fini(struct obd_export *exp)
+{
+        struct client_obd *cli = &exp->exp_obd->u.cli;
+        ENTRY;
+
+        if (cli->cl_seq != NULL) {
+                LASSERT(cli->cl_seq->lcs_exp == exp);
+                seq_client_fini(cli->cl_seq);
+                OBD_FREE_PTR(cli->cl_seq);
+                cli->cl_seq = NULL;
+        }
+
+        RETURN(0);
+}
+
 struct obd_ops mdc_obd_ops = {
         .o_owner        = THIS_MODULE,
         .o_setup        = mdc_setup,
@@ -1401,6 +1644,8 @@ struct obd_ops mdc_obd_ops = {
         .o_del_conn     = client_import_del_conn,
         .o_connect      = client_connect_import,
         .o_disconnect   = client_disconnect_export,
+        .o_fid_init     = mdc_fid_init,
+        .o_fid_fini     = mdc_fid_fini,
         .o_iocontrol    = mdc_iocontrol,
         .o_set_info_async = mdc_set_info_async,
         .o_get_info     = mdc_get_info,
