@@ -64,12 +64,13 @@
 
 int mds_export_stats_init(struct obd_device *obd,
                                  struct obd_export *exp,
+                                 int reconnect,
                                  void *localdata)
 {
         lnet_nid_t *client_nid = localdata;
         int rc, num_stats, newnid = 0;
 
-        rc = lprocfs_exp_setup(exp, client_nid, &newnid);
+        rc = lprocfs_exp_setup(exp, client_nid, reconnect, &newnid);
         if (rc) {
                 /* Mask error for already created
                  * /proc entries */
@@ -85,29 +86,14 @@ int mds_export_stats_init(struct obd_device *obd,
                 num_stats = (sizeof(*obd->obd_type->typ_ops) / sizeof(void *)) +
                              LPROC_MDS_LAST - 1;
                 tmp->nid_stats = lprocfs_alloc_stats(num_stats,
-                                                     LPROCFS_STATS_FLAG_NOPERCPU);
+                                            LPROCFS_STATS_FLAG_NOPERCPU);
                 if (tmp->nid_stats == NULL)
                         return -ENOMEM;
 
                 lprocfs_init_ops_stats(LPROC_MDS_LAST, tmp->nid_stats);
-                rc = lprocfs_register_stats(tmp->nid_proc, "stats",
-                                            tmp->nid_stats);
-                if (rc)
-                        return rc;
-
                 mds_stats_counter_init(tmp->nid_stats);
 
-                /* Always add in ldlm_stats */
-                tmp->nid_ldlm_stats = lprocfs_alloc_stats(LDLM_LAST_OPC -
-                                                          LDLM_FIRST_OPC,
-                                                          0);
-                if (tmp->nid_ldlm_stats == NULL)
-                        return -ENOMEM;
-
-                lprocfs_init_ldlm_stats(tmp->nid_ldlm_stats);
-
-                rc = lprocfs_register_stats(tmp->nid_proc, "ldlm_stats",
-                                            tmp->nid_ldlm_stats);
+                rc = lprocfs_nid_ldlm_stats_init(tmp);
                 if (rc)
                         return rc;
         }
@@ -115,6 +101,58 @@ int mds_export_stats_init(struct obd_device *obd,
         return 0;
 }
 
+/* VBR: to determine the delayed client the lcd should be updated for each new
+ * epoch */
+int mds_update_client_epoch(struct obd_export *exp)
+{
+        struct mds_export_data *med = &exp->exp_mds_data;
+        struct mds_obd *mds = &exp->exp_obd->u.mds;
+        struct lvfs_run_ctxt saved;
+        loff_t off = med->med_lr_off;
+        int rc = 0;
+
+        /* VBR: set client last_epoch to current epoch */
+        if (le32_to_cpu(med->med_lcd->lcd_last_epoch) >=
+                        le32_to_cpu(mds->mds_server_data->lsd_start_epoch))
+                return rc;
+
+        med->med_lcd->lcd_last_epoch = mds->mds_server_data->lsd_start_epoch;
+        push_ctxt(&saved, &exp->exp_obd->obd_lvfs_ctxt, NULL);
+        rc = fsfilt_write_record(exp->exp_obd, mds->mds_rcvd_filp,
+                                 med->med_lcd, sizeof(*med->med_lcd), &off,
+                                 exp->exp_delayed);
+        pop_ctxt(&saved, &exp->exp_obd->obd_lvfs_ctxt, NULL);
+
+        CDEBUG(D_INFO, "update client idx %u last_epoch %#x (%#x)\n",
+               med->med_lr_idx, le32_to_cpu(med->med_lcd->lcd_last_epoch),
+               le32_to_cpu(mds->mds_server_data->lsd_start_epoch));
+
+        return rc;
+}
+
+/* Called after recovery is done on server */
+void mds_update_last_epoch(struct obd_device *obd)
+{
+        struct ptlrpc_request *req;
+        struct mds_obd *mds = &obd->u.mds;
+        __u32 start_epoch;
+
+        /* Increase server epoch after recovery */
+        spin_lock(&mds->mds_transno_lock);
+        start_epoch = lr_epoch(mds->mds_last_transno) + 1;
+        mds->mds_last_transno = (__u64)start_epoch << LR_EPOCH_BITS;
+        mds->mds_server_data->lsd_start_epoch = cpu_to_le32(start_epoch);
+        spin_unlock(&mds->mds_transno_lock);
+
+        /* go through delayed reply queue to find all exports participate in
+         * recovery and set new epoch for them */
+        list_for_each_entry(req, &obd->obd_delayed_reply_queue, rq_list) {
+                LASSERT(!req->rq_export->exp_delayed);
+                mds_update_client_epoch(req->rq_export);
+        }
+        mds_update_server_data(obd, 1);
+}
+
 /* Add client data to the MDS.  We use a bitmap to locate a free space
  * in the last_rcvd file if cl_off is -1 (i.e. a new client).
  * Otherwise, we have just read the data from the last_rcvd file and
@@ -130,7 +168,7 @@ int mds_client_add(struct obd_device *obd, struct obd_export *exp,
         struct mds_export_data *med = &exp->exp_mds_data;
         unsigned long *bitmap = mds->mds_client_bitmap;
         int new_client = (cl_idx == -1);
-        int rc;
+        int rc = 0;
         ENTRY;
 
         LASSERT(bitmap != NULL);
@@ -140,6 +178,10 @@ int mds_client_add(struct obd_device *obd, struct obd_export *exp,
         if (!strcmp(med->med_lcd->lcd_uuid, obd->obd_uuid.uuid))
                 RETURN(0);
 
+        /* VBR: remove expired exports before searching for free slot */
+        if (new_client)
+                class_disconnect_expired_exports(obd);
+
         /* the bitmap operations can handle cl_idx > sizeof(long) * 8, so
          * there's no need for extra complication here
          */
@@ -172,7 +214,7 @@ int mds_client_add(struct obd_device *obd, struct obd_export *exp,
         med->med_lr_off = le32_to_cpu(mds->mds_server_data->lsd_client_start) +
                 (cl_idx * le16_to_cpu(mds->mds_server_data->lsd_client_size));
         LASSERTF(med->med_lr_off > 0, "med_lr_off = %llu\n", med->med_lr_off);
-        mds_export_stats_init(obd, exp, localdata);
+        mds_export_stats_init(obd, exp, 0, localdata);
 
         if (new_client) {
                 struct lvfs_run_ctxt *saved = NULL;
@@ -193,6 +235,14 @@ int mds_client_add(struct obd_device *obd, struct obd_export *exp,
                         rc = PTR_ERR(handle);
                         CERROR("unable to start transaction: rc %d\n", rc);
                 } else {
+                        /* VBR: set client last_transno as mds_last_transno to
+                         * remember last epoch for this client */
+                        med->med_lcd->lcd_last_epoch =
+                                        mds->mds_server_data->lsd_start_epoch;
+                        exp->exp_last_request_time = cfs_time_current_sec();
+                        /* remember first epoch of client for orphan handling */
+                        med->med_lcd->lcd_first_epoch =
+                                  cpu_to_le32(lr_epoch(mds->mds_last_transno));
                         rc = fsfilt_add_journal_cb(obd, 0, handle,
                                                    target_client_add_cb, exp);
                         if (rc == 0) {
@@ -215,11 +265,11 @@ int mds_client_add(struct obd_device *obd, struct obd_export *exp,
                        med->med_lr_idx, med->med_lr_off,
                        (unsigned int)sizeof(*med->med_lcd));
         }
-        return 0;
+        return rc;
 }
 
 struct lsd_client_data zero_lcd; /* globals are implicitly zeroed */
-
+ 
 int mds_client_free(struct obd_export *exp)
 {
         struct mds_export_data *med = &exp->exp_mds_data;
@@ -266,7 +316,6 @@ int mds_client_free(struct obd_export *exp)
                  * or it does num_clients * num_osts syncs.  b=17194 */
                 int need_sync = (!exp->exp_libclient || exp->exp_need_sync) &&
                                  !(exp->exp_flags & OBD_OPT_ABORT_RECOV);
-
                 OBD_SLAB_ALLOC_PTR(saved, obd_lvfs_ctxt_cache);
                 if (saved == NULL) {
                         CERROR("cannot allocate memory for run ctxt\n");
@@ -316,14 +365,95 @@ static int mds_server_free_data(struct mds_obd *mds)
         return 0;
 }
 
+static void mds_add_fake_export(struct obd_device *obd, int num,
+                                struct file *file)
+{
+        struct obd_export *exp;
+        struct lvfs_run_ctxt saved;
+        struct obd_device_target *obt = &obd->u.obt;
+        struct lu_export_data *led;
+        unsigned long *bitmap = obt->obt_client_bitmap;
+        struct lsd_client_data *lcd = NULL;
+        unsigned int idx = 0;
+        loff_t off = 0;
+        int rc = 0;
+
+        while (num > 0) {
+                num--;
+                if (!lcd) {
+                        OBD_ALLOC_PTR(lcd);
+                        if (!lcd)
+                                return;
+                }
+                idx = find_next_zero_bit(bitmap, LR_MAX_CLIENTS, idx);
+                if (idx >= LR_MAX_CLIENTS) {
+                        CERROR("no room for %u clients - fix LR_MAX_CLIENTS\n", idx);
+                        OBD_FREE_PTR(lcd);
+                        break;
+                }
+                if (test_and_set_bit(idx, bitmap)) {
+                        CERROR("Bit %u is set already\n", idx);
+                        continue;
+                }
+                off = le32_to_cpu(obt->obt_lsd->lsd_client_start) +
+                      idx * le16_to_cpu(obt->obt_lsd->lsd_client_size);
+
+                sprintf(lcd->lcd_uuid, "dead-%.16u", idx);
+                CDEBUG(D_INFO, "Create fake export %s, index %u, offset %lu\n",
+                       lcd->lcd_uuid, idx, (unsigned long)off);
+
+                exp = class_new_export(obd, (struct obd_uuid *)lcd->lcd_uuid);
+                if (IS_ERR(exp)) {
+                        if (PTR_ERR(exp) == -EALREADY) {
+                                CERROR("Export %s already exists\n",
+                                       lcd->lcd_uuid);
+                        }
+                        CERROR("Failed to create export %lu\n", PTR_ERR(exp));
+                        OBD_FREE_PTR(lcd);
+                        break;
+                }
+                LASSERT(exp);
+                led = &exp->exp_target_data;
+                led->led_lr_idx = idx;
+                led->led_lr_off = off;
+                led->led_lcd = lcd;
+
+                exp->exp_last_request_time = cfs_time_current_sec();
+                exp->exp_replay_needed = 1;
+                exp->exp_connecting = 0;
+                exp->exp_in_recovery = 0;
+
+                spin_lock_bh(&obd->obd_processing_task_lock);
+                obd->obd_recoverable_clients++;
+                obd->obd_max_recoverable_clients++;
+                spin_unlock_bh(&obd->obd_processing_task_lock);
+
+                class_set_export_delayed(exp);
+                class_export_put(exp);
+
+                lcd->lcd_last_epoch = cpu_to_le32(1);
+                push_ctxt(&saved, &obd->obd_lvfs_ctxt, NULL);
+                rc = fsfilt_write_record(obd, file, lcd, sizeof(*lcd), &off, 0);
+                pop_ctxt(&saved, &obd->obd_lvfs_ctxt, NULL);
+                if (rc) {
+                        CERROR("Failed to create fake client record\n");
+                        OBD_FREE_PTR(lcd);
+                        break;
+                }
+                lcd = NULL;
+        }
+}
+
 static int mds_init_server_data(struct obd_device *obd, struct file *file)
 {
         struct mds_obd *mds = &obd->u.mds;
         struct lr_server_data *lsd;
         struct lsd_client_data *lcd = NULL;
+        struct lustre_mount_info *lmi;
         loff_t off = 0;
         unsigned long last_rcvd_size = i_size_read(file->f_dentry->d_inode);
         __u64 mount_count;
+        __u32 start_epoch;
         int cl_idx, rc = 0;
         ENTRY;
 
@@ -354,6 +484,7 @@ static int mds_init_server_data(struct obd_device *obd, struct file *file)
                 lsd->lsd_server_size = cpu_to_le32(LR_SERVER_SIZE);
                 lsd->lsd_client_start = cpu_to_le32(LR_CLIENT_START);
                 lsd->lsd_client_size = cpu_to_le16(LR_CLIENT_SIZE);
+                lsd->lsd_feature_compat = cpu_to_le32(OBD_COMPAT_MDT);
                 lsd->lsd_feature_rocompat = cpu_to_le32(OBD_ROCOMPAT_LOVOBJID);
                 lsd->lsd_feature_incompat = cpu_to_le32(OBD_INCOMPAT_MDT);
         } else {
@@ -369,6 +500,7 @@ static int mds_init_server_data(struct obd_device *obd, struct file *file)
                                            obd->obd_uuid.uuid, lsd->lsd_uuid);
                         GOTO(err_msd, rc = -EINVAL);
                 }
+                lsd->lsd_feature_compat |= cpu_to_le32(OBD_COMPAT_MDT);
                 /* COMPAT_146 */
                 /* Assume old last_rcvd format unless I_C_LR is set */
                 if (!(lsd->lsd_feature_incompat &
@@ -391,12 +523,24 @@ static int mds_init_server_data(struct obd_device *obd, struct file *file)
                 /* Do something like remount filesystem read-only */
                 GOTO(err_msd, rc = -EINVAL);
         }
+        /* evict all clients as it is first boot with 2.0 last_rcvd */
+        if (lsd->lsd_feature_compat & cpu_to_le32(OBD_COMPAT_20)) {
+                LCONSOLE_WARN("Mounting %s at first time on 2.0 FS, remove all"
+                              " clients for interop needs\n", obd->obd_name);
+                simple_truncate(mds->mds_vfsmnt->mnt_sb->s_root,
+                                mds->mds_vfsmnt, LAST_RCVD,
+                                lsd->lsd_client_start);
+                last_rcvd_size = lsd->lsd_client_start;
+                lsd->lsd_feature_compat &= ~cpu_to_le32(OBD_COMPAT_20);
+        }
 
-        lsd->lsd_feature_compat = cpu_to_le32(OBD_COMPAT_MDT);
-
+        target_trans_table_init(obd);
         mds->mds_last_transno = le64_to_cpu(lsd->lsd_last_transno);
+        start_epoch = le32_to_cpu(lsd->lsd_start_epoch);
 
-        CDEBUG(D_INODE, "%s: server last_transno: "LPU64"\n",
+        CDEBUG(D_INODE, "%s: server start_epoch: %#x\n",
+               obd->obd_name, start_epoch);
+        CDEBUG(D_INODE, "%s: server last_transno: "LPX64"\n",
                obd->obd_name, mds->mds_last_transno);
         CDEBUG(D_INODE, "%s: server mount_count: "LPU64"\n",
                obd->obd_name, mount_count + 1);
@@ -425,6 +569,7 @@ static int mds_init_server_data(struct obd_device *obd, struct file *file)
         for (cl_idx = 0, off = le32_to_cpu(lsd->lsd_client_start);
              off < last_rcvd_size; cl_idx++) {
                 __u64 last_transno;
+                __u32 last_epoch;
                 struct obd_export *exp;
                 struct mds_export_data *med;
 
@@ -452,10 +597,10 @@ static int mds_init_server_data(struct obd_device *obd, struct file *file)
                         continue;
                 }
 
-                last_transno = le64_to_cpu(lcd->lcd_last_transno) >
-                               le64_to_cpu(lcd->lcd_last_close_transno) ?
-                               le64_to_cpu(lcd->lcd_last_transno) :
-                               le64_to_cpu(lcd->lcd_last_close_transno);
+                check_lcd(obd->obd_name, cl_idx, lcd);
+
+                last_transno = lsd_last_transno(lcd);
+                last_epoch = le32_to_cpu(lcd->lcd_last_epoch);
 
                 /* These exports are cleaned up by mds_disconnect(), so they
                  * need to be set up like real exports as mds_connect() does.
@@ -480,26 +625,42 @@ static int mds_init_server_data(struct obd_device *obd, struct file *file)
                         /* can't fail for existing client */
                         LASSERTF(rc == 0, "rc = %d\n", rc);
 
+                        /* VBR: set export last committed version */
+                        exp->exp_last_committed = last_transno;
+                        /* read last time from disk */
+                        exp->exp_last_request_time = target_trans_table_last_time(exp);
                         lcd = NULL;
 
                         spin_lock(&exp->exp_lock);
                         exp->exp_replay_needed = 1;
                         exp->exp_connecting = 0;
+                        exp->exp_in_recovery = 0;
                         spin_unlock(&exp->exp_lock);
 
+                        spin_lock_bh(&obd->obd_processing_task_lock);
                         obd->obd_recoverable_clients++;
                         obd->obd_max_recoverable_clients++;
+                        spin_unlock_bh(&obd->obd_processing_task_lock);
+
+                        /* VBR: if epoch too old mark export as delayed,
+                         * if epoch is zero then client is pre-vbr one */
+                        if (start_epoch > last_epoch && last_epoch != 0)
+                                class_set_export_delayed(exp);
                         class_export_put(exp);
                 }
 
                 /* Need to check last_rcvd even for duplicated exports. */
-                CDEBUG(D_OTHER, "client at idx %d has last_transno = "LPU64"\n",
-                       cl_idx, last_transno);
+                CDEBUG(D_OTHER, "client at idx %d has last_transno = "LPX64","
+                       "last_epoch %#x\n", cl_idx, last_transno, last_epoch);
 
                 if (last_transno > mds->mds_last_transno)
                         mds->mds_last_transno = last_transno;
         }
 
+        if (unlikely(OBD_FAIL_CHECK(OBD_FAIL_TGT_FAKE_EXP))) {
+                mds_add_fake_export(obd, obd_fail_val, file);
+        }
+
         if (lcd)
                 OBD_FREE_PTR(lcd);
 
@@ -507,17 +668,33 @@ static int mds_init_server_data(struct obd_device *obd, struct file *file)
 
         if (obd->obd_recoverable_clients) {
                 CWARN("RECOVERY: service %s, %d recoverable clients, "
-                      "last_transno "LPU64"\n", obd->obd_name,
-                      obd->obd_recoverable_clients, mds->mds_last_transno);
+                      "%d delayed clients, last_transno "LPU64"\n",
+                      obd->obd_name, obd->obd_recoverable_clients,
+                      obd->obd_delayed_clients, mds->mds_last_transno);
                 obd->obd_next_recovery_transno = obd->obd_last_committed + 1;
                 obd->obd_recovering = 1;
                 obd->obd_recovery_start = 0;
                 obd->obd_recovery_end = 0;
-                obd->obd_recovery_timeout = OBD_RECOVERY_FACTOR * obd_timeout;
-#ifdef CRAY_XT3
-                /* bz13079: this won't be changed for mds */
-                obd->obd_recovery_max_time = OBD_RECOVERY_MAX_TIME;
-#endif
+        } else {
+                LASSERT(!obd->obd_recovering);
+                /* VBR: update boot epoch after recovery */
+                mds_update_last_epoch(obd);
+        }
+
+        obd->obd_recovery_timeout = OBD_RECOVERY_TIME_SOFT;
+        obd->obd_recovery_time_hard = OBD_RECOVERY_TIME_HARD;
+
+        lmi = server_find_mount_locked(obd->obd_name);
+        if (lmi) {
+                struct lustre_sb_info *lsi = s2lsi(lmi->lmi_sb);
+
+                if (lsi->lsi_lmd && lsi->lsi_lmd->lmd_recovery_time_soft)
+                        obd->obd_recovery_timeout =
+                                lsi->lsi_lmd->lmd_recovery_time_soft;
+
+                if (lsi->lsi_lmd && lsi->lsi_lmd->lmd_recovery_time_hard)
+                        obd->obd_recovery_time_hard =
+                                lsi->lsi_lmd->lmd_recovery_time_hard;
         }
 
         mds->mds_mount_count = mount_count + 1;
@@ -562,6 +739,8 @@ int mds_fs_setup(struct obd_device *obd, struct vfsmount *mnt)
         mds->mds_vfsmnt = mnt;
         /* why not mnt->mnt_sb instead of mnt->mnt_root->d_inode->i_sb? */
         obd->u.obt.obt_sb = mnt->mnt_root->d_inode->i_sb;
+        obd->u.obt.obt_stale_export_age = STALE_EXPORT_MAXTIME_DEFAULT;
+        spin_lock_init(&obd->u.obt.obt_trans_table_lock);
 
         rc = fsfilt_setup(obd, obd->u.obt.obt_sb);
         if (rc)
@@ -575,7 +754,7 @@ int mds_fs_setup(struct obd_device *obd, struct vfsmount *mnt)
 
         /* setup the directory tree */
         push_ctxt(saved, &obd->obd_lvfs_ctxt, NULL);
-        dentry = simple_mkdir(current->fs->pwd, mnt, "ROOT", 0755, 0);
+        dentry = simple_mkdir(cfs_fs_pwd(current->fs), mnt, "ROOT", 0755, 0);
         if (IS_ERR(dentry)) {
                 rc = PTR_ERR(dentry);
                 CERROR("cannot create ROOT directory: rc = %d\n", rc);
@@ -588,7 +767,7 @@ int mds_fs_setup(struct obd_device *obd, struct vfsmount *mnt)
 
         dput(dentry);
 
-        dentry = lookup_one_len("__iopen__", current->fs->pwd,
+        dentry = lookup_one_len("__iopen__", cfs_fs_pwd(current->fs),
                                 strlen("__iopen__"));
         if (IS_ERR(dentry)) {
                 rc = PTR_ERR(dentry);
@@ -603,7 +782,7 @@ int mds_fs_setup(struct obd_device *obd, struct vfsmount *mnt)
                 GOTO(err_fid, rc);
         }
 
-        dentry = simple_mkdir(current->fs->pwd, mnt, "PENDING", 0777, 1);
+        dentry = simple_mkdir(cfs_fs_pwd(current->fs), mnt, "PENDING", 0777, 1);
         if (IS_ERR(dentry)) {
                 rc = PTR_ERR(dentry);
                 CERROR("cannot create PENDING directory: rc = %d\n", rc);
@@ -612,7 +791,7 @@ int mds_fs_setup(struct obd_device *obd, struct vfsmount *mnt)
         mds->mds_pending_dir = dentry;
 
         /* COMPAT_146 */
-        dentry = simple_mkdir(current->fs->pwd, mnt, MDT_LOGS_DIR, 0777, 1);
+        dentry = simple_mkdir(cfs_fs_pwd(current->fs), mnt, MDT_LOGS_DIR, 0777, 1);
         if (IS_ERR(dentry)) {
                 rc = PTR_ERR(dentry);
                 CERROR("cannot create %s directory: rc = %d\n",
@@ -622,7 +801,7 @@ int mds_fs_setup(struct obd_device *obd, struct vfsmount *mnt)
         mds->mds_logs_dir = dentry;
         /* end COMPAT_146 */
 
-        dentry = simple_mkdir(current->fs->pwd, mnt, "OBJECTS", 0777, 1);
+        dentry = simple_mkdir(cfs_fs_pwd(current->fs), mnt, "OBJECTS", 0777, 1);
         if (IS_ERR(dentry)) {
                 rc = PTR_ERR(dentry);
                 CERROR("cannot create OBJECTS directory: rc = %d\n", rc);
@@ -663,7 +842,7 @@ int mds_fs_setup(struct obd_device *obd, struct vfsmount *mnt)
                 CERROR("cannot open/create %s file: rc = %d\n",HEALTH_CHECK,rc);
                 GOTO(err_lov_objid, rc = PTR_ERR(file));
         }
-        mds->mds_health_check_filp = file;
+        mds->mds_obt.obt_health_check_filp = file;
         if (!S_ISREG(file->f_dentry->d_inode->i_mode)) {
                 CERROR("%s is not a regular file!: mode = %o\n", HEALTH_CHECK,
                        file->f_dentry->d_inode->i_mode);
@@ -678,8 +857,8 @@ err_pop:
         return rc;
 
 err_health_check:
-        if (mds->mds_health_check_filp &&
-            filp_close(mds->mds_health_check_filp, 0))
+        if (mds->mds_obt.obt_health_check_filp &&
+            filp_close(mds->mds_obt.obt_health_check_filp, 0))
                 CERROR("can't close %s after error\n", HEALTH_CHECK);
 err_lov_objid:
          mds_lov_destroy_objids(obd);
@@ -716,6 +895,12 @@ int mds_fs_cleanup(struct obd_device *obd)
                               "will be preserved.\n", obd->obd_name);
 
         class_disconnect_exports(obd); /* cleans up client info too */
+
+       /* some exports may still be in the zombie queue, so we make sure that
+        * all the exports have been processed, otherwise the last_rcvd slot
+        * may not be updated on time */
+        obd_zombie_barrier();
+
         mds_server_free_data(mds);
 
         push_ctxt(saved, &obd->obd_lvfs_ctxt, NULL);
@@ -728,9 +913,9 @@ int mds_fs_cleanup(struct obd_device *obd)
 
         mds_lov_destroy_objids(obd);
 
-        if (mds->mds_health_check_filp) {
-                rc = filp_close(mds->mds_health_check_filp, 0);
-                mds->mds_health_check_filp = NULL;
+        if (mds->mds_obt.obt_health_check_filp) {
+                rc = filp_close(mds->mds_obt.obt_health_check_filp, 0);
+                mds->mds_obt.obt_health_check_filp = NULL;
                 if (rc)
                         CERROR("%s file won't close, rc=%d\n", HEALTH_CHECK,rc);
         }
@@ -751,9 +936,9 @@ int mds_fs_cleanup(struct obd_device *obd)
 
         pop_ctxt(saved, &obd->obd_lvfs_ctxt, NULL);
         OBD_SLAB_FREE_PTR(saved, obd_lvfs_ctxt_cache);
-        shrink_dcache_parent(mds->mds_fid_de);
         dput(mds->mds_fid_de);
-        LL_DQUOT_OFF(obd->u.obt.obt_sb);
+        LL_DQUOT_OFF(obd->u.obt.obt_sb, 0);
+        shrink_dcache_sb(mds->mds_obt.obt_sb);
 
         return rc;
 }
@@ -767,8 +952,8 @@ int mds_obd_create(struct obd_export *exp, struct obdo *oa,
         struct mds_obd *mds = &exp->exp_obd->u.mds;
         struct inode *parent_inode = mds->mds_objects_dir->d_inode;
         unsigned int tmpname = ll_rand();
-        struct file *filp;
-        struct dentry *new_child;
+        struct dentry *dchild, *new_child;
+        struct lvfs_dentry_params dp = LVFS_DENTRY_PARAMS_INIT;
         struct lvfs_run_ctxt *saved = NULL;
         char fidname[LL_FID_NAMELEN];
         void *handle;
@@ -787,31 +972,32 @@ int mds_obd_create(struct obd_export *exp, struct obdo *oa,
 
         push_ctxt(saved, &exp->exp_obd->obd_lvfs_ctxt, &ucred);
 
-        sprintf(fidname, "OBJECTS/%u.%u", tmpname, current->pid);
-        filp = filp_open(fidname, O_CREAT | O_EXCL, 0666);
-        if (IS_ERR(filp)) {
-                rc = PTR_ERR(filp);
-                if (rc == -EEXIST) {
-                        CERROR("impossible object name collision %u\n",
-                               tmpname);
-                        LBUG();
-                }
-                CERROR("error creating tmp object %u: rc %d\n", tmpname, rc);
-                GOTO(out_pop, rc);
+        sprintf(fidname, "%u.%u", tmpname, current->pid);
+        dchild = lookup_one_len(fidname, mds->mds_objects_dir, strlen(fidname));
+        if (IS_ERR(dchild)) {
+                CERROR("getting neg dentry for obj: %u\n", tmpname);
+                GOTO(out_pop, rc = PTR_ERR(dchild));
+        }
+        if (dchild->d_inode != NULL) {
+                CERROR("impossible non-negative obj dentry: %u\n", tmpname);
+                LBUG();
         }
 
-        LASSERT(mds->mds_objects_dir == filp->f_dentry->d_parent);
-
-        oa->o_id = filp->f_dentry->d_inode->i_ino;
-        oa->o_generation = filp->f_dentry->d_inode->i_generation;
-        namelen = ll_fid2str(fidname, oa->o_id, oa->o_generation);
+        dchild->d_fsdata = (void *)&dp;
+        dp.ldp_ptr   = (void *)DP_LASTGROUP_REVERSE;
 
         LOCK_INODE_MUTEX(parent_inode);
+        rc = ll_vfs_create(parent_inode, dchild, S_IFREG | 0666, NULL);
+
+        oa->o_id = dchild->d_inode->i_ino;
+        oa->o_generation = dchild->d_inode->i_generation;
+        namelen = ll_fid2str(fidname, oa->o_id, oa->o_generation);
+
         new_child = lookup_one_len(fidname, mds->mds_objects_dir, namelen);
 
         if (IS_ERR(new_child)) {
                 CERROR("getting neg dentry for obj rename: %d\n", rc);
-                GOTO(out_close, rc = PTR_ERR(new_child));
+                GOTO(out_dput, rc = PTR_ERR(new_child));
         }
         if (new_child->d_inode != NULL) {
                 CERROR("impossible non-negative obj dentry " LPU64":%u!\n",
@@ -822,13 +1008,10 @@ int mds_obd_create(struct obd_export *exp, struct obdo *oa,
         handle = fsfilt_start(exp->exp_obd, mds->mds_objects_dir->d_inode,
                               FSFILT_OP_RENAME, NULL);
         if (IS_ERR(handle))
-                GOTO(out_dput, rc = PTR_ERR(handle));
+                GOTO(out_dput2, rc = PTR_ERR(handle));
 
-        lock_kernel();
-        rc = ll_vfs_rename(mds->mds_objects_dir->d_inode, filp->f_dentry,
-                           filp->f_vfsmnt, mds->mds_objects_dir->d_inode,
-                           new_child, filp->f_vfsmnt);
-        unlock_kernel();
+        rc = ll_vfs_rename(parent_inode, dchild, mds->mds_vfsmnt,
+                           parent_inode, new_child, mds->mds_vfsmnt);
         if (rc)
                 CERROR("error renaming new object "LPU64":%u: rc %d\n",
                        oa->o_id, oa->o_generation, rc);
@@ -839,16 +1022,11 @@ int mds_obd_create(struct obd_export *exp, struct obdo *oa,
                 oa->o_valid |= OBD_MD_FLID | OBD_MD_FLGENER;
         else if (!rc)
                 rc = err;
-out_dput:
+out_dput2:
         dput(new_child);
-out_close:
+out_dput:
+        dput(dchild);
         UNLOCK_INODE_MUTEX(parent_inode);
-        err = filp_close(filp, 0);
-        if (err) {
-                CERROR("closing tmpfile %u: rc %d\n", tmpname, rc);
-                if (!rc)
-                        rc = err;
-        }
 out_pop:
         pop_ctxt(saved, &exp->exp_obd->obd_lvfs_ctxt, &ucred);
         OBD_SLAB_FREE_PTR(saved, obd_lvfs_ctxt_cache);
