@@ -41,9 +41,10 @@
 #include <lustre_mds.h>
 
 #define MDT_ROCOMPAT_SUPP       (OBD_ROCOMPAT_LOVOBJID)
-#define MDT_INCOMPAT_SUPP       (OBD_INCOMPAT_MDT | OBD_INCOMPAT_COMMON_LR)
+#define MDT_INCOMPAT_SUPP       (OBD_INCOMPAT_MDT | OBD_INCOMPAT_COMMON_LR | \
+                                 OBD_INCOMPAT_FID)
 
-#define MDS_SERVICE_WATCHDOG_FACTOR 2000
+#define MDS_SERVICE_WATCHDOG_FACTOR 2
 
 #define MAX_ATIME_DIFF 60
 
@@ -51,7 +52,9 @@ struct mds_filter_data {
         __u64 io_epoch;
 };
 
-#define MDS_FILTERDATA(inode) ((struct mds_filter_data *)(inode)->i_filterdata)
+#define MDS_FILTERDATA(inode) \
+                ((struct mds_filter_data *)INODE_PRIVATE_DATA(inode))
+
 
 static inline struct mds_obd *mds_req2mds(struct ptlrpc_request *req)
 {
@@ -68,7 +71,6 @@ static inline void mds_export_evict(struct obd_export *exp)
 /* Open counts for files.  No longer atomic, must hold inode->i_sem */
 # define mds_inode_oatomic(inode)    ((inode)->i_cindex)
 
-#ifdef HAVE_I_ALLOC_SEM
 #define MDS_UP_READ_ORPHAN_SEM(i)          UP_READ_I_ALLOC_SEM(i)
 #define MDS_DOWN_READ_ORPHAN_SEM(i)        DOWN_READ_I_ALLOC_SEM(i)
 #define LASSERT_MDS_ORPHAN_READ_LOCKED(i)  LASSERT_I_ALLOC_SEM_READ_LOCKED(i)
@@ -77,16 +79,6 @@ static inline void mds_export_evict(struct obd_export *exp)
 #define MDS_DOWN_WRITE_ORPHAN_SEM(i)       DOWN_WRITE_I_ALLOC_SEM(i)
 #define LASSERT_MDS_ORPHAN_WRITE_LOCKED(i) LASSERT_I_ALLOC_SEM_WRITE_LOCKED(i)
 #define MDS_PACK_MD_LOCK 1
-#else
-#define MDS_UP_READ_ORPHAN_SEM(i)          do { up(&(i)->i_sem); } while (0)
-#define MDS_DOWN_READ_ORPHAN_SEM(i)        do { down(&(i)->i_sem); } while (0)
-#define LASSERT_MDS_ORPHAN_READ_LOCKED(i)  LASSERT(down_trylock(&(i)->i_sem)!=0)
-
-#define MDS_UP_WRITE_ORPHAN_SEM(i)         do { up(&(i)->i_sem); } while (0)
-#define MDS_DOWN_WRITE_ORPHAN_SEM(i)       do { down(&(i)->i_sem); } while (0)
-#define LASSERT_MDS_ORPHAN_WRITE_LOCKED(i) LASSERT(down_trylock(&(i)->i_sem)!=0)
-#define MDS_PACK_MD_LOCK 0
-#endif
 
 static inline int mds_orphan_open_count(struct inode *inode)
 {
@@ -94,6 +86,13 @@ static inline int mds_orphan_open_count(struct inode *inode)
         return mds_inode_oatomic(inode);
 }
 
+static inline int mds_orphan_needed(struct obd_device *obd,
+                                    struct inode * inode)
+{
+        return (obd->obd_recovering ||
+                mds_orphan_open_count(inode) > 0);
+}
+
 static inline int mds_orphan_open_inc(struct inode *inode)
 {
         LASSERT_MDS_ORPHAN_WRITE_LOCKED(inode);
@@ -150,9 +149,8 @@ int enqueue_ordered_locks(struct obd_device *obd, struct ldlm_res_id *p1_res_id,
                           struct lustre_handle *p2_lockh, int p2_lock_mode,
                           ldlm_policy_data_t *p2_policy);
 void mds_commit_cb(struct obd_device *, __u64 last_rcvd, void *data, int error);
-int mds_finish_transno(struct mds_obd *mds, struct inode *inode, void *handle,
-                       struct ptlrpc_request *req, int rc, __u32 op_data,
-                       int force_sync);
+int mds_finish_transno(struct mds_obd *, struct inode **, void *,
+                       struct ptlrpc_request *, int, __u32, int force_sync);
 void mds_reconstruct_generic(struct ptlrpc_request *req);
 void mds_req_from_lcd(struct ptlrpc_request *req, struct lsd_client_data *cd);
 int mds_get_parent_child_locked(struct obd_device *obd, struct mds_obd *mds,
@@ -184,9 +182,17 @@ int mds_get_parents_children_locked(struct obd_device *obd,
                                     struct lustre_handle *dlm_handles,
                                     int child_mode);
 
-void mds_shrink_reply(struct obd_device *obd, struct ptlrpc_request *req,
-                      struct mds_body *body, int md_off);
+struct dentry *mds_lookup(struct obd_device *obd,
+                          const char *fid_name,
+                          struct dentry *dparent,
+                          int fid_namelen);
+
+void mds_shrink_body_reply(struct ptlrpc_request *req, int req_mdoff,
+                           int reply_mdoff);
+void mds_shrink_intent_reply(struct ptlrpc_request *req,
+                             int opc, int reply_mdoff);
 int mds_get_cookie_size(struct obd_device *obd, struct lov_mds_md *lmm);
+int mds_version_get_check(struct ptlrpc_request *, struct inode *, int);
 /* mds/mds_lib.c */
 int mds_update_unpack(struct ptlrpc_request *, int offset,
                       struct mds_update_record *);
@@ -202,16 +208,18 @@ int mds_osc_destroy_orphan(struct obd_device *obd, umode_t mode,
                            struct lov_mds_md *lmm, int lmm_size,
                            struct llog_cookie *logcookies, int log_unlink);
 int mds_cleanup_pending(struct obd_device *obd);
+int mds_check_stale_orphan(struct obd_device *obd, struct ll_fid *fid);
 
 /* mds/mds_log.c */
 int mds_log_op_unlink(struct obd_device *obd,
                       struct lov_mds_md *lmm, int lmm_size,
                       struct llog_cookie *logcookies, int cookies_size);
+int mds_log_op_orphan(struct obd_device *, struct lov_stripe_md *, obd_count);
 int mds_log_op_setattr(struct obd_device *obd, struct inode *inode,
                       struct lov_mds_md *lmm, int lmm_size,
                       struct llog_cookie *logcookies, int cookies_size);
-int mds_llog_init(struct obd_device *obd, struct obd_device *tgt, int count,
-                  struct llog_catid *logid, struct obd_uuid *uuid);
+int mds_llog_init(struct obd_device *obd, struct obd_device *disk_obd,
+                  int *index);
 int mds_llog_finish(struct obd_device *obd, int count);
 
 /* mds/mds_lov.c */
@@ -232,7 +240,8 @@ int mds_notify(struct obd_device *obd, struct obd_device *watched,
 int mds_get_default_md(struct obd_device *obd, struct lov_mds_md *lmm,
                        int *lmmsize);
 int mds_convert_lov_ea(struct obd_device *obd, struct inode *inode,
-                       struct lov_mds_md *lmm, int lmm_size);
+                       struct lov_mds_md *lmm, int lmm_size,
+                       __u64 connect_flags);
 int mds_init_lov_desc(struct obd_device *obd, struct obd_export *osc_exp);
 
 /* mds/mds_open.c */
@@ -254,8 +263,11 @@ int mds_join_file(struct mds_update_record *rec, struct ptlrpc_request *req,
                   struct dentry *dchild, struct lustre_handle *lockh);
 
 /* mds/mds_fs.c */
+int mds_update_client_epoch(struct obd_export *exp);
+void mds_update_last_epoch(struct obd_device *obd);
 int mds_export_stats_init(struct obd_device *obd,
                           struct obd_export *exp,
+                          int reconnect,
                           void *client_nid);
 int mds_client_add(struct obd_device *obd, struct obd_export *exp,
                    int cl_off, void *localdata);
@@ -274,10 +286,10 @@ int mds_postrecov(struct obd_device *obd);
 int mds_init_export(struct obd_export *exp);
 #ifdef __KERNEL__
 int mds_get_md(struct obd_device *, struct inode *, void *md, int *size,
-               int lock, int flags);
+               int lock, int flags, __u64 connect_flags);
 int mds_pack_md(struct obd_device *, struct lustre_msg *, int offset,
-                struct mds_body *, struct inode *, int lock, int flags);
-void mds_pack_inode2fid(struct ll_fid *fid, struct inode *inode);
+                struct mds_body *, struct inode *, int lock, int flags,
+                __u64 connect_flags);
 void mds_pack_inode2body(struct mds_body *body, struct inode *inode);
 #endif
 int mds_pack_acl(struct mds_export_data *med, struct inode *inode,
