@@ -210,6 +210,7 @@ static int config_log_add(char *logname, struct config_llog_instance *cfg,
         cld->cld_cfg.cfg_flags = 0;
         cld->cld_cfg.cfg_sb = sb;
         atomic_set(&cld->cld_refcount, 1);
+        init_mutex(&cld->cld_sem);
 
         /* Keep the mgc around until we are done */
         cld->cld_mgcexp = class_export_get(lsi->lsi_mgc->obd_self_export);
@@ -232,8 +233,6 @@ static int config_log_add(char *logname, struct config_llog_instance *cfg,
         RETURN(rc);
 }
 
-DECLARE_MUTEX(llog_process_lock);
-
 /* Stop watching for updates on this log. */
 static int config_log_end(char *logname, struct config_llog_instance *cfg)
 {
@@ -247,9 +246,9 @@ static int config_log_end(char *logname, struct config_llog_instance *cfg)
         /* drop the ref from the find */
         config_log_put(cld);
 
-        down(&llog_process_lock);
+        down(&cld->cld_sem);
         cld->cld_stopping = 1;
-        up(&llog_process_lock);
+        up(&cld->cld_sem);
 
         /* drop the start ref */
         config_log_put(cld);
@@ -277,7 +276,7 @@ static int mgc_requeue_thread(void *data)
         int rc = 0;
         ENTRY;
 
-        ptlrpc_daemonize(name);
+        cfs_daemonize(name);
 
         CDEBUG(D_MGC, "Starting requeue thread\n");
 
@@ -415,7 +414,7 @@ static int mgc_fs_setup(struct obd_device *obd, struct super_block *sb,
         obd->obd_lvfs_ctxt.fs = get_ds();
 
         push_ctxt(&saved, &obd->obd_lvfs_ctxt, NULL);
-        dentry = lookup_one_len(MOUNT_CONFIGS_DIR, current->fs->pwd,
+        dentry = lookup_one_len(MOUNT_CONFIGS_DIR, cfs_fs_pwd(current->fs),
                                 strlen(MOUNT_CONFIGS_DIR));
         pop_ctxt(&saved, &obd->obd_lvfs_ctxt, NULL);
         if (IS_ERR(dentry)) {
@@ -489,6 +488,17 @@ static int mgc_precleanup(struct obd_device *obd, enum obd_cleanup_stage stage)
                         spin_unlock(&config_list_lock);
                         cfs_waitq_signal(&rq_waitq);
                 }
+                /* client import will not have been cleaned. */
+                down_write(&obd->u.cli.cl_sem);
+                if (obd->u.cli.cl_import) {
+                        struct obd_import *imp;
+                        imp = obd->u.cli.cl_import;
+                        CERROR("client import never connected\n");
+                        class_destroy_import(imp);
+                        obd->u.cli.cl_import = NULL;
+                }
+                up_write(&obd->u.cli.cl_sem);
+
                 rc = obd_llog_finish(obd, 0);
                 if (rc != 0)
                         CERROR("failed to cleanup llogging subsystems\n");
@@ -534,7 +544,7 @@ static int mgc_setup(struct obd_device *obd, obd_count len, void *buf)
         if (rc)
                 GOTO(err_decref, rc);
 
-        rc = obd_llog_init(obd, obd, 0, NULL, NULL);
+        rc = obd_llog_init(obd, obd, NULL);
         if (rc) {
                 CERROR("failed to setup llogging subsystems\n");
                 GOTO(err_cleanup, rc);
@@ -662,7 +672,8 @@ static int mgc_enqueue(struct obd_export *exp, struct lov_stripe_md *lsm,
 }
 
 static int mgc_cancel(struct obd_export *exp, struct lov_stripe_md *md,
-                      __u32 mode, struct lustre_handle *lockh)
+                      __u32 mode, struct lustre_handle *lockh, int flags,
+                      obd_off end)
 {
         ENTRY;
 
@@ -897,7 +908,6 @@ static int mgc_import_event(struct obd_device *obd,
 
         switch (event) {
         case IMP_EVENT_DISCON:
-                /* MGC imports should not wait for recovery */
                 break;
         case IMP_EVENT_INACTIVE:
                 break;
@@ -920,20 +930,19 @@ static int mgc_import_event(struct obd_device *obd,
         RETURN(rc);
 }
 
-static int mgc_llog_init(struct obd_device *obd, struct obd_device *tgt,
-                         int count, struct llog_catid *logid,
-                         struct obd_uuid *uuid)
+static int mgc_llog_init(struct obd_device *obd, struct obd_device *disk_obd,
+                         int *index)
 {
         struct llog_ctxt *ctxt;
         int rc;
         ENTRY;
 
-        rc = llog_setup(obd, LLOG_CONFIG_ORIG_CTXT, tgt, 0, NULL,
+        rc = llog_setup(obd, LLOG_CONFIG_ORIG_CTXT, disk_obd, 0, NULL,
                         &llog_lvfs_ops);
         if (rc)
                 RETURN(rc);
 
-        rc = llog_setup(obd, LLOG_CONFIG_REPL_CTXT, tgt, 0, NULL,
+        rc = llog_setup(obd, LLOG_CONFIG_REPL_CTXT, disk_obd, 0, NULL,
                         &llog_client_ops);
         if (rc == 0) {
                 ctxt = llog_get_context(obd, LLOG_CONFIG_REPL_CTXT);
@@ -1101,14 +1110,10 @@ static int mgc_process_log(struct obd_device *mgc,
                 RETURN(-EINVAL);
         }
 
-        /* I don't want mutliple processes running process_log at once --
-           sounds like badness.  It actually might be fine, as long as
-           we're not trying to update from the same log
-           simultaneously (in which case we should use a per-log sem.) */
-        down(&llog_process_lock);
-
+        /* Serialize update from the same log */
+        down(&cld->cld_sem);
         if (cld->cld_stopping) {
-                up(&llog_process_lock);
+                up(&cld->cld_sem);
                 RETURN(0);
         }
 
@@ -1121,8 +1126,8 @@ static int mgc_process_log(struct obd_device *mgc,
 
         ctxt = llog_get_context(mgc, LLOG_CONFIG_REPL_CTXT);
         if (!ctxt) {
+                up(&cld->cld_sem);
                 CERROR("missing llog context\n");
-                up(&llog_process_lock);
                 RETURN(-EINVAL);
         }
 
@@ -1178,16 +1183,15 @@ out_pop:
         /* Now drop the lock so MGS can revoke it */
         if (!rcl) {
                 rcl = mgc_cancel(mgc->u.cli.cl_mgc_mgsexp, NULL,
-                                 LCK_CR, &lockh);
+                                 LCK_CR, &lockh, 0, 0);
                 if (rcl)
                         CERROR("Can't drop cfg lock: %d\n", rcl);
         }
+        up(&cld->cld_sem);
 
         CDEBUG(D_MGC, "%s: configuration from log '%s' %sed (%d).\n",
                mgc->obd_name, cld->cld_logname, rc ? "fail" : "succeed", rc);
 
-        up(&llog_process_lock);
-
         RETURN(rc);
 }
 
