@@ -47,6 +47,7 @@
 #include <obd_class.h>
 #include <lprocfs_status.h>
 #include <class_hash.h>
+#include <lustre_export.h>
 
 extern struct list_head obd_types;
 spinlock_t obd_types_lock;
@@ -115,12 +116,12 @@ struct obd_type *class_get_type(const char *name)
 {
         struct obd_type *type = class_search_type(name);
 
-#ifdef CONFIG_KMOD
+#ifdef HAVE_MODULE_LOADING_SUPPORT
         if (!type) {
                 const char *modname = name;
                 if (strcmp(modname, LUSTRE_MDT_NAME) == 0)
                         modname = LUSTRE_MDS_NAME;
-                if (!request_module(modname)) {
+                if (!request_module("%s", modname)) {
                         CDEBUG(D_INFO, "Loaded module '%s'\n", modname);
                         type = class_search_type(name);
                 } else {
@@ -643,6 +644,8 @@ static void export_handle_addref(void *export)
         class_export_get(export);
 }
 
+/* called from mds_commit_cb() in context of journal commit callback
+ * and cannot call any blocking functions. */
 void __class_export_put(struct obd_export *exp)
 {
         if (atomic_dec_and_test(&exp->exp_refcount)) {
@@ -676,6 +679,7 @@ void class_export_destroy(struct obd_export *exp)
                 ptlrpc_put_connection_superhack(exp->exp_connection);
 
         LASSERT(list_empty(&exp->exp_outstanding_replies));
+        LASSERT(list_empty(&exp->exp_uncommitted_replies));
         LASSERT(list_empty(&exp->exp_req_replay_queue));
         LASSERT(list_empty(&exp->exp_queued_rpc));
         obd_destroy_export(exp);
@@ -703,8 +707,11 @@ struct obd_export *class_new_export(struct obd_device *obd,
         atomic_set(&export->exp_rpc_count, 0);
         export->exp_obd = obd;
         CFS_INIT_LIST_HEAD(&export->exp_outstanding_replies);
+        spin_lock_init(&export->exp_uncommitted_replies_lock);
+        CFS_INIT_LIST_HEAD(&export->exp_uncommitted_replies);
         CFS_INIT_LIST_HEAD(&export->exp_req_replay_queue);
         CFS_INIT_LIST_HEAD(&export->exp_queued_rpc);
+
         CFS_INIT_LIST_HEAD(&export->exp_handle.h_link);
         class_handle_hash(&export->exp_handle, export_handle_addref);
         export->exp_last_request_time = cfs_time_current_sec();
@@ -757,8 +764,27 @@ void class_unlink_export(struct obd_export *exp)
         spin_unlock(&exp->exp_obd->obd_dev_lock);
         /* Keep these counter valid always */
         spin_lock_bh(&exp->exp_obd->obd_processing_task_lock);
-        if (exp->exp_replay_needed)
-                exp->exp_obd->obd_recoverable_clients--;
+        if (exp->exp_delayed) {
+                spin_lock(&exp->exp_lock);
+                exp->exp_delayed = 0;
+                spin_unlock(&exp->exp_lock);
+                LASSERT(exp->exp_obd->obd_delayed_clients);
+                exp->exp_obd->obd_delayed_clients--;
+        } else if (exp->exp_replay_needed) {
+                        spin_lock(&exp->exp_lock);
+                        exp->exp_replay_needed = 0;
+                        spin_unlock(&exp->exp_lock);
+                        LASSERT(exp->exp_obd->obd_recoverable_clients);
+                        exp->exp_obd->obd_recoverable_clients--;
+        }
+
+        if (exp->exp_obd->obd_recovering && exp->exp_in_recovery) {
+                spin_lock(&exp->exp_lock);
+                exp->exp_in_recovery = 0;
+                spin_unlock(&exp->exp_lock);
+                LASSERT(exp->exp_obd->obd_connected_clients);
+                exp->exp_obd->obd_connected_clients--;
+        }
         spin_unlock_bh(&exp->exp_obd->obd_processing_task_lock);
         class_export_put(exp);
 }
@@ -872,6 +898,10 @@ struct obd_import *class_new_import(struct obd_device *obd)
         class_handle_hash(&imp->imp_handle, import_handle_addref);
         init_imp_at(&imp->imp_at);
 
+/* b1_8 supports both v1 & v2. but HEAD only supports v2.
+ * So let's use v2.
+ */
+#define HAVE_DEFAULT_V2_CONNECT 1
 #ifdef HAVE_DEFAULT_V2_CONNECT
         /* the default magic is V2, will be used in connect RPC, and
          * then adjusted according to the flags in request/reply. */
@@ -927,10 +957,12 @@ int class_connect(struct lustre_handle *conn, struct obd_device *obd,
 }
 EXPORT_SYMBOL(class_connect);
 
-/* This function removes two references from the export: one for the
- * hash entry and one for the export pointer passed in.  The export
- * pointer passed to this function is destroyed should not be used
- * again. */
+/* This function removes 1-3 references from the export:
+ * 1 - for export pointer passed
+ * and if disconnect really need
+ * 2 - removing from hash
+ * 3 - in client_unlink_export
+ * The export pointer passed to this function can destroyed */
 int class_disconnect(struct obd_export *export)
 {
         int already_disconnected;
@@ -945,40 +977,57 @@ int class_disconnect(struct obd_export *export)
         spin_lock(&export->exp_lock);
         already_disconnected = export->exp_disconnected;
         export->exp_disconnected = 1;
+        spin_unlock(&export->exp_lock);
+
+
+        /* class_cleanup(), abort_recovery(), and class_fail_export()
+         * all end up in here, and if any of them race we shouldn't
+         * call extra class_export_puts(). */
+        if (already_disconnected)
+                GOTO(no_disconn, already_disconnected);
+
+        CDEBUG(D_IOCTL, "disconnect: cookie "LPX64"\n",
+               export->exp_handle.h_cookie);
+
 
         if (!hlist_unhashed(&export->exp_nid_hash))
                 lustre_hash_del(export->exp_obd->obd_nid_hash,
                                 &export->exp_connection->c_peer.nid,
                                 &export->exp_nid_hash);
 
-        spin_unlock(&export->exp_lock);
-
-        /* class_cleanup(), abort_recovery(), and class_fail_export()
-         * all end up in here, and if any of them race we shouldn't
-         * call extra class_export_puts(). */
-        if (already_disconnected)
-                RETURN(0);
-
-        CDEBUG(D_IOCTL, "disconnect: cookie "LPX64"\n",
-               export->exp_handle.h_cookie);
-
         class_unlink_export(export);
+
+no_disconn:
         class_export_put(export);
         RETURN(0);
 }
 
+/* Return non-zero for a fully connected export */
+int class_connected_export(struct obd_export *exp)
+{
+        if (exp) {
+                int connected;
+                spin_lock(&exp->exp_lock);
+                connected = (exp->exp_conn_cnt > 0);
+                spin_unlock(&exp->exp_lock);
+                return connected;
+        }
+        return 0;
+}
+EXPORT_SYMBOL(class_connected_export);
+
 static void class_disconnect_export_list(struct list_head *list,
                                          enum obd_option flags)
 {
         int rc;
-        struct lustre_handle fake_conn;
-        struct obd_export *fake_exp, *exp;
+        struct obd_export *exp;
         ENTRY;
 
         /* It's possible that an export may disconnect itself, but
          * nothing else will be added to this list. */
         while (!list_empty(list)) {
                 exp = list_entry(list->next, struct obd_export, exp_obd_chain);
+                /* need for safe call CDEBUG after obd_disconnect */
                 class_export_get(exp);
 
                 spin_lock(&exp->exp_lock);
@@ -997,22 +1046,16 @@ static void class_disconnect_export_list(struct list_head *list,
                         continue;
                 }
 
-                fake_conn.cookie = exp->exp_handle.h_cookie;
-                fake_exp = class_conn2export(&fake_conn);
-                if (!fake_exp) {
-                        class_export_put(exp);
-                        continue;
-                }
-
-                spin_lock(&fake_exp->exp_lock);
-                fake_exp->exp_flags = flags;
-                spin_unlock(&fake_exp->exp_lock);
-
+                class_export_get(exp);
                 CDEBUG(D_HA, "%s: disconnecting export at %s (%p), "
                        "last request at %ld\n",
                        exp->exp_obd->obd_name, obd_export_nid2str(exp),
                        exp, exp->exp_last_request_time);
-                rc = obd_disconnect(fake_exp);
+
+                /* release one export reference anyway */
+                rc = obd_disconnect(exp);
+                CDEBUG(D_HA, "disconnected export at %s (%p): rc %d\n",
+                       obd_export_nid2str(exp), exp, rc);
                 class_export_put(exp);
         }
         EXIT;
@@ -1024,9 +1067,10 @@ void class_disconnect_exports(struct obd_device *obd)
         ENTRY;
 
         /* Move all of the exports from obd_exports to a work list, en masse. */
+        CFS_INIT_LIST_HEAD(&work_list);
         spin_lock(&obd->obd_dev_lock);
-        list_add(&work_list, &obd->obd_exports);
-        list_del_init(&obd->obd_exports);
+        list_splice_init(&obd->obd_delayed_exports, &work_list);
+        list_splice_init(&obd->obd_exports, &work_list);
         spin_unlock(&obd->obd_dev_lock);
 
         CDEBUG(D_HA, "OBD device %d (%p) has exports, "
@@ -1036,15 +1080,13 @@ void class_disconnect_exports(struct obd_device *obd)
 }
 EXPORT_SYMBOL(class_disconnect_exports);
 
-/* Remove exports that have not completed recovery.
- */
+/* Remove exports that have not completed recovery. */
 void class_disconnect_stale_exports(struct obd_device *obd,
                                     enum obd_option flags)
 {
         struct list_head work_list;
         struct list_head *pos, *n;
         struct obd_export *exp;
-        int cnt = 0;
         ENTRY;
 
         CFS_INIT_LIST_HEAD(&work_list);
@@ -1052,19 +1094,128 @@ void class_disconnect_stale_exports(struct obd_device *obd,
         list_for_each_safe(pos, n, &obd->obd_exports) {
                 exp = list_entry(pos, struct obd_export, exp_obd_chain);
                 if (exp->exp_replay_needed) {
-                        list_del(&exp->exp_obd_chain);
-                        list_add(&exp->exp_obd_chain, &work_list);
+                        list_move(&exp->exp_obd_chain, &work_list);
+                        obd->obd_stale_clients++;
+                }
+        }
+        spin_unlock(&obd->obd_dev_lock);
+
+        CDEBUG(D_HA, "%s: disconnecting %d stale clients\n",
+               obd->obd_name, obd->obd_stale_clients);
+        class_disconnect_export_list(&work_list, flags);
+        EXIT;
+}
+EXPORT_SYMBOL(class_disconnect_stale_exports);
+
+void class_disconnect_expired_exports(struct obd_device *obd)
+{
+        struct list_head expired_list;
+        struct obd_export *exp, *n;
+        int cnt = 0;
+        ENTRY;
+
+        CFS_INIT_LIST_HEAD(&expired_list);
+        spin_lock(&obd->obd_dev_lock);
+        list_for_each_entry_safe(exp, n, &obd->obd_delayed_exports,
+                                 exp_obd_chain) {
+                if (exp_expired(exp, obd->u.obt.obt_stale_export_age)) {
+                        list_move(&exp->exp_obd_chain, &expired_list);
                         cnt++;
                 }
         }
         spin_unlock(&obd->obd_dev_lock);
 
-        CDEBUG(D_ERROR, "%s: disconnecting %d stale clients\n",
+        if (cnt == 0)
+                return;
+
+        CDEBUG(D_INFO, "%s: disconnecting %d expired exports\n",
                obd->obd_name, cnt);
-        class_disconnect_export_list(&work_list, flags);
+        class_disconnect_export_list(&expired_list, exp_flags_from_obd(obd));
+
         EXIT;
 }
-EXPORT_SYMBOL(class_disconnect_stale_exports);
+EXPORT_SYMBOL(class_disconnect_expired_exports);
+
+void class_set_export_delayed(struct obd_export *exp)
+{
+        struct obd_device *obd = class_exp2obd(exp);
+
+        LASSERT(!exp->exp_delayed);
+
+        /* no need to ping delayed exports */
+        spin_lock(&obd->obd_dev_lock);
+        list_del_init(&exp->exp_obd_chain_timed);
+        list_move_tail(&exp->exp_obd_chain, &obd->obd_delayed_exports);
+        spin_unlock(&obd->obd_dev_lock);
+
+        LASSERT(obd->obd_recoverable_clients > 0);
+
+        spin_lock_bh(&obd->obd_processing_task_lock);
+        /* race with target_queue_last_replay_reply? */
+        if (exp->exp_replay_needed) {
+                spin_lock(&exp->exp_lock);
+                exp->exp_delayed = 1;
+                spin_unlock(&exp->exp_lock);
+
+                obd->obd_delayed_clients++;
+                obd->obd_recoverable_clients--;
+        }
+        spin_unlock_bh(&obd->obd_processing_task_lock);
+
+        CDEBUG(D_HA, "%s: set client %s as delayed\n",
+               obd->obd_name, exp->exp_client_uuid.uuid);
+}
+EXPORT_SYMBOL(class_set_export_delayed);
+
+/*
+ * Manage exports that have not completed recovery.
+ */
+void class_handle_stale_exports(struct obd_device *obd)
+{
+        struct list_head delay_list, evict_list;
+        struct obd_export *exp, *n;
+        int delayed = 0;
+        ENTRY;
+
+        CFS_INIT_LIST_HEAD(&delay_list);
+        CFS_INIT_LIST_HEAD(&evict_list);
+        spin_lock(&obd->obd_dev_lock);
+        list_for_each_entry_safe(exp, n, &obd->obd_exports, exp_obd_chain) {
+                LASSERT(!exp->exp_delayed);
+                /* clients finished recovery */
+                if (!exp->exp_replay_needed)
+                        continue;
+                /* connected non-vbr clients are evicted */
+                if (exp->exp_in_recovery && !exp_connect_vbr(exp)) {
+                        obd->obd_stale_clients++;
+                        list_move_tail(&exp->exp_obd_chain, &evict_list);
+                        continue;
+                }
+                if (obd->obd_version_recov || !exp->exp_in_recovery) {
+                        list_move_tail(&exp->exp_obd_chain, &delay_list);
+                        delayed++;
+                }
+        }
+#ifndef HAVE_DELAYED_RECOVERY
+        /* delayed recovery is turned off, evict all delayed exports */
+        list_splice_init(&delay_list, &evict_list);
+        list_splice_init(&obd->obd_delayed_exports, &evict_list);
+        obd->obd_stale_clients += delayed;
+#endif
+        spin_unlock(&obd->obd_dev_lock);
+
+        list_for_each_entry_safe(exp, n, &delay_list, exp_obd_chain) {
+                class_set_export_delayed(exp);
+                exp->exp_last_request_time = cfs_time_current_sec();
+        }
+        LASSERT(list_empty(&delay_list));
+
+        /* evict clients without VBR support */
+        class_disconnect_export_list(&evict_list, exp_flags_from_obd(obd));
+
+        EXIT;
+}
+EXPORT_SYMBOL(class_handle_stale_exports);
 
 int oig_init(struct obd_io_group **oig_out)
 {
@@ -1342,7 +1493,7 @@ void obd_zombie_impexp_cull(void)
 
                 if (export != NULL)
                         class_export_destroy(export);
-
+                cfs_cond_resched();
         } while (import != NULL || export != NULL);
 }
 
@@ -1350,6 +1501,7 @@ static struct completion        obd_zombie_start;
 static struct completion        obd_zombie_stop;
 static unsigned long            obd_zombie_flags;
 static cfs_waitq_t              obd_zombie_waitq;
+static pid_t                    obd_zombie_pid;
 
 enum {
         OBD_ZOMBIE_STOP = 1
@@ -1374,6 +1526,35 @@ static void obd_zombie_impexp_notify(void)
         cfs_waitq_signal(&obd_zombie_waitq);
 }
 
+/**
+ * check whether obd_zombie is idle
+ */
+static int obd_zombie_is_idle(void)
+{
+        int rc;
+
+        LASSERT(!test_bit(OBD_ZOMBIE_STOP, &obd_zombie_flags));
+        spin_lock(&obd_zombie_impexp_lock);
+        rc = list_empty(&obd_zombie_imports) &&
+             list_empty(&obd_zombie_exports);
+        spin_unlock(&obd_zombie_impexp_lock);
+        return rc;
+}
+
+/**
+ * wait when obd_zombie import/export queues become empty
+ */
+void obd_zombie_barrier(void)
+{
+        struct l_wait_info lwi = { 0 };
+
+        if (obd_zombie_pid == cfs_curproc_pid())
+                /* don't wait for myself */
+                return;
+        l_wait_event(obd_zombie_waitq, obd_zombie_is_idle(), &lwi);
+}
+EXPORT_SYMBOL(obd_zombie_barrier);
+
 #ifdef __KERNEL__
 
 static int obd_zombie_impexp_thread(void *unused)
@@ -1387,12 +1568,20 @@ static int obd_zombie_impexp_thread(void *unused)
 
         complete(&obd_zombie_start);
 
+        obd_zombie_pid = cfs_curproc_pid();
+
         while(!test_bit(OBD_ZOMBIE_STOP, &obd_zombie_flags)) {
                 struct l_wait_info lwi = { 0 };
 
                 l_wait_event(obd_zombie_waitq, !obd_zombi_impexp_check(NULL), &lwi);
 
                 obd_zombie_impexp_cull();
+
+                /*
+                 * Notify obd_zombie_barrier callers that queues
+                 * may be empty.
+                 */
+                cfs_waitq_signal(&obd_zombie_waitq);
         }
 
         complete(&obd_zombie_stop);
@@ -1430,6 +1619,7 @@ int obd_zombie_impexp_init(void)
         init_completion(&obd_zombie_start);
         init_completion(&obd_zombie_stop);
         cfs_waitq_init(&obd_zombie_waitq);
+        obd_zombie_pid = 0;
 
 #ifdef __KERNEL__
         rc = cfs_kernel_thread(obd_zombie_impexp_thread, NULL, 0);
