@@ -56,7 +56,9 @@
 
 cfs_sysctl_table_header_t *obd_table_header = NULL;
 
-#define OBD_SYSCTL 300
+#ifndef HAVE_SYSCTL_UNNUMBERED
+
+#define CTL_LUSTRE      300
 
 enum {
         OBD_FAIL_LOC = 1,       /* control test failures instrumentation */
@@ -73,7 +75,34 @@ enum {
         OBD_DEBUG_PEER_ON_TIMEOUT, /* dump peer debug when RPC times out */
         OBD_ALLOC_FAIL_RATE,    /* memory allocation random failure rate */
         OBD_MAX_DIRTY_PAGES,    /* maximum dirty pages */
+        OBD_AT_MIN,             /* Adaptive timeouts params */
+        OBD_AT_MAX,
+        OBD_AT_EXTRA,
+        OBD_AT_EARLY_MARGIN,
+        OBD_AT_HISTORY,
 };
+#else
+#define CTL_LUSTRE              CTL_UNNUMBERED
+#define OBD_FAIL_LOC            CTL_UNNUMBERED
+#define OBD_FAIL_VAL            CTL_UNNUMBERED
+#define OBD_TIMEOUT             CTL_UNNUMBERED
+#define OBD_DUMP_ON_TIMEOUT     CTL_UNNUMBERED
+#define OBD_MEMUSED             CTL_UNNUMBERED
+#define OBD_PAGESUSED           CTL_UNNUMBERED
+#define OBD_MAXMEMUSED          CTL_UNNUMBERED
+#define OBD_MAXPAGESUSED        CTL_UNNUMBERED
+#define OBD_SYNCFILTER          CTL_UNNUMBERED
+#define OBD_LDLM_TIMEOUT        CTL_UNNUMBERED
+#define OBD_DUMP_ON_EVICTION    CTL_UNNUMBERED
+#define OBD_DEBUG_PEER_ON_TIMEOUT CTL_UNNUMBERED
+#define OBD_ALLOC_FAIL_RATE     CTL_UNNUMBERED
+#define OBD_MAX_DIRTY_PAGES     CTL_UNNUMBERED
+#define OBD_AT_MIN              CTL_UNNUMBERED
+#define OBD_AT_MAX              CTL_UNNUMBERED
+#define OBD_AT_EXTRA            CTL_UNNUMBERED
+#define OBD_AT_EARLY_MARGIN     CTL_UNNUMBERED
+#define OBD_AT_HISTORY          CTL_UNNUMBERED
+#endif
 
 int LL_PROC_PROTO(proc_fail_loc)
 {
@@ -120,7 +149,8 @@ int LL_PROC_PROTO(proc_max_dirty_pages_in_mb)
                         obd_max_dirty_pages = 4 << (20 - CFS_PAGE_SHIFT);
                 }
         } else {
-                char buf[21];
+                char buf[22];
+                struct ctl_table dummy;
                 int len;
 
                 len = lprocfs_read_frac_helper(buf, sizeof(buf),
@@ -129,7 +159,13 @@ int LL_PROC_PROTO(proc_max_dirty_pages_in_mb)
                 if (len > *lenp)
                         len = *lenp;
                 buf[len] = '\0';
-                if (copy_to_user(buffer, buf, len))
+
+                dummy = *table;
+                dummy.data = buf;
+                dummy.maxlen = sizeof(buf);
+
+                rc = ll_proc_dostring(&dummy,write,filp,buffer,lenp, ppos);
+                if (rc)
                         return -EFAULT;
                 *lenp = len;
         }
@@ -152,7 +188,8 @@ int LL_PROC_PROTO(proc_alloc_fail_rate)
                                                (unsigned int*)table->data,
                                                OBD_ALLOC_FAIL_MULT);
         } else {
-                char buf[21];
+                char buf[22];
+                struct ctl_table dummy;
                 int  len;
 
                 len = lprocfs_read_frac_helper(buf, sizeof(buf),
@@ -161,7 +198,12 @@ int LL_PROC_PROTO(proc_alloc_fail_rate)
                 if (len > *lenp)
                         len = *lenp;
                 buf[len] = '\0';
-                if (copy_to_user(buffer, buf, len))
+                dummy = *table;
+                dummy.data = buf;
+                dummy.maxlen = sizeof(buf);
+
+                rc = ll_proc_dostring(&dummy,write,filp,buffer,lenp, ppos);
+                if(rc)
                         return -EFAULT;
                 *lenp = len;
         }
@@ -172,6 +214,7 @@ int LL_PROC_PROTO(proc_alloc_fail_rate)
 
 int LL_PROC_PROTO(proc_memory_alloc)
 {
+        struct ctl_table dummy;
         char buf[22];
         int len;
         DECLARE_LL_PROC_PPOS_DECL;
@@ -187,15 +230,17 @@ int LL_PROC_PROTO(proc_memory_alloc)
         if (len > *lenp)
                 len = *lenp;
         buf[len] = '\0';
-        if (copy_to_user(buffer, buf, len))
-                return -EFAULT;
-        *lenp = len;
-        *ppos += *lenp;
-        return 0;
+
+        dummy = *table;
+        dummy.data = buf;
+        dummy.maxlen = sizeof(buf);
+
+        return ll_proc_dostring(&dummy,write,filp,buffer,lenp, ppos);
 }
 
 int LL_PROC_PROTO(proc_pages_alloc)
 {
+        struct ctl_table dummy;
         char buf[22];
         int len;
         DECLARE_LL_PROC_PPOS_DECL;
@@ -211,15 +256,17 @@ int LL_PROC_PROTO(proc_pages_alloc)
         if (len > *lenp)
                 len = *lenp;
         buf[len] = '\0';
-        if (copy_to_user(buffer, buf, len))
-                return -EFAULT;
-        *lenp = len;
-        *ppos += *lenp;
-        return 0;
+
+        dummy = *table;
+        dummy.data = buf;
+        dummy.maxlen = sizeof(buf);
+
+        return ll_proc_dostring(&dummy,write,filp,buffer,lenp, ppos);
 }
 
 int LL_PROC_PROTO(proc_mem_max)
 {
+        struct ctl_table dummy;
         char buf[22];
         int len;
         DECLARE_LL_PROC_PPOS_DECL;
@@ -235,17 +282,19 @@ int LL_PROC_PROTO(proc_mem_max)
         if (len > *lenp)
                 len = *lenp;
         buf[len] = '\0';
-        if (copy_to_user(buffer, buf, len))
-                return -EFAULT;
-        *lenp = len;
-        *ppos += *lenp;
-        return 0;
+
+        dummy = *table;
+        dummy.data = buf;
+        dummy.maxlen = sizeof(buf);
+
+        return ll_proc_dostring(&dummy,write,filp,buffer,lenp, ppos);
 }
 
 int LL_PROC_PROTO(proc_pages_max)
 {
         char buf[22];
         int len;
+        struct ctl_table dummy;
         DECLARE_LL_PROC_PPOS_DECL;
 
         if (!*lenp || (*ppos && !write)) {
@@ -254,16 +303,38 @@ int LL_PROC_PROTO(proc_pages_max)
         }
         if (write)
                 return -EINVAL;
+         dummy = *table;
+         dummy.data = buf;
+         dummy.maxlen = sizeof(buf);
+         len = snprintf(buf, sizeof(buf), LPU64,
+                        obd_pages_max());
 
-        len = snprintf(buf, sizeof(buf), LPU64"\n", obd_pages_max());
-        if (len > *lenp)
-                len = *lenp;
-        buf[len] = '\0';
-        if (copy_to_user(buffer, buf, len))
-                return -EFAULT;
-        *lenp = len;
-        *ppos += *lenp;
-        return 0;
+         if (len > *lenp)
+                 len = *lenp;
+         buf[len] = '\0';
+
+         return ll_proc_dostring(&dummy,write,filp,buffer,lenp, ppos);
+}
+
+int LL_PROC_PROTO(proc_at_min)
+{
+        return ll_proc_dointvec(table, write, filp, buffer, lenp, ppos);
+}
+int LL_PROC_PROTO(proc_at_max)
+{
+        return ll_proc_dointvec(table, write, filp, buffer, lenp, ppos);
+}
+int LL_PROC_PROTO(proc_at_extra)
+{
+        return ll_proc_dointvec(table, write, filp, buffer, lenp, ppos);
+}
+int LL_PROC_PROTO(proc_at_early_margin)
+{
+        return ll_proc_dointvec(table, write, filp, buffer, lenp, ppos);
+}
+int LL_PROC_PROTO(proc_at_history)
+{
+        return ll_proc_dointvec(table, write, filp, buffer, lenp, ppos);
 }
 
 static cfs_sysctl_table_t obd_table[] = {
@@ -281,7 +352,8 @@ static cfs_sysctl_table_t obd_table[] = {
                 .data     = &obd_fail_val,
                 .maxlen   = sizeof(int),
                 .mode     = 0644,
-                .proc_handler = &proc_dointvec
+                .proc_handler = &proc_dointvec,
+                .strategy = &sysctl_intvec,
         },
         {
                 .ctl_name = OBD_TIMEOUT,
@@ -297,7 +369,7 @@ static cfs_sysctl_table_t obd_table[] = {
                 .data     = &obd_debug_peer_on_timeout,
                 .maxlen   = sizeof(int),
                 .mode     = 0644,
-                .proc_handler = &proc_dointvec
+                .proc_handler = &proc_dointvec,
         },
         {
                 .ctl_name = OBD_DUMP_ON_TIMEOUT,
@@ -305,7 +377,7 @@ static cfs_sysctl_table_t obd_table[] = {
                 .data     = &obd_dump_on_timeout,
                 .maxlen   = sizeof(int),
                 .mode     = 0644,
-                .proc_handler = &proc_dointvec
+                .proc_handler = &proc_dointvec,
         },
         {
                 .ctl_name = OBD_DUMP_ON_EVICTION,
@@ -313,7 +385,7 @@ static cfs_sysctl_table_t obd_table[] = {
                 .data     = &obd_dump_on_eviction,
                 .maxlen   = sizeof(int),
                 .mode     = 0644,
-                .proc_handler = &proc_dointvec
+                .proc_handler = &proc_dointvec,
         },
         {
                 .ctl_name = OBD_MEMUSED,
@@ -321,7 +393,7 @@ static cfs_sysctl_table_t obd_table[] = {
                 .data     = NULL,
                 .maxlen   = 0,
                 .mode     = 0444,
-                .proc_handler = &proc_memory_alloc
+                .proc_handler = &proc_memory_alloc,
         },
         {
                 .ctl_name = OBD_PAGESUSED,
@@ -329,7 +401,7 @@ static cfs_sysctl_table_t obd_table[] = {
                 .data     = NULL,
                 .maxlen   = 0,
                 .mode     = 0444,
-                .proc_handler = &proc_pages_alloc
+                .proc_handler = &proc_pages_alloc,
         },
         {
                 .ctl_name = OBD_MAXMEMUSED,
@@ -337,7 +409,7 @@ static cfs_sysctl_table_t obd_table[] = {
                 .data     = NULL,
                 .maxlen   = 0,
                 .mode     = 0444,
-                .proc_handler = &proc_mem_max
+                .proc_handler = &proc_mem_max,
         },
         {
                 .ctl_name = OBD_MAXPAGESUSED,
@@ -345,7 +417,7 @@ static cfs_sysctl_table_t obd_table[] = {
                 .data     = NULL,
                 .maxlen   = 0,
                 .mode     = 0444,
-                .proc_handler = &proc_pages_max
+                .proc_handler = &proc_pages_max,
         },
         {
                 .ctl_name = OBD_LDLM_TIMEOUT,
@@ -373,12 +445,52 @@ static cfs_sysctl_table_t obd_table[] = {
                 .mode     = 0644,
                 .proc_handler = &proc_max_dirty_pages_in_mb
         },
+        {
+                .ctl_name = OBD_AT_MIN,
+                .procname = "at_min",
+                .data     = &at_min,
+                .maxlen   = sizeof(int),
+                .mode     = 0644,
+                .proc_handler = &proc_at_min
+        },
+        {
+                .ctl_name = OBD_AT_MAX,
+                .procname = "at_max",
+                .data     = &at_max,
+                .maxlen   = sizeof(int),
+                .mode     = 0644,
+                .proc_handler = &proc_at_max
+        },
+        {
+                .ctl_name = OBD_AT_EXTRA,
+                .procname = "at_extra",
+                .data     = &at_extra,
+                .maxlen   = sizeof(int),
+                .mode     = 0644,
+                .proc_handler = &proc_at_extra
+        },
+        {
+                .ctl_name = OBD_AT_EARLY_MARGIN,
+                .procname = "at_early_margin",
+                .data     = &at_early_margin,
+                .maxlen   = sizeof(int),
+                .mode     = 0644,
+                .proc_handler = &proc_at_early_margin
+        },
+        {
+                .ctl_name = OBD_AT_HISTORY,
+                .procname = "at_history",
+                .data     = &at_history,
+                .maxlen   = sizeof(int),
+                .mode     = 0644,
+                .proc_handler = &proc_at_history
+        },
         { 0 }
 };
 
 static cfs_sysctl_table_t parent_table[] = {
        {
-               .ctl_name = OBD_SYSCTL,
+               .ctl_name = CTL_LUSTRE,
                .procname = "lustre",
                .data     = NULL,
                .maxlen   = 0,
