@@ -99,6 +99,14 @@ void lustre_swab_ll_fid(struct ll_fid *fid)
 }
 EXPORT_SYMBOL(lustre_swab_ll_fid);
 
+void lustre_swab_lu_fid(struct lu_fid *fid)
+{
+        __swab64s(&fid->f_seq);
+        __swab32s(&fid->f_oid);
+        __swab32s(&fid->f_ver);
+}
+EXPORT_SYMBOL(lustre_swab_lu_fid);
+
 void lustre_swab_llog_rec(struct llog_rec_hdr *rec, struct llog_rec_tail *tail)
 {
         __swab32s(&rec->lrh_len);
@@ -123,8 +131,8 @@ void lustre_swab_llog_rec(struct llog_rec_hdr *rec, struct llog_rec_tail *tail)
                 struct llog_unlink_rec *lur = (struct llog_unlink_rec *)rec;
 
                 __swab64s(&lur->lur_oid);
-                __swab32s(&lur->lur_ogen);
-
+                __swab32s(&lur->lur_ogr);
+                __swab32s(&lur->lur_count);
                 break;
         }
 
@@ -132,7 +140,18 @@ void lustre_swab_llog_rec(struct llog_rec_hdr *rec, struct llog_rec_tail *tail)
                 struct llog_setattr_rec *lsr = (struct llog_setattr_rec *)rec;
 
                 __swab64s(&lsr->lsr_oid);
-                __swab32s(&lsr->lsr_ogen);
+                __swab32s(&lsr->lsr_ogr);
+                __swab32s(&lsr->lsr_uid);
+                __swab32s(&lsr->lsr_gid);
+
+                break;
+        }
+
+        case MDS_SETATTR64_REC: {
+                struct llog_setattr64_rec *lsr = (struct llog_setattr64_rec *)rec;
+
+                __swab64s(&lsr->lsr_oid);
+                __swab32s(&lsr->lsr_ogr);
                 __swab32s(&lsr->lsr_uid);
                 __swab32s(&lsr->lsr_gid);
 
