@@ -254,19 +254,19 @@ static int llog_test_3(struct obd_device *obd, struct llog_handle *llh)
                         hdr.lrh_len = 32;
                         hdr.lrh_type = OBD_CFG_REC;
                         rc = llog_write_rec(llh, &hdr, NULL, 0, buf_odd, -1);
-		}
+                }
                 if (rc) {
                         if (rc == -ENOSPC) {
                                 break;
                         } else {
-                                CERROR("3d: write recs failed at #%d: %d\n",
+                       	        CERROR("3c: write recs failed at #%d: %d\n",
                                	        i + 1, rc);
-                                RETURN(rc);
+                       	        RETURN(rc);
                         }
                 }
                 num_recs++;
         }
-        if (rc != -ENOSPC) {
+	if (rc != -ENOSPC) {
                 CWARN("3d: write record more than BITMAP size!\n");
                 RETURN(-EINVAL);
         }
@@ -502,9 +502,6 @@ static int llog_test_6(struct obd_device *obd, char *name)
         struct obd_device *mdc_obd;
         struct llog_ctxt *ctxt = llog_get_context(obd, LLOG_TEST_ORIG_CTXT);
         struct obd_uuid *mds_uuid = &ctxt->loc_exp->exp_obd->obd_uuid;
-        struct lustre_handle exph = {0, };
-        struct obd_export *exp;
-        struct obd_uuid uuid = {"LLOG_TEST6_UUID"};
         struct llog_handle *llh = NULL;
         struct llog_ctxt *nctxt;
         int rc;
@@ -517,14 +514,6 @@ static int llog_test_6(struct obd_device *obd, char *name)
                 GOTO(ctxt_release, rc = -ENOENT);
         }
 
-        rc = obd_connect(&exph, mdc_obd, &uuid, NULL /* obd_connect_data */,
-                         NULL);
-        if (rc) {
-                CERROR("6: failed to connect to MDC: %s\n", mdc_obd->obd_name);
-                GOTO(ctxt_release, rc);
-        }
-        exp = class_conn2export(&exph);
-
         nctxt = llog_get_context(mdc_obd, LLOG_CONFIG_REPL_CTXT);
         rc = llog_create(nctxt, &llh, NULL, name);
         if (rc) {
@@ -553,7 +542,7 @@ parse_out:
         if (rc) {
                 CERROR("6: llog_close failed: rc = %d\n", rc);
         }
-        rc = obd_disconnect(exp);
+
 ctxt_release:
         llog_ctxt_put(ctxt);
         RETURN(rc);
@@ -657,14 +646,13 @@ static int llog_run_tests(struct obd_device *obd)
 }
 
 
-static int llog_test_llog_init(struct obd_device *obd, struct obd_device *tgt,
-                               int count, struct llog_catid *logid,
-                               struct obd_uuid *uuid)
+static int llog_test_llog_init(struct obd_device *obd, struct obd_device *disk_obd,
+                               int *index)
 {
         int rc;
         ENTRY;
 
-        rc = llog_setup(obd, LLOG_TEST_ORIG_CTXT, tgt, 0, NULL, &llog_lvfs_ops);
+        rc = llog_setup(obd, LLOG_TEST_ORIG_CTXT, disk_obd, 0, NULL, &llog_lvfs_ops);
         RETURN(rc);
 }
 
@@ -723,7 +711,7 @@ static int llog_test_setup(struct obd_device *obd, obd_count len, void *buf)
                 RETURN(-EINVAL);
         }
 
-        rc = obd_llog_init(obd, tgt, 0, NULL, NULL);
+        rc = obd_llog_init(obd, tgt, NULL);
         if (rc)
                 RETURN(rc);
 
