@@ -78,6 +78,17 @@ static struct lustre_mount_info *server_find_mount(char *name)
         RETURN(NULL);
 }
 
+struct lustre_mount_info *server_find_mount_locked(char *name)
+{
+        struct lustre_mount_info *lmi;
+
+        down(&lustre_mount_info_lock);
+        lmi = server_find_mount(name);
+        up(&lustre_mount_info_lock);
+
+        return lmi;
+}
+
 /* we must register an obd for a mount before we call the setup routine.
    *_setup will call lustre_get_mount to get the mnt struct
    by obd_name, since we can't pass the pointer to setup. */
@@ -167,7 +178,7 @@ struct lustre_mount_info *server_get_mount(char *name)
         lsi = s2lsi(lmi->lmi_sb);
         mntget(lmi->lmi_mnt);
         atomic_inc(&lsi->lsi_mounts);
-        
+
         CDEBUG(D_MOUNT, "get_mnt %p from %s, refs=%d, vfscount=%d\n",
                lmi->lmi_mnt, name, atomic_read(&lsi->lsi_mounts),
                atomic_read(&lmi->lmi_mnt->mnt_count));
@@ -228,7 +239,6 @@ int server_put_mount(char *name, struct vfsmount *mnt)
         RETURN(0);
 }
 
-
 /******* mount helper utilities *********/
 
 static void ldd_print(struct lustre_disk_data *ldd)
@@ -510,7 +520,6 @@ static int server_start_mgs(struct super_block *sb)
                 LCONSOLE_ERROR_MSG(0x15e, "Failed to start MGS '%s' (%d).  Is "
                                    "the 'mgs' module loaded?\n",
                                    LUSTRE_MGS_OBDNAME, rc);
-
         RETURN(rc);
 }
 
@@ -729,7 +738,8 @@ static int lustre_start_mgc(struct super_block *sb)
         OBD_ALLOC_PTR(data);
         if (data == NULL)
                 GOTO(out, rc = -ENOMEM);
-        data->ocd_connect_flags = OBD_CONNECT_VERSION | OBD_CONNECT_AT;
+        data->ocd_connect_flags = OBD_CONNECT_VERSION | OBD_CONNECT_AT |
+                                  OBD_CONNECT_FID;
         data->ocd_version = LUSTRE_VERSION_CODE;
         /* We connect to the MGS at setup, and don't disconnect until cleanup */
         rc = obd_connect(&mgc_conn, obd, &(obd->obd_uuid), data, &exp);
@@ -1160,6 +1170,8 @@ struct lustre_sb_info *lustre_init_lsi(struct super_block *sb)
         }
 
         lsi->lsi_lmd->lmd_exclude_count = 0;
+        lsi->lsi_lmd->lmd_recovery_time_soft = 0;
+        lsi->lsi_lmd->lmd_recovery_time_hard = 0;
         s2lsi_nocast(sb) = lsi;
         /* we take 1 extra ref for our setup */
         atomic_set(&lsi->lsi_mounts, 1);
@@ -1338,27 +1350,38 @@ out_free:
         RETURN(ERR_PTR(rc));
 }
 
+/* Wait here forever until the mount refcount is 0 before completing umount,
+ * else we risk dereferencing a null pointer.
+ * LNET may take e.g. 165s before killing zombies. 
+*/
 static void server_wait_finished(struct vfsmount *mnt)
 {
-        wait_queue_head_t   waitq;
-        struct l_wait_info  lwi;
-        int                 retries = 330;
+        cfs_waitq_t             waitq;
+        int                     rc, waited = 0;
+        cfs_sigset_t            blocked;
 
-        init_waitqueue_head(&waitq);
+        cfs_waitq_init(&waitq);
 
-        while ((atomic_read(&mnt->mnt_count) > 1) && (retries > 0)) {
-                LCONSOLE_WARN("Mount still busy with %d refs, waiting for "
-                              "%d secs...\n",
-                              atomic_read(&mnt->mnt_count), retries);
-
-                /* Wait for a bit */
-                retries -= 5;
-                lwi = LWI_TIMEOUT(5 * HZ, NULL, NULL);
-                l_wait_event(waitq, 0, &lwi);
-        }
-        if (atomic_read(&mnt->mnt_count) > 1) {
-                CERROR("Mount %p is still busy (%d refs), giving up.\n",
-                       mnt, atomic_read(&mnt->mnt_count));
+        while (cfs_atomic_read(&mnt->mnt_count) > 1) {
+                if (waited && (waited % 30 == 0))
+                        LCONSOLE_WARN("Mount still busy with %d refs after "
+                                       "%d secs.\n",
+                                       atomic_read(&mnt->mnt_count),
+                                       waited);
+                /* Cannot use l_event_wait() for an interruptible sleep. */
+                waited += 3;
+                blocked = l_w_e_set_sigs(sigmask(SIGKILL)); 
+                rc = cfs_waitq_wait_event_interruptible_timeout(
+                        waitq, 
+                        (cfs_atomic_read(&mnt->mnt_count) == 1),
+                        cfs_time_seconds(3));
+                cfs_block_sigs(blocked);
+                if (rc < 0) {
+                        LCONSOLE_EMERG("Danger: interrupted umount %p with "
+                                       "%d refs!\n",
+                                       mnt, atomic_read(&mnt->mnt_count)); 
+                        break;
+                }
         }
 }
 
@@ -1380,6 +1403,8 @@ static void server_put_super(struct super_block *sb)
         OBD_ALLOC(tmpname, tmpname_sz);
         memcpy(tmpname, lsi->lsi_ldd->ldd_svname, tmpname_sz);
         CDEBUG(D_MOUNT, "server put_super %s\n", tmpname);
+        if (IS_MDT(lsi->lsi_ldd) && (lsi->lsi_lmd->lmd_flags & LMD_FLG_NOSVC))
+                snprintf(tmpname, tmpname_sz, "MGS");
 
         /* Stop the target */
         if (!(lsi->lsi_lmd->lmd_flags & LMD_FLG_NOSVC) &&
@@ -1643,9 +1668,6 @@ static int server_fill_super(struct super_block *sb)
         if (rc)
                 GOTO(out_mnt, rc);
 
-        LCONSOLE_WARN("Server %s on device %s has started\n",
-                      lsi->lsi_ldd->ldd_svname, lsi->lsi_lmd->lmd_dev);
-
         RETURN(0);
 out_mnt:
         /* We jump here in case of failure while starting targets or MGS.
@@ -1663,10 +1685,18 @@ int server_name2index(char *svname, __u32 *idx, char **endptr)
 {
         unsigned long index;
         int rc;
-        char *dash = strchr(svname, '-');
+        char *dash = strrchr(svname, '-');
         if (!dash)
                 return(-EINVAL);
 
+        /* ignore trailing "-mdc" */
+        if (!strcmp(LUSTRE_MDC_NAME, dash + 1)) {
+                dash--;
+                for (; dash > svname && *dash != '-'; dash--);
+                if (dash == svname)
+                        return(-EINVAL);
+        }
+
         if (strncmp(dash + 1, "MDT", 3) == 0)
                 rc = LDD_F_SV_TYPE_MDT;
         else if (strncmp(dash + 1, "OST", 3) == 0)
@@ -1714,8 +1744,18 @@ static void lmd_print(struct lustre_mount_data *lmd)
                 PRINT_CMD(PRINT_MASK, "profile: %s\n", lmd->lmd_profile);
         PRINT_CMD(PRINT_MASK, "device:  %s\n", lmd->lmd_dev);
         PRINT_CMD(PRINT_MASK, "flags:   %x\n", lmd->lmd_flags);
+
         if (lmd->lmd_opts)
                 PRINT_CMD(PRINT_MASK, "options: %s\n", lmd->lmd_opts);
+
+        if (lmd->lmd_recovery_time_soft)
+                PRINT_CMD(PRINT_MASK, "recovery time soft: %d\n",
+                          lmd->lmd_recovery_time_soft);
+
+        if (lmd->lmd_recovery_time_hard)
+                PRINT_CMD(PRINT_MASK, "recovery time hard: %d\n",
+                          lmd->lmd_recovery_time_hard);
+
         for (i = 0; i < lmd->lmd_exclude_count; i++) {
                 PRINT_CMD(PRINT_MASK, "exclude %d:  OST%04x\n", i,
                           lmd->lmd_exclude[i]);
@@ -1833,6 +1873,9 @@ static int lmd_parse(char *options, struct lustre_mount_data *lmd)
         s1 = options;
         while (*s1) {
                 int clear = 0;
+                int time_min = 2 * (CONNECTION_SWITCH_MAX +
+                               2 * INITIAL_CONNECT_TIMEOUT);
+
                 /* Skip whitespace and extra commas */
                 while (*s1 == ' ' || *s1 == ',')
                         s1++;
@@ -1845,6 +1888,14 @@ static int lmd_parse(char *options, struct lustre_mount_data *lmd)
                 if (strncmp(s1, "abort_recov", 11) == 0) {
                         lmd->lmd_flags |= LMD_FLG_ABORT_RECOV;
                         clear++;
+                } else if (strncmp(s1, "recovery_time_soft=", 19) == 0) {
+                        lmd->lmd_recovery_time_soft = max_t(int,
+                                simple_strtoul(s1 + 19, NULL, 10), time_min);
+                        clear++;
+                } else if (strncmp(s1, "recovery_time_hard=", 19) == 0) {
+                        lmd->lmd_recovery_time_hard = max_t(int,
+                                simple_strtoul(s1 + 19, NULL, 10), time_min);
+                        clear++;
                 } else if (strncmp(s1, "nosvc", 5) == 0) {
                         lmd->lmd_flags |= LMD_FLG_NOSVC;
                         clear++;
@@ -1867,8 +1918,11 @@ static int lmd_parse(char *options, struct lustre_mount_data *lmd)
                            must be the last one. */
                         *s1 = '\0';
                         break;
+                } else if (strncmp(s1, "loop=", 5) == 0) {
+                        clear++;
                 }
 
+
                 /* Find next opt */
                 s2 = strchr(s1, ',');
                 if (s2 == NULL) {
@@ -1990,7 +2044,7 @@ out:
                 CERROR("Unable to mount %s (%d)\n",
                        s2lsi(sb) ? lmd->lmd_dev : "", rc);
         } else {
-                CDEBUG(D_SUPER, "Mount %s complete\n", 
+                CDEBUG(D_SUPER, "Mount %s complete\n",
                        lmd->lmd_dev);
         }
         return rc;
@@ -2049,6 +2103,9 @@ struct file_system_type lustre_fs_type = {
         .get_sb       = lustre_get_sb,
         .kill_sb      = lustre_kill_super,
         .fs_flags     = FS_BINARY_MOUNTDATA | FS_REQUIRES_DEV |
+#ifdef FS_HAS_FIEMAP
+                        FS_HAS_FIEMAP |
+#endif
                         LL_RENAME_DOES_D_MOVE,
 };
 
@@ -2067,6 +2124,7 @@ EXPORT_SYMBOL(lustre_register_kill_super_cb);
 EXPORT_SYMBOL(lustre_common_put_super);
 EXPORT_SYMBOL(lustre_process_log);
 EXPORT_SYMBOL(lustre_end_log);
+EXPORT_SYMBOL(server_find_mount_locked);
 EXPORT_SYMBOL(server_get_mount);
 EXPORT_SYMBOL(server_put_mount);
 EXPORT_SYMBOL(server_register_target);
