@@ -169,9 +169,9 @@ echo_free_object (struct ec_object *eco)
         struct echo_client_obd *ec = &obd->u.echo_client;
 
         LASSERT (eco->eco_refcount == 0);
-        if (!eco->eco_lsm) 
+        if (!eco->eco_lsm)
                 CERROR("No object %s\n", obd->obd_name);
-        else 
+        else
                 obd_free_memmd(ec->ec_exp, &eco->eco_lsm);
         OBD_FREE (eco, sizeof (*eco));
 }
@@ -196,7 +196,7 @@ static int echo_create_object(struct obd_device *obd, int on_target,
 
         if (ulsm != NULL) {
                 eco = echo_allocate_object (obd);
-                if (eco == NULL) 
+                if (eco == NULL)
                         return (-ENOMEM);
 
                 lsm = eco->eco_lsm;
@@ -227,7 +227,7 @@ static int echo_create_object(struct obd_device *obd, int on_target,
                 }
         } else {
                 OBD_ALLOC(eco, sizeof(*eco));
-                if (!eco) 
+                if (!eco)
                         return (-ENOMEM);
                 eco->eco_device = obd;
                 lsm = NULL;
@@ -281,7 +281,7 @@ static int echo_create_object(struct obd_device *obd, int on_target,
 
  failed:
         echo_free_object (eco);
-        if (rc) 
+        if (rc)
                 CERROR("%s: err %d on create\n", obd->obd_name, rc);
         return (rc);
 }
@@ -447,9 +447,9 @@ echo_get_stripe_off_id (struct lov_stripe_md *lsm, obd_off *offp, obd_id *idp)
         *offp = offset * stripe_size + woffset % stripe_size;
 }
 
-static void 
-echo_client_page_debug_setup(struct lov_stripe_md *lsm, 
-                             cfs_page_t *page, int rw, obd_id id, 
+static void
+echo_client_page_debug_setup(struct lov_stripe_md *lsm,
+                             cfs_page_t *page, int rw, obd_id id,
                              obd_off offset, obd_off count)
 {
         char    *addr;
@@ -471,7 +471,7 @@ echo_client_page_debug_setup(struct lov_stripe_md *lsm,
                         stripe_off = 0xdeadbeef00c0ffeeULL;
                         stripe_id = 0xdeadbeef00c0ffeeULL;
                 }
-                block_debug_setup(addr + delta, OBD_ECHO_BLOCK_SIZE, 
+                block_debug_setup(addr + delta, OBD_ECHO_BLOCK_SIZE,
                                   stripe_off, stripe_id);
         }
 
@@ -479,8 +479,8 @@ echo_client_page_debug_setup(struct lov_stripe_md *lsm,
 }
 
 static int
-echo_client_page_debug_check(struct lov_stripe_md *lsm, 
-                             cfs_page_t *page, obd_id id, 
+echo_client_page_debug_check(struct lov_stripe_md *lsm,
+                             cfs_page_t *page, obd_id id,
                              obd_off offset, obd_off count)
 {
         obd_off stripe_off;
@@ -500,8 +500,8 @@ echo_client_page_debug_check(struct lov_stripe_md *lsm,
                 stripe_id = id;
                 echo_get_stripe_off_id (lsm, &stripe_off, &stripe_id);
 
-                rc2 = block_debug_check("test_brw", 
-                                        addr + delta, OBD_ECHO_BLOCK_SIZE, 
+                rc2 = block_debug_check("test_brw",
+                                        addr + delta, OBD_ECHO_BLOCK_SIZE,
                                         stripe_off, stripe_id);
                 if (rc2 != 0) {
                         CERROR ("Error in echo object "LPX64"\n", id);
@@ -520,6 +520,7 @@ static int echo_client_kbrw(struct obd_device *obd, int rw, struct obdo *oa,
         struct echo_client_obd *ec = &obd->u.echo_client;
         struct obd_info         oinfo = { { { 0 } } };
         obd_count               npages;
+        struct ptlrpc_request_set *set = NULL;
         struct brw_page        *pga;
         struct brw_page        *pgp;
         obd_off                 off;
@@ -542,6 +543,10 @@ static int echo_client_kbrw(struct obd_device *obd, int rw, struct obdo *oa,
             (count & (~CFS_PAGE_MASK)) != 0)
                 return (-EINVAL);
 
+        set =  ptlrpc_prep_set();
+        if (set == NULL)
+                RETURN(-ENOMEM);
+
         /* XXX think again with misaligned I/O */
         npages = count >> CFS_PAGE_SHIFT;
 
@@ -571,8 +576,24 @@ static int echo_client_kbrw(struct obd_device *obd, int rw, struct obdo *oa,
 
         oinfo.oi_oa = oa;
         oinfo.oi_md = lsm;
-        rc = obd_brw(rw, ec->ec_exp, &oinfo, npages, pga, oti);
 
+        /* OST/filter device don't support o_brw_async ops, turn to o_brw ops */
+        if (ec->ec_exp && ec->ec_exp->exp_obd &&
+            OBT(ec->ec_exp->exp_obd) && OBP(ec->ec_exp->exp_obd, brw_async)) {
+                rc = obd_brw_async(rw, ec->ec_exp, &oinfo, npages, pga, oti,
+                                   set, 0);
+                if (rc == 0) {
+                        rc = ptlrpc_set_wait(set);
+                        if (rc)
+                                CERROR("error from callback: rc = %d\n", rc);
+                }
+        } else {
+                rc = obd_brw(rw, ec->ec_exp, &oinfo, npages, pga, oti);
+        }
+        if (rc)
+                CDEBUG_LIMIT(rc == -ENOSPC ? D_INODE : D_ERROR,
+                             "error from obd_brw_async: rc = %d\n", rc);
+        ptlrpc_set_destroy(set);
  out:
         if (rc != 0 || rw != OBD_BRW_READ)
                 verify = 0;
@@ -649,7 +670,7 @@ static void ec_ap_fill_obdo(void *data, int cmd, struct obdo *oa)
 {
         struct echo_async_page *eap = EAP_FROM_COOKIE(data);
 
-        memcpy(oa, &eap->eap_eas->eas_oa, sizeof(*oa));
+        lustre_set_wire_obdo(oa, &eap->eap_eas->eas_oa);
 }
 
 static int ec_ap_completion(void *data, int cmd, struct obdo *oa, int rc)
@@ -663,7 +684,7 @@ static int ec_ap_completion(void *data, int cmd, struct obdo *oa, int rc)
             eas->eas_oa.o_id != ECHO_PERSISTENT_OBJID &&
             (eas->eas_oa.o_valid & OBD_MD_FLFLAGS) != 0 &&
             (eas->eas_oa.o_flags & OBD_FL_DEBUG_CHECK) != 0)
-                echo_client_page_debug_check(eas->eas_lsm, eap->eap_page, 
+                echo_client_page_debug_check(eas->eas_lsm, eap->eap_page,
                                              eas->eas_oa.o_id, eap->eap_off,
                                              CFS_PAGE_SIZE);
 
@@ -691,7 +712,7 @@ static int echo_client_async_page(struct obd_export *exp, int rw,
 {
         obd_count npages, i;
         struct echo_async_page *eap;
-        struct echo_async_state eas;
+        struct echo_async_state *eas;
         int rc = 0;
         struct echo_async_page **aps = NULL;
 
@@ -718,19 +739,23 @@ static int echo_client_async_page(struct obd_export *exp, int rw,
         /* XXX think again with misaligned I/O */
         npages = batching >> CFS_PAGE_SHIFT;
 
-        memcpy(&eas.eas_oa, oa, sizeof(*oa));
-        eas.eas_next_offset = offset;
-        eas.eas_end_offset = offset + count;
-        spin_lock_init(&eas.eas_lock);
-        cfs_waitq_init(&eas.eas_waitq);
-        eas.eas_in_flight = 0;
-        eas.eas_rc = 0;
-        eas.eas_lsm = lsm;
-        CFS_INIT_LIST_HEAD(&eas.eas_avail);
+        OBD_ALLOC_PTR(eas);
+        if (NULL == eas)
+                return(-ENOMEM);
+
+        memcpy(&eas->eas_oa, oa, sizeof(*oa));
+        eas->eas_next_offset = offset;
+        eas->eas_end_offset = offset + count;
+        spin_lock_init(&eas->eas_lock);
+        cfs_waitq_init(&eas->eas_waitq);
+        eas->eas_in_flight = 0;
+        eas->eas_rc = 0;
+        eas->eas_lsm = lsm;
+        CFS_INIT_LIST_HEAD(&eas->eas_avail);
 
         OBD_ALLOC(aps, npages * sizeof aps[0]);
         if (aps == NULL)
-                return (-ENOMEM);
+                GOTO(free_eas, rc = -ENOMEM);
 
         /* prepare the group of pages that we're going to be keeping
          * in flight */
@@ -748,56 +773,57 @@ static int echo_client_async_page(struct obd_export *exp, int rw,
 
                 eap->eap_magic = EAP_MAGIC;
                 eap->eap_page = page;
-                eap->eap_eas = &eas;
-                list_add_tail(&eap->eap_item, &eas.eas_avail);
+                eap->eap_eas = eas;
+                list_add_tail(&eap->eap_item, &eas->eas_avail);
                 aps[i] = eap;
         }
 
         /* first we spin queueing io and being woken by its completion */
-        spin_lock(&eas.eas_lock);
+        spin_lock(&eas->eas_lock);
         for(;;) {
                 int rc;
 
                 /* sleep until we have a page to send */
-                spin_unlock(&eas.eas_lock);
-                rc = wait_event_interruptible(eas.eas_waitq, 
-                                              eas_should_wake(&eas));
-                spin_lock(&eas.eas_lock);
-                if (rc && !eas.eas_rc)
-                        eas.eas_rc = rc;
-                if (eas.eas_rc)
+                spin_unlock(&eas->eas_lock);
+                rc = wait_event_interruptible(eas->eas_waitq,
+                                              eas_should_wake(eas));
+                spin_lock(&eas->eas_lock);
+                if (rc && !eas->eas_rc)
+                        eas->eas_rc = rc;
+                if (eas->eas_rc)
                         break;
-                if (list_empty(&eas.eas_avail))
+                if (list_empty(&eas->eas_avail))
                         continue;
-                eap = list_entry(eas.eas_avail.next, struct echo_async_page,
+                eap = list_entry(eas->eas_avail.next, struct echo_async_page,
                                  eap_item);
                 list_del(&eap->eap_item);
-                spin_unlock(&eas.eas_lock);
+                spin_unlock(&eas->eas_lock);
 
                 /* unbind the eap from its old page offset */
                 if (eap->eap_cookie != NULL) {
-                        obd_teardown_async_page(exp, lsm, NULL, 
+                        obd_teardown_async_page(exp, lsm, NULL,
                                                 eap->eap_cookie);
                         eap->eap_cookie = NULL;
                 }
 
-                eas.eas_next_offset += CFS_PAGE_SIZE;
-                eap->eap_off = eas.eas_next_offset;
+                eas->eas_next_offset += CFS_PAGE_SIZE;
+                eap->eap_off = eas->eas_next_offset;
 
                 rc = obd_prep_async_page(exp, lsm, NULL, eap->eap_page,
                                          eap->eap_off, &ec_async_page_ops,
-                                         eap, &eap->eap_cookie, 1, NULL);
+                                         eap, &eap->eap_cookie,
+                                         OBD_PAGE_NO_CACHE, NULL);
                 if (rc) {
-                        spin_lock(&eas.eas_lock);
-                        eas.eas_rc = rc;
+                        spin_lock(&eas->eas_lock);
+                        eas->eas_rc = rc;
                         break;
                 }
 
                 if (oa->o_id != ECHO_PERSISTENT_OBJID &&
                     (oa->o_valid & OBD_MD_FLFLAGS) != 0 &&
                     (oa->o_flags & OBD_FL_DEBUG_CHECK) != 0)
-                        echo_client_page_debug_setup(lsm, eap->eap_page, rw, 
-                                                     oa->o_id, 
+                        echo_client_page_debug_setup(lsm, eap->eap_page, rw,
+                                                     oa->o_id,
                                                      eap->eap_off, CFS_PAGE_SIZE);
 
                 /* always asserts urgent, which isn't quite right */
@@ -805,26 +831,26 @@ static int echo_client_async_page(struct obd_export *exp, int rw,
                                         rw, 0, CFS_PAGE_SIZE, 0,
                                         ASYNC_READY | ASYNC_URGENT |
                                         ASYNC_COUNT_STABLE);
-                spin_lock(&eas.eas_lock);
-                if (rc && !eas.eas_rc) {
-                        eas.eas_rc = rc;
+                spin_lock(&eas->eas_lock);
+                if (rc && !eas->eas_rc) {
+                        eas->eas_rc = rc;
                         break;
                 }
-                eas.eas_in_flight++;
-                if (eas.eas_next_offset == eas.eas_end_offset)
+                eas->eas_in_flight++;
+                if (eas->eas_next_offset == eas->eas_end_offset)
                         break;
-        } 
+        }
 
         /* still hold the eas_lock here.. */
 
         /* now we just spin waiting for all the rpcs to complete */
-        while(eas.eas_in_flight) {
-                spin_unlock(&eas.eas_lock);
-                wait_event_interruptible(eas.eas_waitq, 
-                                         eas.eas_in_flight == 0);
-                spin_lock(&eas.eas_lock);
+        while(eas->eas_in_flight) {
+                spin_unlock(&eas->eas_lock);
+                wait_event_interruptible(eas->eas_waitq,
+                                         eas->eas_in_flight == 0);
+                spin_lock(&eas->eas_lock);
         }
-        spin_unlock(&eas.eas_lock);
+        spin_unlock(&eas->eas_lock);
 
 out:
         if (aps != NULL) {
@@ -841,13 +867,15 @@ out:
                 }
                 OBD_FREE(aps, npages * sizeof aps[0]);
         }
+free_eas:
+        OBD_FREE_PTR(eas);
 
         RETURN(rc);
 }
 
 static int echo_client_prep_commit(struct obd_export *exp, int rw,
                                    struct obdo *oa, struct lov_stripe_md *lsm,
-                                   obd_off offset, obd_size count,  
+                                   obd_off offset, obd_size count,
                                    obd_size batch, struct obd_trans_info *oti)
 {
         struct obd_ioobj ioo;
@@ -876,6 +904,8 @@ static int echo_client_prep_commit(struct obd_export *exp, int rw,
         off = offset;
 
         for(; tot_pages; tot_pages -= npages) {
+                int lpages;
+
                 if (tot_pages < npages)
                         npages = tot_pages;
 
@@ -887,11 +917,13 @@ static int echo_client_prep_commit(struct obd_export *exp, int rw,
                 ioo.ioo_bufcnt = npages;
                 oti->oti_transno = 0;
 
-                ret = obd_preprw(rw, exp, oa, 1, &ioo, npages, rnb, lnb, oti);
+                lpages = npages;
+                ret = obd_preprw(rw, exp, oa, 1, &ioo, rnb, &lpages, lnb, oti);
                 if (ret != 0)
                         GOTO(out, ret);
+                LASSERT(lpages == npages);
 
-                for (i = 0; i < npages; i++) {
+                for (i = 0; i < lpages; i++) {
                         cfs_page_t *page = lnb[i].page;
 
                         /* read past eof? */
@@ -915,7 +947,7 @@ static int echo_client_prep_commit(struct obd_export *exp, int rw,
                                                              rnb[i].len);
                 }
 
-                ret = obd_commitrw(rw, exp, oa, 1, &ioo, npages, lnb, oti, ret);
+                ret = obd_commitrw(rw, exp, oa, 1,&ioo,rnb,npages,lnb,oti,ret);
                 if (ret != 0)
                         GOTO(out, ret);
         }
@@ -1115,7 +1147,7 @@ echo_client_cancel(struct obd_export *exp, struct obdo *oa)
                 return (-ENOENT);
 
         rc = obd_cancel(ec->ec_exp, ecl->ecl_object->eco_lsm, ecl->ecl_mode,
-                        &ecl->ecl_lock_handle);
+                        &ecl->ecl_lock_handle, 0, 0);
 
         echo_put_object (ecl->ecl_object);
         OBD_FREE (ecl, sizeof (*ecl));
@@ -1256,7 +1288,7 @@ echo_client_iocontrol(unsigned int cmd, struct obd_export *exp,
  out:
 
         /* XXX this should be in a helper also called by target_send_reply */
-        for (ack_lock = dummy_oti.oti_ack_locks, i = 0; i < 4; 
+        for (ack_lock = dummy_oti.oti_ack_locks, i = 0; i < 4;
              i++, ack_lock++) {
                 if (!ack_lock->mode)
                         break;
@@ -1296,25 +1328,54 @@ echo_client_setup(struct obd_device *obddev, obd_count len, void *buf)
         CFS_INIT_LIST_HEAD (&ec->ec_objects);
         ec->ec_unique = 0;
 
+        ec->ec_exp = lustre_hash_lookup(tgt->obd_uuid_hash, &echo_uuid);
+        if (ec->ec_exp)
+                RETURN(0);
+
         OBD_ALLOC(ocd, sizeof(*ocd));
         if (ocd == NULL) {
                 CERROR("Can't alloc ocd connecting to %s\n",
                        lustre_cfg_string(lcfg, 1));
                 return -ENOMEM;
         }
-
         ocd->ocd_connect_flags = OBD_CONNECT_VERSION | OBD_CONNECT_REQPORTAL;
         ocd->ocd_version = LUSTRE_VERSION_CODE;
 
-        rc = obd_connect(&conn, tgt, &echo_uuid, ocd, &ec->ec_exp);
+        if ((strncmp(tgt->obd_type->typ_name, LUSTRE_OSC_NAME,
+                     strlen(LUSTRE_OSC_NAME)) == 0) ||
+            (strncmp(tgt->obd_type->typ_name, LUSTRE_LOV_NAME,
+                     strlen(LUSTRE_LOV_NAME)) == 0)) {
+                rc = obd_connect(&conn, tgt, &echo_uuid, ocd, &ec->ec_exp);
+        } else {
+                rc = obd_connect(&conn, tgt, &echo_uuid, ocd, NULL);
+                if (rc == 0) {
+                        ec->ec_exp = class_conn2export(&conn);
+
+                        /* Turn off pinger because it connects to tgt obd directly */
+                        spin_lock(&tgt->obd_dev_lock);
+                        list_del_init(&ec->ec_exp->exp_obd_chain_timed);
+                        spin_unlock(&tgt->obd_dev_lock);
+                }
+        }
 
         OBD_FREE(ocd, sizeof(*ocd));
 
-        if (rc != 0) {
+        if (rc == -EALREADY && (strncmp(tgt->obd_type->typ_name,LUSTRE_OSC_NAME,
+                                        strlen(LUSTRE_OSC_NAME)) == 0)) {
+                /* OSC obd forbid reconnect already connected import,
+                 * so we hack creating another export here */
+                down_write(&tgt->u.cli.cl_sem);
+                rc = class_connect(&conn, tgt, &echo_uuid);
+                if (rc == 0) {
+                        ++tgt->u.cli.cl_conn_count;
+                        ec->ec_exp = class_conn2export(&conn);
+                }
+                up_write(&tgt->u.cli.cl_sem);
+        }
+
+        if (rc != 0)
                 CERROR("fail to connect to device %s\n",
                        lustre_cfg_string(lcfg, 1));
-                return (rc);
-        }
 
         RETURN(rc);
 }
@@ -1389,7 +1450,7 @@ static int echo_client_disconnect(struct obd_export *exp)
                 list_del (&ecl->ecl_exp_chain);
 
                 rc = obd_cancel(ec->ec_exp, ecl->ecl_object->eco_lsm,
-                                 ecl->ecl_mode, &ecl->ecl_lock_handle);
+                                 ecl->ecl_mode, &ecl->ecl_lock_handle, 0, 0);
 
                 CDEBUG (D_INFO, "Cancel lock on object "LPX64" on disconnect "
                         "(%d)\n", ecl->ecl_object->eco_id, rc);
