@@ -100,14 +100,17 @@ int cache_add_extent(struct lustre_cache *cache, struct ldlm_res_id *res,
                 if (!lock)
                         RETURN(-ENOLCK);
 
-                LASSERTF(lock->l_policy_data.l_extent.start <=
-                         extent->oap_obj_off &&
-                         extent->oap_obj_off + CFS_PAGE_SIZE - 1 <=
-                         lock->l_policy_data.l_extent.end,
-                         "Got wrong lock [" LPU64 "," LPU64 "] for page with "
-                         "offset " LPU64 "\n",
-                         lock->l_policy_data.l_extent.start,
-                         lock->l_policy_data.l_extent.end, extent->oap_obj_off);
+                if(lock->l_policy_data.l_extent.start > extent->oap_obj_off ||
+                   extent->oap_obj_off + CFS_PAGE_SIZE - 1 >
+                   lock->l_policy_data.l_extent.end) {
+                         CDEBUG(D_CACHE, "Got wrong lock [" LPU64 "," LPU64 "] "
+                                         "for page with offset " LPU64 "\n",
+                                         lock->l_policy_data.l_extent.start,
+                                         lock->l_policy_data.l_extent.end,
+                                         extent->oap_obj_off);
+                         LDLM_LOCK_PUT(lock);
+                         RETURN(-ENOLCK);
+                }
         } else {
                 /* Real extent width calculation here once we have real
                  * extents
@@ -138,6 +141,16 @@ int cache_add_extent(struct lustre_cache *cache, struct ldlm_res_id *res,
                                "under us\n");
                         RETURN(-ENOLCK);
                 }
+               
+                /* XXX Note! if the caller passed a unused lock handle,
+                 * it expects us to return the lockh of the lock we matched,
+                 * reference(LCK_PR) of the lock is increased here to assure
+                 * its validity, and the caller should decrease the reference
+                 * when it isn't used any more. */ 
+                if (lockh && !lustre_handle_is_used(lockh)) {
+                        ldlm_lock_addref(&tmplockh, LCK_PR);
+                        lustre_handle_copy(lockh, &tmplockh);
+                }
         }
 
         spin_lock(&lock->l_extents_list_lock);
