@@ -88,14 +88,19 @@ struct osc_cache_waiter {
 #define OSCC_FLAG_SYNC_IN_PROGRESS   0x08 /* only allow one thread to sync */
 #define OSCC_FLAG_LOW                0x10
 #define OSCC_FLAG_EXITING            0x20
+#define OSCC_FLAG_DEGRADED           0x40
+#define OSCC_FLAG_RDONLY             0x80
 
 int osc_precreate(struct obd_export *exp);
 int osc_create(struct obd_export *exp, struct obdo *oa,
-	       struct lov_stripe_md **ea, struct obd_trans_info *oti);
+               struct lov_stripe_md **ea, struct obd_trans_info *oti);
+int osc_create_async(struct obd_export *exp, struct obd_info *oinfo,
+                     struct lov_stripe_md **ea, struct obd_trans_info *oti);
 int osc_real_create(struct obd_export *exp, struct obdo *oa,
-	       struct lov_stripe_md **ea, struct obd_trans_info *oti);
+                    struct lov_stripe_md **ea, struct obd_trans_info *oti);
 void oscc_init(struct obd_device *obd);
 void osc_wake_cache_waiters(struct client_obd *cli);
+int osc_shrink_grant_to_target(struct client_obd *cli, long target);
 
 #ifdef LPROCFS
 int lproc_osc_attach_seqstat(struct obd_device *dev);
@@ -121,9 +126,14 @@ static inline int osc_recoverable_error(int rc)
 /* return 1 if osc should be resend request */
 static inline int osc_should_resend(int resend, struct client_obd *cli)
 {
-        return atomic_read(&cli->cl_resends) ? 
-                atomic_read(&cli->cl_resends) > resend : 1; 
+        return atomic_read(&cli->cl_resends) ?
+                atomic_read(&cli->cl_resends) > resend : 1;
 }
 
+static inline int osc_exp_is_2_0_server(struct obd_export *exp) {
+       LASSERT(exp);
+       return !!(exp->exp_connect_flags & OBD_CONNECT_FID);
+}
 
+#define OSC_FILE2MEM_OFF(fileoff,pshift) ((fileoff) + (pshift))
 #endif /* OSC_INTERNAL_H */
