@@ -46,10 +46,52 @@
 
 #include <obd_support.h>
 #include <obd_class.h>
+#include <lustre_net.h>
 #include "ptlrpc_internal.h"
 
+#ifdef __KERNEL__
+/* What time pinger is supposed to wake up next.
+ * pd_next_ping is the equivalent for liblustre. */
+cfs_time_t pinger_next_wake;
+#endif
+
 struct semaphore pinger_sem;
 static struct list_head pinger_imports = CFS_LIST_HEAD_INIT(pinger_imports);
+static struct list_head timeout_list = CFS_LIST_HEAD_INIT(timeout_list);
+
+struct ptlrpc_request *
+ptlrpc_prep_ping(struct obd_import *imp)
+{
+        struct ptlrpc_request *req;
+
+        req = ptlrpc_prep_req(imp, LUSTRE_OBD_VERSION,
+                              OBD_PING, 1, NULL, NULL);
+        if (req) {
+                ptlrpc_req_set_repsize(req, 1, NULL);
+                req->rq_no_resend = req->rq_no_delay = 1;
+        }
+        return req;
+}
+
+int ptlrpc_obd_ping(struct obd_device *obd)
+{
+        int rc;
+        struct ptlrpc_request *req;
+        ENTRY;
+
+        req = ptlrpc_prep_ping(obd->u.cli.cl_import);
+        if (req == NULL)
+                RETURN(-ENOMEM);
+
+        req->rq_send_state = LUSTRE_IMP_FULL;
+
+        rc = ptlrpc_queue_wait(req);
+
+        ptlrpc_req_finished(req);
+
+        RETURN(rc);
+}
+EXPORT_SYMBOL(ptlrpc_obd_ping);
 
 int ptlrpc_ping(struct obd_import *imp)
 {
@@ -57,14 +99,25 @@ int ptlrpc_ping(struct obd_import *imp)
         int rc = 0;
         ENTRY;
 
-        req = ptlrpc_prep_req(imp, LUSTRE_OBD_VERSION, OBD_PING, 
-                              1, NULL, NULL);
+        req = ptlrpc_prep_ping(imp);
         if (req) {
                 DEBUG_REQ(D_INFO, req, "pinging %s->%s",
                           imp->imp_obd->obd_uuid.uuid,
                           obd2cli_tgt(imp->imp_obd));
-                req->rq_no_resend = req->rq_no_delay = 1;
-                ptlrpc_req_set_repsize(req, 1, NULL);
+
+                /* To quickly detect server failure ping timeouts must be
+                 * kept small.  Therefore we must override/ignore the server
+                 * rpc completion estimate which may be very large since
+                 * it includes non-ping service times.  The right long term
+                 * fix will be to add a per-server (not per-service) thread
+                 * in order to reduce the number of pings in the system in
+                 * general (see bug 12471). */
+                if (!AT_OFF) {
+                        req->rq_timeout = PING_SVC_TIMEOUT +
+                                          at_get(&imp->imp_at.iat_net_latency);
+                        lustre_msg_set_timeout(req->rq_reqmsg, req->rq_timeout);
+                }
+
                 ptlrpcd_add_req(req);
         } else {
                 CERROR("OOM trying to ping %s->%s\n",
@@ -75,24 +128,79 @@ int ptlrpc_ping(struct obd_import *imp)
 
         RETURN(rc);
 }
+EXPORT_SYMBOL(ptlrpc_ping);
 
-void ptlrpc_update_next_ping(struct obd_import *imp)
+static void ptlrpc_update_next_ping(struct obd_import *imp, int soon)
 {
 #ifdef ENABLE_PINGER
-        int time = PING_INTERVAL;
-        if (imp->imp_state == LUSTRE_IMP_DISCON) {
-                int dtime = max_t(int, CONNECTION_SWITCH_MIN,
-                                  AT_OFF ? 0 :
-                                  at_get(&imp->imp_at.iat_net_latency));
-                time = min(time, dtime);
+        cfs_time_t delay, dtime, ctime = cfs_time_current();
+
+        if (imp->imp_state == LUSTRE_IMP_DISCON ||
+            imp->imp_state == LUSTRE_IMP_CONNECTING) {
+                /* In the disconnected case aggressively reconnect, for
+                 * this request the AT service timeout will be set to
+                 * INITIAL_CONNECT_TIMEOUT.  To ensure the request times
+                 * out before we send another we add one extra second. */
+                dtime = cfs_time_seconds(max_t(int, CONNECTION_SWITCH_MIN,
+                                AT_OFF ? 0 : INITIAL_CONNECT_TIMEOUT + 1 +
+                                at_get(&imp->imp_at.iat_net_latency)));
+        } else {
+                /* In the common case we want to cluster the pings at
+                 * at regular intervals to minimize system noise. */
+                delay = cfs_time_seconds(soon ? PING_INTERVAL_SHORT :
+                                         PING_INTERVAL);
+                dtime = delay - (ctime % delay);
         }
-        imp->imp_next_ping = cfs_time_shift(time);
+
+        dtime = cfs_time_add(ctime, dtime);
+
+        if (soon && cfs_time_after(imp->imp_next_ping, ctime) &&
+            cfs_time_after(dtime, imp->imp_next_ping)) {
+                /* if the next ping is due to be sent before the
+                 * new deadline, don't delay it */
+                 return;
+        }
+
+        /* May harmlessly race with ptlrpc_update_next_ping() */
+        imp->imp_next_ping = dtime;
+
+#ifdef __KERNEL__
+        if (pinger_next_wake != 0 && cfs_time_after(pinger_next_wake, dtime))
+                /* pinger is supposed to sleep until after the new ping
+                 * deadline, wake it up to take into account our update.
+                 * no needed for liblustre which updates pd_next_ping. */
+                ptlrpc_pinger_wake_up();
+#endif
+
+        CDEBUG(D_INFO, "Setting %s next ping to "CFS_TIME_T" ("CFS_TIME_T")\n",
+               obd2cli_tgt(imp->imp_obd), imp->imp_next_ping, dtime);
+
 #endif /* ENABLE_PINGER */
 }
 
-void ptlrpc_ping_import_soon(struct obd_import *imp)
+static inline int imp_is_deactive(struct obd_import *imp)
 {
-        imp->imp_next_ping = cfs_time_current();
+        return (imp->imp_deactive ||
+                OBD_FAIL_CHECK(OBD_FAIL_PTLRPC_IMP_DEACTIVE));
+}
+
+cfs_duration_t pinger_check_timeout(cfs_time_t time)
+{
+        struct timeout_item *item;
+        cfs_time_t timeout = PING_INTERVAL;
+
+        /* The timeout list is a increase order sorted list */
+        mutex_down(&pinger_sem);
+        list_for_each_entry(item, &timeout_list, ti_chain) {
+                int ti_timeout = item->ti_timeout;
+                if (timeout > ti_timeout)
+                         timeout = ti_timeout;
+                break;
+        }
+        mutex_up(&pinger_sem);
+
+        return cfs_time_sub(cfs_time_add(time, cfs_time_seconds(timeout)),
+                                         cfs_time_current());
 }
 
 #ifdef __KERNEL__
@@ -112,10 +220,17 @@ static int ptlrpc_pinger_main(void *arg)
         while (1) {
                 cfs_time_t this_ping = cfs_time_current();
                 struct l_wait_info lwi;
-                cfs_duration_t time_to_next_ping;
+                cfs_duration_t time_to_next_wake;
+                cfs_time_t time_of_next_wake;
+                struct timeout_item *item;
                 struct list_head *iter;
 
+                time_of_next_wake = cfs_time_shift(PING_INTERVAL);
+
                 mutex_down(&pinger_sem);
+                list_for_each_entry(item, &timeout_list, ti_chain) {
+                        item->ti_cb(item, item->ti_cb_data);
+                }
                 list_for_each(iter, &pinger_imports) {
                         struct obd_import *imp =
                                 list_entry(iter, struct obd_import,
@@ -133,20 +248,21 @@ static int ptlrpc_pinger_main(void *arg)
                                ptlrpc_import_state_name(level), level,
                                force, imp->imp_deactive, imp->imp_pingable);
 
+                        /* Include any ping which misses the deadline by up to
+                         * 1/10 of a second.  The pings are designed to clump
+                         * and this helps ensure the entire batch gets sent
+                         * promptly, which minimizes system noise from pings */
+
                         if (force ||
-                            /* if the next ping is within, say, 5 jiffies from
-                               now, go ahead and ping. See note below. */
-                            cfs_time_aftereq(this_ping, 
-                                             imp->imp_next_ping - 5 * CFS_TICK)) {
+                            cfs_time_aftereq(this_ping, imp->imp_next_ping -
+                                             (cfs_time_seconds(1) + 9) / 10)) {
                                 if (level == LUSTRE_IMP_DISCON &&
-                                    !imp->imp_deactive) {
-                                        /* wait at least a timeout before
-                                           trying recovery again. */
-                                        imp->imp_next_ping = cfs_time_shift(obd_timeout);
+                                    !imp_is_deactive(imp)) {
+                                        ptlrpc_update_next_ping(imp, 0);
                                         ptlrpc_initiate_recovery(imp);
                                 } else if (level != LUSTRE_IMP_FULL ||
                                          imp->imp_obd->obd_no_recov ||
-                                         imp->imp_deactive) {
+                                         imp_is_deactive(imp)) {
                                         CDEBUG(D_HA, "not pinging %s "
                                                "(in recovery: %s or recovery "
                                                "disabled: %u/%u)\n",
@@ -155,7 +271,12 @@ static int ptlrpc_pinger_main(void *arg)
                                                imp->imp_deactive,
                                                imp->imp_obd->obd_no_recov);
                                 } else if (imp->imp_pingable || force) {
-                                        ptlrpc_ping(imp);
+                                                ptlrpc_ping(imp);
+                                                /* ptlrpc_pinger_sending_on_import()
+                                                 * will asynch update imp_next_ping
+                                                 * so it must not be used below to
+                                                 * calculate minimum wait time. */
+                                                continue;
                                 }
                         } else {
                                 if (!imp->imp_pingable)
@@ -167,33 +288,24 @@ static int ptlrpc_pinger_main(void *arg)
                                        imp->imp_next_ping, this_ping);
                         }
 
-                        /* obd_timeout might have changed */
-                        if (cfs_time_after(imp->imp_next_ping,
-                                           cfs_time_add(this_ping, 
-                                                        cfs_time_seconds(PING_INTERVAL))))
-                                ptlrpc_update_next_ping(imp);
+                        /* Wait time until next ping, or until we stopped. */
+                        if (cfs_time_before(imp->imp_next_ping,
+                                            time_of_next_wake))
+                                time_of_next_wake = imp->imp_next_ping;
                 }
+                pinger_next_wake = time_of_next_wake;
                 mutex_up(&pinger_sem);
-                /* update memory usage info */
                 obd_update_maxusage();
 
-                /* Wait until the next ping time, or until we're stopped. */
-                time_to_next_ping = cfs_time_sub(cfs_time_add(this_ping, 
-                                                              cfs_time_seconds(PING_INTERVAL)), 
-                                                 cfs_time_current());
+                time_to_next_wake = max_t(cfs_duration_t,
+                                          cfs_time_seconds(1),
+                                          cfs_time_sub(time_of_next_wake,
+                                                       cfs_time_current()));
+                CDEBUG(D_INFO, "next ping in "CFS_DURATION_T" ("CFS_TIME_T")\n",
+                               time_to_next_wake, time_of_next_wake);
 
-                /* The ping sent by ptlrpc_send_rpc may get sent out
-                   say .01 second after this.
-                   ptlrpc_pinger_eending_on_import will then set the
-                   next ping time to next_ping + .01 sec, which means
-                   we will SKIP the next ping at next_ping, and the
-                   ping will get sent 2 timeouts from now!  Beware. */
-                CDEBUG(D_INFO, "next ping in "CFS_DURATION_T" ("CFS_TIME_T")\n", 
-                               time_to_next_ping, 
-                               cfs_time_add(this_ping, cfs_time_seconds(PING_INTERVAL)));
-                if (time_to_next_ping > 0) {
-                        lwi = LWI_TIMEOUT(max_t(cfs_duration_t, time_to_next_ping, cfs_time_seconds(1)),
-                                          NULL, NULL);
+                if (time_to_next_wake > 0) {
+                        lwi = LWI_TIMEOUT(time_to_next_wake, NULL, NULL);
                         l_wait_event(thread->t_ctl_waitq,
                                      thread->t_flags & (SVC_STOPPING|SVC_EVENT),
                                      &lwi);
@@ -239,7 +351,7 @@ int ptlrpc_start_pinger(void)
         d.thread = pinger_thread;
 
         /* CLONE_VM and CLONE_FILES just avoid a needless copy, because we
-         * just drop the VM and FILES in ptlrpc_daemonize() right away. */
+         * just drop the VM and FILES in cfs_daemonize_ctxt() right away. */
         rc = cfs_kernel_thread(ptlrpc_pinger_main, &d, CLONE_VM | CLONE_FILES);
         if (rc < 0) {
                 CERROR("cannot start thread: %d\n", rc);
@@ -253,6 +365,8 @@ int ptlrpc_start_pinger(void)
         RETURN(0);
 }
 
+int ptlrpc_pinger_remove_timeouts(void);
+
 int ptlrpc_stop_pinger(void)
 {
         struct l_wait_info lwi = { 0 };
@@ -264,6 +378,8 @@ int ptlrpc_stop_pinger(void)
 
         if (pinger_thread == NULL)
                 RETURN(-EALREADY);
+
+        ptlrpc_pinger_remove_timeouts();
         mutex_down(&pinger_sem);
         pinger_thread->t_flags = SVC_STOPPING;
         cfs_waitq_signal(&pinger_thread->t_ctl_waitq);
@@ -279,7 +395,12 @@ int ptlrpc_stop_pinger(void)
 
 void ptlrpc_pinger_sending_on_import(struct obd_import *imp)
 {
-        ptlrpc_update_next_ping(imp);
+        ptlrpc_update_next_ping(imp, 0);
+}
+
+void ptlrpc_pinger_commit_expected(struct obd_import *imp)
+{
+        ptlrpc_update_next_ping(imp, 1);
 }
 
 int ptlrpc_pinger_add_import(struct obd_import *imp)
@@ -294,7 +415,7 @@ int ptlrpc_pinger_add_import(struct obd_import *imp)
         /* if we add to pinger we want recovery on this import */
         imp->imp_obd->obd_no_recov = 0;
 
-        ptlrpc_update_next_ping(imp);
+        ptlrpc_update_next_ping(imp, 0);
         /* XXX sort, blah blah */
         list_add_tail(&imp->imp_pinger_chain, &pinger_imports);
         class_import_get(imp);
@@ -322,6 +443,121 @@ int ptlrpc_pinger_del_import(struct obd_import *imp)
         RETURN(0);
 }
 
+/**
+ * Register a timeout callback to the pinger list, and the callback will
+ * be called when timeout happens.
+ */
+struct timeout_item* ptlrpc_new_timeout(int time, enum timeout_event event,
+                                        timeout_cb_t cb, void *data)
+{
+        struct timeout_item *ti;
+
+        OBD_ALLOC_PTR(ti);
+        if (!ti)
+                return(NULL);
+
+        CFS_INIT_LIST_HEAD(&ti->ti_obd_list);
+        CFS_INIT_LIST_HEAD(&ti->ti_chain);
+        ti->ti_timeout = time;
+        ti->ti_event = event;
+        ti->ti_cb = cb;
+        ti->ti_cb_data = data;
+
+        return ti;
+}
+
+/**
+ * Register timeout event on the the pinger thread.
+ * Note: the timeout list is an sorted list with increased timeout value.
+ */
+static struct timeout_item*
+ptlrpc_pinger_register_timeout(int time, enum timeout_event event,
+                               timeout_cb_t cb, void *data)
+{
+        struct timeout_item *item, *tmp;
+
+        LASSERT_SEM_LOCKED(&pinger_sem);
+
+        list_for_each_entry(item, &timeout_list, ti_chain)
+                if (item->ti_event == event)
+                        goto out;
+
+        item = ptlrpc_new_timeout(time, event, cb, data);
+        if (item) {
+                list_for_each_entry_reverse(tmp, &timeout_list, ti_chain) {
+                        if (tmp->ti_timeout < time) {
+                                list_add(&item->ti_chain, &tmp->ti_chain);
+                                goto out;
+                        }
+                }
+                list_add(&item->ti_chain, &timeout_list);
+        }
+out:
+        return item;
+}
+
+/* Add a client_obd to the timeout event list, when timeout(@time)
+ * happens, the callback(@cb) will be called.
+ */
+int ptlrpc_add_timeout_client(int time, enum timeout_event event,
+                              timeout_cb_t cb, void *data,
+                              struct list_head *obd_list)
+{
+        struct timeout_item *ti;
+
+        mutex_down(&pinger_sem);
+        ti = ptlrpc_pinger_register_timeout(time, event, cb, data);
+        if (!ti) {
+                mutex_up(&pinger_sem);
+                return (-EINVAL);
+        }
+        list_add(obd_list, &ti->ti_obd_list);
+        mutex_up(&pinger_sem);
+        return 0;
+}
+
+int ptlrpc_del_timeout_client(struct list_head *obd_list,
+                              enum timeout_event event)
+{
+        struct timeout_item *ti = NULL, *item;
+
+        if (list_empty(obd_list))
+                return 0;
+        mutex_down(&pinger_sem);
+        list_del_init(obd_list);
+        /**
+         * If there are no obd attached to the timeout event
+         * list, remove this timeout event from the pinger
+         */
+        list_for_each_entry(item, &timeout_list, ti_chain) {
+                if (item->ti_event == event) {
+                        ti = item;
+                        break;
+                }
+        }
+        LASSERTF(ti != NULL, "ti is NULL ! \n");
+        if (list_empty(&ti->ti_obd_list)) {
+                list_del(&ti->ti_chain);
+                OBD_FREE_PTR(ti);
+        }
+        mutex_up(&pinger_sem);
+        return 0;
+}
+
+int ptlrpc_pinger_remove_timeouts(void)
+{
+        struct timeout_item *item, *tmp;
+
+        mutex_down(&pinger_sem);
+        list_for_each_entry_safe(item, tmp, &timeout_list, ti_chain) {
+                LASSERT(list_empty(&item->ti_obd_list));
+                list_del(&item->ti_chain);
+                OBD_FREE_PTR(item);
+        }
+        mutex_up(&pinger_sem);
+        return 0;
+}
+
 void ptlrpc_pinger_wake_up()
 {
 #ifdef ENABLE_PINGER
@@ -337,23 +573,25 @@ void ptlrpc_pinger_wake_up()
 static int               pet_refcount = 0;
 static int               pet_state;
 static wait_queue_head_t pet_waitq;
-static struct obd_export *pet_exp = NULL;
+CFS_LIST_HEAD(pet_list);
 static spinlock_t        pet_lock = SPIN_LOCK_UNLOCKED;
 
 int ping_evictor_wake(struct obd_export *exp)
 {
+        struct obd_device *obd;
+
         spin_lock(&pet_lock);
-        if (pet_exp || (pet_state != PET_READY)) {
+        if (pet_state != PET_READY) {
                 /* eventually the new obd will call here again. */
                 spin_unlock(&pet_lock);
                 return 1;
         }
 
-        /* We have to make sure the obd isn't destroyed between now and when
-         * the ping evictor runs.  We'll take a reference here, and drop it
-         * when we finish in the evictor.  We don't really care about this
-         * export in particular; we just need one to keep the obd alive. */
-        pet_exp = class_export_get(exp);
+        obd = class_exp2obd(exp);
+        if (list_empty(&obd->obd_evict_list)) {
+                class_incref(obd);
+                list_add(&obd->obd_evict_list, &pet_list);
+        }
         spin_unlock(&pet_lock);
 
         wake_up(&pet_waitq);
@@ -368,24 +606,31 @@ static int ping_evictor_main(void *arg)
         time_t expire_time;
         ENTRY;
 
-        ptlrpc_daemonize("ll_evictor");
+        cfs_daemonize_ctxt("ll_evictor");
 
         CDEBUG(D_HA, "Starting Ping Evictor\n");
-        pet_exp = NULL;
         pet_state = PET_READY;
         while (1) {
-                l_wait_event(pet_waitq, pet_exp ||
+                l_wait_event(pet_waitq, (!list_empty(&pet_list)) ||
                              (pet_state == PET_TERMINATE), &lwi);
-                if (pet_state == PET_TERMINATE)
+
+                /* loop until all obd's will be removed */
+                if ((pet_state == PET_TERMINATE) && list_empty(&pet_list))
                         break;
 
                 /* we only get here if pet_exp != NULL, and the end of this
                  * loop is the only place which sets it NULL again, so lock
                  * is not strictly necessary. */
                 spin_lock(&pet_lock);
-                obd = pet_exp->exp_obd;
+                obd = list_entry(pet_list.next, struct obd_device,
+                                 obd_evict_list);
                 spin_unlock(&pet_lock);
 
+                /* bug 18948: ensure recovery is aborted in a timely fashion */
+                if (target_recovery_check_and_stop(obd) ||
+                    obd->obd_recovering /* no evictor during recovery */)
+                       GOTO(skip, 0);
+
                 expire_time = cfs_time_current_sec() - PING_EVICT_TIMEOUT;
 
                 CDEBUG(D_HA, "evicting all exports of obd %s older than %ld\n",
@@ -421,13 +666,14 @@ static int ping_evictor_main(void *arg)
                         }
                 }
                 spin_unlock(&obd->obd_dev_lock);
-
-                class_export_put(pet_exp);
-
+skip:
                 spin_lock(&pet_lock);
-                pet_exp = NULL;
+                list_del_init(&obd->obd_evict_list);
                 spin_unlock(&pet_lock);
+
+                class_decref(obd);
         }
+
         CDEBUG(D_HA, "Exiting Ping Evictor\n");
 
         RETURN(0);
@@ -517,8 +763,9 @@ static int pinger_check_rpcs(void *arg)
                         list_entry(iter, struct obd_import, imp_pinger_chain);
                 int generation, level;
 
-                if (cfs_time_aftereq(pd->pd_this_ping, 
-                                     imp->imp_next_ping - 5 * CFS_TICK)) {
+                /* Include any ping within 1/10 of a second of the deadline */
+                if (cfs_time_aftereq(pd->pd_this_ping, imp->imp_next_ping -
+                                     (cfs_time_seconds(1) + 9) / 10)) {
                         /* Add a ping. */
                         spin_lock(&imp->imp_lock);
                         generation = imp->imp_generation;
@@ -554,7 +801,7 @@ static int pinger_check_rpcs(void *arg)
         mutex_up(&pinger_sem);
 
         /* Might be empty, that's OK. */
-        if (set->set_remaining == 0)
+        if (atomic_read(&set->set_remaining) == 0)
                 CDEBUG(D_RPCTRACE, "nothing to ping\n");
 
         list_for_each(iter, &set->set_requests) {
@@ -603,7 +850,7 @@ do_check_set:
                         atomic_dec(&imp->imp_inflight);
                 }
                 spin_unlock(&imp->imp_lock);
-                set->set_remaining--;
+                atomic_dec(&set->set_remaining);
         }
         mutex_up(&pinger_sem);
 
@@ -647,7 +894,7 @@ void ptlrpc_pinger_sending_on_import(struct obd_import *imp)
 {
 #ifdef ENABLE_PINGER
         mutex_down(&pinger_sem);
-        ptlrpc_update_next_ping(imp);
+        ptlrpc_update_next_ping(imp, 0);
         if (pinger_args.pd_set == NULL &&
             time_before(imp->imp_next_ping, pinger_args.pd_next_ping)) {
                 CDEBUG(D_HA, "set next ping to "CFS_TIME_T"(cur "CFS_TIME_T")\n",
@@ -658,6 +905,34 @@ void ptlrpc_pinger_sending_on_import(struct obd_import *imp)
 #endif
 }
 
+void ptlrpc_pinger_commit_expected(struct obd_import *imp)
+{
+#ifdef ENABLE_PINGER
+        mutex_down(&pinger_sem);
+        ptlrpc_update_next_ping(imp, 1);
+        if (pinger_args.pd_set == NULL &&
+            time_before(imp->imp_next_ping, pinger_args.pd_next_ping)) {
+                CDEBUG(D_HA, "set next ping to "CFS_TIME_T"(cur "CFS_TIME_T")\n",
+                        imp->imp_next_ping, cfs_time_current());
+                pinger_args.pd_next_ping = imp->imp_next_ping;
+        }
+        mutex_up(&pinger_sem);
+#endif
+}
+
+int ptlrpc_add_timeout_client(int time, enum timeout_event event,
+                              timeout_cb_t cb, void *data,
+                              struct list_head *obd_list)
+{
+        return 0;
+}
+
+int ptlrpc_del_timeout_client(struct list_head *obd_list,
+                              enum timeout_event event)
+{
+        return 0;
+}
+
 int ptlrpc_pinger_add_import(struct obd_import *imp)
 {
         ENTRY;
@@ -701,11 +976,13 @@ void ptlrpc_pinger_wake_up()
                 CDEBUG(D_RPCTRACE, "checking import %s->%s\n",
                        imp->imp_obd->obd_uuid.uuid, obd2cli_tgt(imp->imp_obd));
 #ifdef ENABLE_LIBLUSTRE_RECOVERY
-                if (imp->imp_state == LUSTRE_IMP_DISCON && !imp->imp_deactive)
+                if (imp->imp_state == LUSTRE_IMP_DISCON &&
+                    !imp_is_deactive(imp))
 #else
                 /*XXX only recover for the initial connection */
                 if (!lustre_handle_is_used(&imp->imp_remote_handle) &&
-                    imp->imp_state == LUSTRE_IMP_DISCON && !imp->imp_deactive)
+                    imp->imp_state == LUSTRE_IMP_DISCON &&
+                    !imp_is_deactive(imp))
 #endif
                         ptlrpc_initiate_recovery(imp);
                 else if (imp->imp_state != LUSTRE_IMP_FULL)
@@ -713,7 +990,7 @@ void ptlrpc_pinger_wake_up()
                                      "state %d, deactive %d\n",
                                      imp->imp_obd->obd_uuid.uuid,
                                      obd2cli_tgt(imp->imp_obd), imp->imp_state,
-                                     imp->imp_deactive);
+                                     imp_is_deactive(imp));
         }
 #endif
         EXIT;
