@@ -203,13 +203,12 @@ int lprocfs_quota_rd_type(char *page, char **start, off_t off, int count,
 
         /* Collect the needed information */
         oq_type = obd->u.obt.obt_qctxt.lqc_flags;
-        oq_version = obt->obt_qfmt;
         if (is_mds) {
-                rc = mds_quota_get_version(obd, &aq_version);
+                rc = mds_quota_get_version(obd, &aq_version, &oq_version);
                 if (rc)
                         return -EPROTO;
-                /* Here we can also assert that aq_type == oq_type
-                 * except for quota startup/shutdown states     */
+        } else {
+                oq_version = obt->obt_qfmt;
         }
 
         /* Transform the collected data into a user-readable string */
@@ -236,22 +235,17 @@ int lprocfs_quota_rd_type(char *page, char **start, off_t off, int count,
 }
 EXPORT_SYMBOL(lprocfs_quota_rd_type);
 
-static int auto_quota_on(struct obd_device *obd, int type,
-                         struct super_block *sb, int is_master)
+/*
+ * generic_quota_on is very lazy and tolerant about current quota settings
+ * @global means to turn on quotas on each OST additionally to local quotas;
+ * should not be called from filter_quota_ctl on MDS nodes (as it starts
+ * admin quotas on MDS nodes).
+ */
+int generic_quota_on(struct obd_device *obd, struct obd_quotactl *oqctl, int global)
 {
-        struct obd_quotactl *oqctl;
+        struct obd_device_target *obt = &obd->u.obt;
         struct lvfs_run_ctxt saved;
-        int rc = 0, id;
-        struct obd_device_target *obt;
-        ENTRY;
-
-        LASSERT(type == USRQUOTA || type == GRPQUOTA || type == UGQUOTA);
-
-        obt = &obd->u.obt;
-
-        OBD_ALLOC_PTR(oqctl);
-        if (!oqctl)
-                RETURN(-ENOMEM);
+        int id, is_master, rc = 0, local; /* means we need a local quotaon */
 
         if (!atomic_dec_and_test(&obt->obt_quotachecking)) {
                 CDEBUG(D_INFO, "other people are doing quotacheck\n");
@@ -259,45 +253,74 @@ static int auto_quota_on(struct obd_device *obd, int type,
                 RETURN(-EBUSY);
         }
 
-        id = UGQUOTA2LQC(type);
-        /* quota already turned on */
-        if ((obt->obt_qctxt.lqc_flags & id) == id) {
-                rc = 0;
-                goto out;
-        }
+        id = UGQUOTA2LQC(oqctl->qc_type);
+        local = (obt->obt_qctxt.lqc_flags & id) != id;
 
-        oqctl->qc_type = type;
         oqctl->qc_cmd = Q_QUOTAON;
         oqctl->qc_id = obt->obt_qfmt;
 
-        push_ctxt(&saved, &obd->obd_lvfs_ctxt, NULL);
+        is_master = !strcmp(obd->obd_type->typ_name, LUSTRE_MDS_NAME);
         if (is_master) {
-                struct mds_obd *mds = &obd->u.mds;
-
-                down(&mds->mds_qonoff_sem);
-                /* turn on cluster wide quota */
-                rc = mds_admin_quota_on(obd, oqctl);
-                if (rc)
-                        CDEBUG(rc == -ENOENT ? D_QUOTA : D_ERROR,
-                               "auto-enable admin quota failed. rc=%d\n", rc);
-                up(&mds->mds_qonoff_sem);
-
-        }
-        if (!rc) {
-                /* turn on local quota */
-                rc = fsfilt_quotactl(obd, sb, oqctl);
-                if (rc)
-                        CDEBUG(rc == -ENOENT ? D_QUOTA : D_ERROR,
-                               "auto-enable local quota failed. rc=%d\n", rc);
-                else
-                        obt->obt_qctxt.lqc_flags |= UGQUOTA2LQC(type);
+                down(&obd->u.mds.mds_qonoff_sem);
+                if (local) {
+                        push_ctxt(&saved, &obd->obd_lvfs_ctxt, NULL);
+                        /* turn on cluster wide quota */
+                        rc = mds_admin_quota_on(obd, oqctl);
+                        pop_ctxt(&saved, &obd->obd_lvfs_ctxt, NULL);
+                        if (rc && rc != -ENOENT)
+                                CERROR("%s: %s admin quotaon failed. rc=%d\n",
+                                       obd->obd_name, global? "global":"local",
+                                       rc);
+                }
         }
 
-        pop_ctxt(&saved, &obd->obd_lvfs_ctxt, NULL);
+        if (rc == 0) {
+                if (local) {
+                        push_ctxt(&saved, &obd->obd_lvfs_ctxt, NULL);
+                        /* turn on local quota */
+                        rc = fsfilt_quotactl(obd, obt->obt_sb, oqctl);
+                        if (rc) {
+                                if (rc != -ENOENT)
+                                        CERROR("%s: %s quotaon failed with"
+                                               " rc=%d\n", obd->obd_name,
+                                               global ? "global" : "local", rc);
+                        } else {
+                                obt->obt_qctxt.lqc_flags |= UGQUOTA2LQC(oqctl->qc_type);
+                                /* when quotaon, create lqs for every
+                                 * quota uid/gid b=18574 */
+                                build_lqs(obd);
+                        }
+                        pop_ctxt(&saved, &obd->obd_lvfs_ctxt, NULL);
+                }
+
+                if (rc == 0 && global && is_master)
+                        rc = obd_quotactl(obd->u.mds.mds_lov_exp, oqctl);
+        }
+
+        if (is_master)
+                up(&obd->u.mds.mds_qonoff_sem);
 
-out:
         atomic_inc(&obt->obt_quotachecking);
 
+        return rc;
+}
+
+static int auto_quota_on(struct obd_device *obd, int type)
+{
+        struct obd_quotactl *oqctl;
+        int rc;
+        ENTRY;
+
+        LASSERT(type == USRQUOTA || type == GRPQUOTA || type == UGQUOTA);
+
+        OBD_ALLOC_PTR(oqctl);
+        if (!oqctl)
+                RETURN(-ENOMEM);
+
+        oqctl->qc_type = type;
+
+        rc = generic_quota_on(obd, oqctl, 0);
+
         OBD_FREE_PTR(oqctl);
         RETURN(rc);
 }
@@ -320,13 +343,15 @@ static int filter_quota_set_version(struct obd_device *obd,
                 return -EBUSY;
         }
 
-        if (obt->obt_qctxt.lqc_flags & (LQC_USRQUOTA_FLAG | LQC_GRPQUOTA_FLAG)) {
-                atomic_inc(&obt->obt_quotachecking);
-                return -EBUSY;
+        /* do not complain of being busy if we actually have nothing to do */
+        if (obt->obt_qfmt != version) {
+                if (obt->obt_qctxt.lqc_flags&(LQC_USRQUOTA_FLAG|LQC_GRPQUOTA_FLAG)){
+                        atomic_inc(&obt->obt_quotachecking);
+                        return -EBUSY;
+                }
+                obt->obt_qfmt = version;
         }
 
-        obt->obt_qfmt = version;
-
         atomic_inc(&obt->obt_quotachecking);
 
         return 0;
@@ -391,19 +416,22 @@ int lprocfs_quota_wr_type(struct file *file, const char *buffer,
                                 return -EINVAL;
 #endif
                         if (is_mds) {
-                                rc = mds_quota_set_version(obd, s2av[idx]);
+                                rc = mds_quota_set_version(obd, s2av[idx],
+                                                           s2ov[idx]);
                                 if (rc) {
                                         CDEBUG(D_QUOTA, "failed to set admin "
                                                "quota to spec %c! %d\n",
                                                stype[i], rc);
                                         return rc;
                                 }
-                        }
-                        rc = filter_quota_set_version(obd, s2ov[idx]);
-                        if (rc) {
-                                CDEBUG(D_QUOTA, "failed to set operational quota"
-                                       " to spec %c! %d\n", stype[i], rc);
-                                return rc;
+                        } else {
+                                rc = filter_quota_set_version(obd, s2ov[idx]);
+                                if (rc) {
+                                        CDEBUG(D_QUOTA, "failed to set op"
+                                               " quota to spec %c! %d\n",
+                                               stype[i], rc);
+                                        return rc;
+                                }
                         }
                         break;
                 default  : /* just skip stray symbols like \n */
@@ -411,8 +439,18 @@ int lprocfs_quota_wr_type(struct file *file, const char *buffer,
                 }
         }
 
-        if (type != 0)
-                auto_quota_on(obd, type - 1, obt->obt_sb, is_mds);
+        if (type != 0) {
+                int rc = auto_quota_on(obd, type - 1);
+
+                if (rc == -ENOENT)
+                        CWARN("%s: quotaon failed because quota files don't "
+                              "exist, please run quotacheck firstly\n",
+                              obd->obd_name);
+                else if (rc == -EALREADY)
+                        CWARN("%s: quota is on already!\n", obd->obd_name);
+                else if (rc != 0)
+                        return rc;
+        }
 
         return count;
 }
@@ -679,8 +717,8 @@ int lquota_proc_setup(struct obd_device *obd, int is_master)
                                                lprocfs_quota_common_vars, obd);
         if (IS_ERR(qctxt->lqc_proc_dir)) {
                 rc = PTR_ERR(qctxt->lqc_proc_dir);
-                CERROR("error %d setting up lprocfs for %s\n", rc,
-                       obd->obd_name);
+                CERROR("%s: error %d setting up lprocfs\n",
+                       obd->obd_name, rc);
                 qctxt->lqc_proc_dir = NULL;
                 GOTO(out, rc);
         }
@@ -689,8 +727,8 @@ int lquota_proc_setup(struct obd_device *obd, int is_master)
                 rc = lprocfs_add_vars(qctxt->lqc_proc_dir,
                                       lprocfs_quota_master_vars, obd);
                 if (rc) {
-                        CERROR("error %d setting up lprocfs for %s"
-                               "(quota master)\n", rc, obd->obd_name);
+                        CERROR("%s: error %d setting up lprocfs for "
+                               "quota master\n", obd->obd_name, rc);
                         GOTO(out_free_proc, rc);
                 }
         }
@@ -762,7 +800,7 @@ int lquota_proc_cleanup(struct lustre_quota_ctxt *qctxt)
                 return -EINVAL;
 
         if (qctxt->lqc_stats != NULL)
-                 lprocfs_free_stats(&qctxt->lqc_stats);
+                lprocfs_free_stats(&qctxt->lqc_stats);
 
         lprocfs_remove(&qctxt->lqc_proc_dir);
         return 0;
