@@ -13,7 +13,12 @@ TMP=${TMP:-/tmp}
 
 MDSDEV=${MDSDEV:-$TMP/${FSNAME}-mdt}
 MDSSIZE=${MDSSIZE:-400000}
-MDSOPT=${MDSOPT:-"--mountfsoptions=acl"}
+MDSOPT=${MDSOPT:-"--mountfsoptions=errors=remount-ro,iopen_nopriv,user_xattr,acl"}
+
+mdsfailover_dev=${mdsfailover_dev:-$MDSDEV}
+
+MGSDEV=${MGSDEV:-$MDSDEV}
+MGSSIZE=${MGSSIZE:-$MDSSIZE}
 
 OSTCOUNT=${OSTCOUNT:-2}
 OSTDEVBASE=${OSTDEVBASE:-$TMP/${FSNAME}-ost}
@@ -36,6 +41,10 @@ SUBSYSTEM=${SUBSYSTEM:- 0xffb7e3ff}
 
 L_GETGROUPS=${L_GETGROUPS:-`do_facet mds which l_getgroups || echo`}
 
+ENABLE_QUOTA=${ENABLE_QUOTA:-""}
+QUOTA_TYPE=${QUOTA_TYPE:-"ug3"}
+QUOTA_USERS=${QUOTA_USERS:-"quota_usr quota_2usr sanityusr sanityusr1"}
+
 MKFSOPT=""
 [ "x$MDSJOURNALSIZE" != "x" ] &&
     MKFSOPT=$MKFSOPT" -J size=$MDSJOURNALSIZE"
@@ -51,7 +60,13 @@ MKFSOPT=""
     MDSOPT=$MDSOPT" --param lov.stripecount=$STRIPES_PER_OBJ"
 [ "x$L_GETGROUPS" != "x" ] &&
     MDSOPT=$MDSOPT" --param mdt.group_upcall=$L_GETGROUPS"
-MDS_MKFS_OPTS="--mgs --mdt --fsname=$FSNAME --device-size=$MDSSIZE --param sys.timeout=$TIMEOUT $MKFSOPT $MDSOPT $MDS_MKFS_OPTS"
+MDS_MKFS_OPTS="--mdt --fsname=$FSNAME --device-size=$MDSSIZE --param sys.timeout=$TIMEOUT $MKFSOPT $MDSOPT $MDS_MKFS_OPTS"
+if [[ $mds_HOST == $mgs_HOST ]] && [[ $MDSDEV == $MGSDEV ]]; then
+    MDS_MKFS_OPTS="--mgs $MDS_MKFS_OPTS"
+else
+    MDS_MKFS_OPTS="--mgsnode=$MGSNID $MDS_MKFS_OPTS"
+    mgs_MKFS_OPTS="--mgs "
+fi
 
 MKFSOPT=""
 [ "x$OSTJOURNALSIZE" != "x" ] &&
@@ -64,6 +79,7 @@ OST_MKFS_OPTS="--ost --fsname=$FSNAME --device-size=$OSTSIZE --mgsnode=$MGSNID -
 
 MDS_MOUNT_OPTS=${MDS_MOUNT_OPTS:-"-o loop"}
 OST_MOUNT_OPTS=${OST_MOUNT_OPTS:-"-o loop"}
+mgs_MOUNT_OPTS=${mgs_MOUNT_OPTS:-$MDS_MOUNT_OPTS}
 
 #client
 MOUNT=${MOUNT:-/mnt/${FSNAME}}
@@ -88,3 +104,8 @@ POWER_DOWN=${POWER_DOWN:-"powerman --off"}
 POWER_UP=${POWER_UP:-"powerman --on"}
 SLOW=${SLOW:-no}
 FAIL_ON_ERROR=${FAIL_ON_ERROR:-true}
+
+MPIRUN=$(which mpirun 2>/dev/null) || true
+MPI_USER=${MPI_USER:-mpiuser}
+SHARED_DIR_LOGS=${SHARED_DIR_LOGS:-""}
+
