@@ -93,18 +93,27 @@ int main(int argc, char *argv[])
 					argv[0], name, st.st_mode, mode);
 				exit(12);
 			}
+			if (i == S_IFCHR || i == S_IFBLK) {
+				if (st.st_rdev != 0x1234) {
+					fprintf(stderr, "%s: ERROR rdev %s: "
+						"%llu != 0x1234",
+						argv[0], name,
+						(unsigned long long)st.st_rdev);
+					exit(13);
+				}
+			}
 			rc = unlink(name);
 			if (rc < 0) {
 				fprintf(stderr, "%s: ERROR unlink %s: %s",
 					argv[0], name, strerror(errno));
-				exit(13);
+				exit(14);
 			}
 			break;
 		default:
 			if (rc == 0) {
 				fprintf(stderr, "%s: ERROR: %s created\n",
 					argv[0], name);
-				exit(14);
+				exit(15);
 			}
 		}
 	}
