@@ -42,6 +42,7 @@ remote_ost_nodsh && skip "remote OST with nodsh" && exit 0
 FAIL_CLIENTS=$(echo " $FAIL_CLIENTS " | sed -re "s/\s+$LIVE_CLIENT\s+/ /g")
 
 DIR=${DIR:-$MOUNT}
+TESTDIR=$DIR/d0.$(basename $0 .sh)
 
 #####
 # fail clients round robin
@@ -60,19 +61,6 @@ set_fail_client() {
     echo "fail $FAIL_CLIENT, next is $FAIL_NEXT"
 }
 
-shutdown_client() {
-    client=$1
-    if [ "$FAILURE_MODE" = HARD ]; then
-       $POWER_DOWN $client
-       while ping -w 3 -c 1 $client > /dev/null 2>&1; do 
-	   echo "waiting for node $client to fail"
-	   sleep 1
-       done  
-    elif [ "$FAILURE_MODE" = SOFT ]; then
-       zconf_umount $client $MOUNT -f
-    fi
-}
-
 fail_clients() {
     num=$1
 
@@ -124,29 +112,28 @@ client_touch() {
     file=$1
     for c in $LIVE_CLIENT $FAIL_CLIENTS;  do
 	if echo $DOWN_CLIENTS | grep -q $c; then continue; fi
-	$PDSH $c touch $MOUNT/${c}_$file || return 1
+	$PDSH $c touch $TESTDIR/${c}_$file || return 1
     done
 }
 
 client_rm() {
     file=$1
     for c in $LIVE_CLIENT $FAIL_CLIENTS;  do
-	$PDSH $c rm $MOUNT/${c}_$file
+	$PDSH $c rm $TESTDIR/${c}_$file
     done
 }
 
 client_mkdirs() {
     for c in $LIVE_CLIENT $FAIL_CLIENTS;  do
-	echo "$c mkdir $MOUNT/$c"
-	$PDSH $c "mkdir $MOUNT/$c"
-	$PDSH $c "ls -l $MOUNT/$c" 
+	echo "$c mkdir $TESTDIR/$c"
+	$PDSH $c "mkdir $TESTDIR/$c && ls -l $TESTDIR/$c"
     done
 }
 
 client_rmdirs() {
     for c in $LIVE_CLIENT $FAIL_CLIENTS;  do
-	echo "rmdir $MOUNT/$c"
-	$PDSH $LIVE_CLIENT "rmdir $MOUNT/$c"
+	echo "rmdir $TESTDIR/$c"
+	$PDSH $LIVE_CLIENT "rmdir $TESTDIR/$c"
     done
 }
 
@@ -157,6 +144,9 @@ clients_recover_osts() {
 
 check_and_setup_lustre
 
+rm -rf $TESTDIR
+mkdir -p $TESTDIR
+
 # 9 Different Failure Modes Combinations
 echo "Starting Test 17 at `date`"
 
@@ -187,7 +177,7 @@ test_2() {
     echo "Verify Lustre filesystem is up and running"
     [ -z "$(mounted_lustre_filesystems)" ] && error "Lustre is not running"
 
-    client_df
+    clients_up
 
     shutdown_facet mds
     reboot_facet mds
@@ -196,7 +186,7 @@ test_2() {
     change_active mds
     reboot_facet mds
 
-    client_df &
+    clients_up &
     DFPID=$!
     sleep 5
 
@@ -214,7 +204,7 @@ test_2() {
     wait $DFPID
     clients_recover_osts ost1
     echo "Verify reintegration"
-    client_df || return 1
+    clients_up || return 1
 
 }
 run_test 2 "Second Failure Mode: MDS/OST `date`"
@@ -226,28 +216,28 @@ test_3() {
     #Create files
     echo "Verify Lustre filesystem is up and running"
     [ -z "$(mounted_lustre_filesystems)" ] && error "Lustre is not running"
-    
+
     #MDS Portion
     facet_failover mds
     wait $DFPID || echo df failed: $?
     #Check FS
 
     echo "Test Lustre stability after MDS failover"
-    client_df
+    clients_up
 
     #CLIENT Portion
     echo "Failing 2 CLIENTS"
     fail_clients 2
-    
+
     #Check FS
     echo "Test Lustre stability after CLIENT failure"
-    client_df
-    
+    clients_up
+
     #Reintegration
     echo "Reintegrating CLIENTS"
     reintegrate_clients || return 1
 
-    client_df || return 3
+    clients_up || return 3
 }
 run_test 3  "Thirdb Failure Mode: MDS/CLIENT `date`"
 ###################################################
@@ -258,10 +248,10 @@ test_4() {
 
     #OST Portion
     shutdown_facet ost1
- 
+
     #Check FS
     echo "Test Lustre stability after OST failure"
-    client_df &
+    clients_up &
     DFPIDA=$!
     sleep 5
 
@@ -273,7 +263,7 @@ test_4() {
     change_active mds
     reboot_facet mds
 
-    client_df &
+    clients_up &
     DFPIDB=$!
     sleep 5
 
@@ -282,23 +272,23 @@ test_4() {
     reboot_facet ost1
     wait_for ost1
     start_ost 1
-    
+
     wait_for mds
     start mds $MDSDEV $MDS_MOUNT_OPTS
     #Check FS
-    
+
     wait $DFPIDA
     wait $DFPIDB
     clients_recover_osts ost1
     echo "Test Lustre stability after MDS failover"
-    client_df || return 1
+    clients_up || return 1
 }
 run_test 4 "Fourth Failure Mode: OST/MDS `date`"
 ###################################################
 
 ############### Fifth Failure Mode ###############
 test_5() {
-    [ $OSTCOUNT -lt 2 ] && skip "$OSTCOUNT < 2, not enough OSTs" && return 0
+    [ $OSTCOUNT -lt 2 ] && skip_env "$OSTCOUNT < 2, not enough OSTs" && return 0
 
     echo "Fifth Failure Mode: OST/OST `date`"
 
@@ -306,25 +296,25 @@ test_5() {
     echo "Verify Lustre filesystem is up and running"
     [ -z "$(mounted_lustre_filesystems)" ] && error "Lustre is not running"
 
-    client_df
-    
+    clients_up
+
     #OST Portion
     shutdown_facet ost1
     reboot_facet ost1
-    
+
     #Check FS
     echo "Test Lustre stability after OST failure"
-    client_df &
+    clients_up &
     DFPIDA=$!
     sleep 5
-    
+
     #OST Portion
     shutdown_facet ost2
     reboot_facet ost2
 
     #Check FS
     echo "Test Lustre stability after OST failure"
-    client_df &
+    clients_up &
     DFPIDB=$!
     sleep 5
 
@@ -334,14 +324,14 @@ test_5() {
     start_ost 1
     wait_for ost2
     start_ost 2
-    
+
     clients_recover_osts ost1
     clients_recover_osts ost2
     sleep $TIMEOUT
 
     wait $DFPIDA
     wait $DFPIDB
-    client_df || return 2
+    clients_up || return 2
 }
 run_test 5 "Fifth Failure Mode: OST/OST `date`"
 ###################################################
@@ -354,16 +344,16 @@ test_6() {
     echo "Verify Lustre filesystem is up and running"
     [ -z "$(mounted_lustre_filesystems)" ] && error "Lustre is not running"
 
-    client_df
+    clients_up
     client_touch testfile || return 2
-	
+
     #OST Portion
     shutdown_facet ost1
     reboot_facet ost1
 
     #Check FS
     echo "Test Lustre stability after OST failure"
-    client_df &
+    clients_up &
     DFPIDA=$!
     echo DFPIDA=$DFPIDA
     sleep 5
@@ -371,28 +361,28 @@ test_6() {
     #CLIENT Portion
     echo "Failing CLIENTs"
     fail_clients
-    
+
     #Check FS
     echo "Test Lustre stability after CLIENTs failure"
-    client_df &
+    clients_up &
     DFPIDB=$!
     echo DFPIDB=$DFPIDB
     sleep 5
-    
+
     #Reintegration
     echo "Reintegrating OST/CLIENTs"
     wait_for ost1
     start_ost 1
     reintegrate_clients || return 1
-    sleep 5 
+    sleep 5
 
-    wait_remote_prog df $((TIMEOUT * 3 + 10)) 
+    wait_remote_prog df $((TIMEOUT * 3 + 10))
     wait $DFPIDA
     wait $DFPIDB
 
     echo "Verifying mount"
     [ -z "$(mounted_lustre_filesystems)" ] && return 3
-    client_df
+    clients_up
 }
 run_test 6 "Sixth Failure Mode: OST/CLIENT `date`"
 ###################################################
@@ -406,19 +396,19 @@ test_7() {
     echo "Verify Lustre filesystem is up and running"
     [ -z "$(mounted_lustre_filesystems)" ] && error "Lustre is not running"
 
-    client_df
+    clients_up
     client_touch testfile  || return 1
 
     #CLIENT Portion
     echo "Part 1: Failing CLIENT"
     fail_clients 2
-    
+
     #Check FS
     echo "Test Lustre stability after CLIENTs failure"
-    client_df
-    $PDSH $LIVE_CLIENT "ls -l $MOUNT"
-    $PDSH $LIVE_CLIENT "rm -f $MOUNT/*_testfile"
-    
+    clients_up
+    $PDSH $LIVE_CLIENT "ls -l $TESTDIR"
+    $PDSH $LIVE_CLIENT "rm -f $TESTDIR/*_testfile"
+
     #Sleep
     echo "Wait 1 minutes"
     sleep 60
@@ -427,7 +417,7 @@ test_7() {
     echo "Verify Lustre filesystem is up and running"
     [ -z "$(mounted_lustre_filesystems)" ] && return 2
 
-    client_df
+    clients_up
     client_rm testfile
 
     #MDS Portion
@@ -436,14 +426,14 @@ test_7() {
     #Check FS
     echo "Test Lustre stability after MDS failover"
     wait $DFPID || echo "df on down clients fails " || return 1
-    $PDSH $LIVE_CLIENT "ls -l $MOUNT"
-    $PDSH $LIVE_CLIENT "rm -f $MOUNT/*_testfile"
+    $PDSH $LIVE_CLIENT "ls -l $TESTDIR"
+    $PDSH $LIVE_CLIENT "rm -f $TESTDIR/*_testfile"
 
     #Reintegration
     echo "Reintegrating CLIENTs"
     reintegrate_clients || return 2
-    client_df
-    
+    clients_up
+
     #Sleep
     echo "wait 1 minutes"
     sleep 60
@@ -460,18 +450,18 @@ test_8() {
     echo "Verify Lustre filesystem is up and running"
     [ -z "$(mounted_lustre_filesystems)" ] && error "Lustre is not running"
 
-    client_df
+    clients_up
     client_touch testfile
-	
+
     #CLIENT Portion
     echo "Failing CLIENTs"
     fail_clients 2
 
     #Check FS
     echo "Test Lustre stability after CLIENTs failure"
-    client_df
-    $PDSH $LIVE_CLIENT "ls -l $MOUNT"
-    $PDSH $LIVE_CLIENT "rm -f $MOUNT/*_testfile"
+    clients_up
+    $PDSH $LIVE_CLIENT "ls -l $TESTDIR"
+    $PDSH $LIVE_CLIENT "rm -f $TESTDIR/*_testfile"
 
     #Sleep
     echo "Wait 1 minutes"
@@ -481,7 +471,7 @@ test_8() {
     echo "Verify Lustre filesystem is up and running"
     [ -z "$(mounted_lustre_filesystems)" ] && error "Lustre is not running"
 
-    client_df
+    clients_up
     client_touch testfile
 
 
@@ -491,20 +481,20 @@ test_8() {
 
     #Check FS
     echo "Test Lustre stability after OST failure"
-    client_df &
+    clients_up &
     DFPID=$!
     sleep 5
     #non-failout hangs forever here
-    #$PDSH $LIVE_CLIENT "ls -l $MOUNT"
-    #$PDSH $LIVE_CLIENT "rm -f $MOUNT/*_testfile"
-    
+    #$PDSH $LIVE_CLIENT "ls -l $TESTDIR"
+    #$PDSH $LIVE_CLIENT "rm -f $TESTDIR/*_testfile"
+
     #Reintegration
     echo "Reintegrating CLIENTs/OST"
     reintegrate_clients || return 3
     wait_for ost1
     start_ost 1
     wait $DFPID
-    client_df || return 1
+    clients_up || return 1
     client_touch testfile2 || return 2
 
     #Sleep
@@ -523,18 +513,18 @@ test_9() {
     echo "Verify Lustre filesystem is up and running"
     [ -z "$(mounted_lustre_filesystems)" ] && error "Lustre is not running"
 
-    client_df
+    clients_up
     client_touch testfile || return 1
-	
+
     #CLIENT Portion
     echo "Failing CLIENTs"
     fail_clients 2
 
     #Check FS
     echo "Test Lustre stability after CLIENTs failure"
-    client_df
-    $PDSH $LIVE_CLIENT "ls -l $MOUNT" || return 1
-    $PDSH $LIVE_CLIENT "rm -f $MOUNT/*_testfile" || return 2
+    clients_up
+    $PDSH $LIVE_CLIENT "ls -l $TESTDIR" || return 1
+    $PDSH $LIVE_CLIENT "rm -f $TESTDIR/*_testfile" || return 2
 
     #Sleep
     echo "Wait 1 minutes"
@@ -542,25 +532,25 @@ test_9() {
 
     #Create files
     echo "Verify Lustre filesystem is up and running"
-    $PDSH $LIVE_CLIENT "grep -e $DIR /proc/mounts" || return 3
-    $PDSH $LIVE_CLIENT df $MOUNT
+    $PDSH $LIVE_CLIENT "grep -e $MOUNT /proc/mounts" || return 3
+    client_up $LIVE_CLIENT
     client_touch testfile || return 4
 
     #CLIENT Portion
     echo "Failing CLIENTs"
     fail_clients 2
-    
+
     #Check FS
     echo "Test Lustre stability after CLIENTs failure"
-    client_df
-    $PDSH $LIVE_CLIENT "ls -l $MOUNT" || return 5
-    $PDSH $LIVE_CLIENT "rm -f $MOUNT/*_testfile" || return 6
+    clients_up
+    $PDSH $LIVE_CLIENT "ls -l $TESTDIR" || return 5
+    $PDSH $LIVE_CLIENT "rm -f $TESTDIR/*_testfile" || return 6
 
     #Reintegration
     echo "Reintegrating  CLIENTs/CLIENTs"
     reintegrate_clients || return 7
-    client_df
-    
+    clients_up
+
     #Sleep
     echo "Wait 1 minutes"
     sleep 60
