@@ -43,7 +43,7 @@ $LFS setstripe $TESTDIR -c -1
 get_stripe $TESTDIR
 
 # Make sure we start with a clean slate
-rm -f ${LOG} PI*
+rm -f ${LOG}
 
 if [ -n "$NOSINGLE" ]; then
     echo "NO Test for creates for a single client."
@@ -52,7 +52,6 @@ else
     mdsrate_cleanup $NUM_CLIENTS $MACHINEFILE $NUM_FILES $TESTDIR_SINGLE 'f%%d' --ignore
 
     log "===== $0 ### 1 NODE CREATE ###"
-    echo "Running creates on 1 node(s)."
 
     COMMAND="${MDSRATE} ${MDSRATE_DEBUG} --create --time ${TIME_PERIOD}
                 --nfiles ${NUM_FILES} --dir ${TESTDIR_SINGLE} --filefmt 'f%%d'"
@@ -60,21 +59,25 @@ else
     mpi_run -np 1 -machinefile ${MACHINEFILE} ${COMMAND} | tee ${LOG}
 
     if [ ${PIPESTATUS[0]} != 0 ]; then
-	[ -f $LOG ] && cat $LOG
-	error "mpirun ... mdsrate ... failed, aborting"
+	[ -f $LOG ] && sed -e "s/^/log: /" $LOG
+	error "mdsrate creates for a single client failed, aborting"
     fi
     
     log "===== $0 ### 1 NODE UNLINK ###"
-    echo "Running unlinks on 1 node(s)."
 
-    COMMAND="${MDSRATE} ${MDSRATE_DEBUG} --unlink --time ${TIME_PERIOD}
+    if [ -f "$LOG" ]; then
+        CREATED=$(awk '/total:/ { print $7 }' $LOG)
+        [ $CREATED -gt 0 ] && NUM_FILES=$CREATED
+    fi
+
+    COMMAND="${MDSRATE} ${MDSRATE_DEBUG} --unlink
                 --nfiles ${NUM_FILES} --dir ${TESTDIR_SINGLE} --filefmt 'f%%d'"
     echo "+ ${COMMAND}"
     mpi_run -np 1 -machinefile ${MACHINEFILE} ${COMMAND} | tee ${LOG}
  
     if [ ${PIPESTATUS[0]} != 0 ]; then
-	[ -f $LOG ] && cat $LOG
-	error "mpirun ... mdsrate ... failed, aborting"
+	[ -f $LOG ] && sed -e "s/^/log: /" $LOG
+	error "mdsrate unlink on a single client failed, aborting"
     fi
 fi
 
@@ -83,13 +86,13 @@ if [ $IFree -lt $NUM_FILES ]; then
     NUM_FILES=$IFree
 fi
 
+[ $NUM_CLIENTS -eq 1 ] && NOMULTI=yes
 if [ -n "$NOMULTI" ]; then
     echo "NO test for create on multiple nodes."
 else
     mdsrate_cleanup $NUM_CLIENTS $MACHINEFILE $NUM_FILES $TESTDIR_MULTI 'f%%d' --ignore
 
     log "===== $0 ### $NUM_CLIENTS NODES CREATE ###"
-    echo "Running creates on ${NUM_CLIENTS} node(s)."
 
     COMMAND="${MDSRATE} ${MDSRATE_DEBUG} --create --time ${TIME_PERIOD}
                 --nfiles $NUM_FILES --dir ${TESTDIR_MULTI} --filefmt 'f%%d'"
@@ -97,27 +100,28 @@ else
     mpi_run -np ${NUM_CLIENTS} -machinefile ${MACHINEFILE} ${COMMAND} | tee ${LOG}
 
     if [ ${PIPESTATUS[0]} != 0 ]; then
-	[ -f $LOG ] && cat $LOG
-	error "mpirun ... mdsrate ... failed, aborting"
+	[ -f $LOG ] && sed -e "s/^/log: /" $LOG
+	error "mdsrate create on multiple nodes failed, aborting"
     fi
 
-    echo "Running unlinks on ${NUM_CLIENTS} node(s)."
+    log "===== $0 ### $NUM_CLIENTS NODES UNLINK ###"
 
-    COMMAND="${MDSRATE} ${MDSRATE_DEBUG} --unlink --time ${TIME_PERIOD}
+    if [ -f "$LOG" ]; then
+        CREATED=$(awk '/total:/ { print $7 }' $LOG)
+        [ $CREATED -gt 0 ] && NUM_FILES=$CREATED
+    fi
+
+    COMMAND="${MDSRATE} ${MDSRATE_DEBUG} --unlink
                 --nfiles ${NUM_FILES} --dir ${TESTDIR_MULTI} --filefmt 'f%%d'"
     echo "+ ${COMMAND}"
     mpi_run -np ${NUM_CLIENTS} -machinefile ${MACHINEFILE} ${COMMAND} | tee ${LOG}
 
     if [ ${PIPESTATUS[0]} != 0 ]; then
-	[ -f $LOG ] && cat $LOG
-	error "mpirun ... mdsrate ... failed, aborting"
+	[ -f $LOG ] && sed -e "s/^/log: /" $LOG
+	error "mdsrate unlink on multiple nodes failed, aborting"
     fi
-
 fi
 
-equals_msg `basename $0`: test complete, cleaning up
-mdsrate_cleanup $NUM_CLIENTS $MACHINEFILE $NUM_FILES $TESTDIR_SINGLE 'f%%d' --ignore
-mdsrate_cleanup $NUM_CLIENTS $MACHINEFILE $NUM_FILES $TESTDIR_MULTI 'f%%d' --ignore
 rm -f $MACHINEFILE
 check_and_cleanup_lustre
 #rm -f $LOG
