@@ -2,8 +2,8 @@
 
 set -e
 
-# bug number:  10124 
-ALWAYS_EXCEPT="15c   $REPLAY_DUAL_EXCEPT"
+# bug number:  10124 19884
+ALWAYS_EXCEPT="15c 14b  $REPLAY_DUAL_EXCEPT"
 
 SAVE_PWD=$PWD
 PTLDEBUG=${PTLDEBUG:--1}
@@ -198,7 +198,7 @@ test_11() {
     #sleep for while, let both clients reconnect and timeout
     sleep $((TIMEOUT * 2))
     do_facet mds lctl set_param fail_loc=0
-    client_df
+    clients_up
     while [ -z "$(ls $MOUNT1/$tfile-[1-5] 2>/dev/null)" ]; do
 	sleep 5
 	echo -n "."
@@ -220,7 +220,7 @@ test_12() {
     do_facet mds lctl set_param fail_loc=0x80000302
     facet_failover mds
     do_facet mds lctl set_param fail_loc=0
-    df $MOUNT || { kill -USR1 $MULTIPID  && return 1; }
+    clients_up || { kill -USR1 $MULTIPID  && return 1; }
 
     ls $DIR/$tfile
     kill -USR1 $MULTIPID || return 3
@@ -245,7 +245,7 @@ test_13() {
     do_facet mds lctl set_param fail_loc=0x80000115
     facet_failover mds
     do_facet mds lctl set_param fail_loc=0
-    df $MOUNT || return 1
+    clients_up || return 1
 
     ls $DIR/$tfile
     $CHECKSTAT -t file $DIR/$tfile || return 2
@@ -255,7 +255,7 @@ test_13() {
 }
 run_test 13 "close resend timeout"
 
-test_14() {
+test_14a() {
     replay_barrier mds
     createmany -o $MOUNT1/$tfile- 25
     createmany -o $MOUNT2/$tfile-2- 1
@@ -264,16 +264,44 @@ test_14() {
 
     facet_failover mds
     # expect recovery to fail due to missing client 2
-    df $MOUNT && return 1
+    client_evicted || return 1
     sleep 1
 
-    # first 25 files should have been replayed 
+    # first 25 files should have been replayed
     unlinkmany $MOUNT1/$tfile- 25 || return 2
 
-    zconf_mount `hostname` $MOUNT2 || error "mount $MOUNT2 fail" 
+    zconf_mount `hostname` $MOUNT2 || error "mount $MOUNT2 fail"
     return 0
 }
-run_test 14 "timeouts waiting for lost client during replay"
+run_test 14a "timeouts waiting for lost client during replay"
+
+test_14b() {
+    BEFOREUSED=`df -P $DIR | tail -1 | awk '{ print $3 }'`
+    #lfs setstripe --index=0 --count=1 $MOUNT1
+    mkdir -p $MOUNT1/$tdir
+    #lfs setstripe --index=0 --count=1 $MOUNT1/$tdir
+    replay_barrier mds
+    createmany -o $MOUNT1/$tfile- 5
+    echo "data" > $MOUNT2/$tdir/$tfile-2
+    createmany -o $MOUNT1/$tfile-3- 5
+    umount $MOUNT2
+
+    fail mds
+
+    # first 25 files should have been replayed
+    unlinkmany $MOUNT1/$tfile- 5 || return 2
+    unlinkmany $MOUNT1/$tfile-3- 5 || return 3
+
+    zconf_mount `hostname` $MOUNT2 || error "mount $MOUNT2 fail"
+    # give ost time to process llogs
+    sleep 3
+    AFTERUSED=`df -P $DIR | tail -1 | awk '{ print $3 }'`
+    log "before $BEFOREUSED, after $AFTERUSED"
+    [ $AFTERUSED -ne $BEFOREUSED ] && \
+        error "after $AFTERUSED > before $BEFOREUSED" && return 4
+    return 0
+}
+run_test 14b "delete ost orphans if gap occured in objids due to VBR"
 
 test_15a() {	# was test_15
     replay_barrier mds
@@ -281,8 +309,7 @@ test_15a() {	# was test_15
     createmany -o $MOUNT2/$tfile-2- 1
     umount $MOUNT2
 
-    facet_failover mds
-    df $MOUNT || return 1
+    fail mds
 
     unlinkmany $MOUNT1/$tfile- 25 || return 2
     [ -e $MOUNT1/$tfile-2-0 ] && error "$tfile-2-0 exists"
@@ -295,14 +322,12 @@ run_test 15a "timeout waiting for lost client during replay, 1 client completes"
 test_15c() {
     replay_barrier mds
     for ((i = 0; i < 2000; i++)); do
-	echo "data" > "$MOUNT2/${tfile}-$i" || error "create ${tfile}-$i failed"
+        echo "data" > "$MOUNT2/${tfile}-$i" || error "create ${tfile}-$i failed"
     done
-    
-    umount $MOUNT2
-    facet_failover mds
 
-    df $MOUNT || return 1
-    
+    umount $MOUNT2
+    fail mds
+
     zconf_mount `hostname` $MOUNT2 || error "mount $MOUNT2 fail"
     return 0
 }
@@ -316,8 +341,7 @@ test_16() {
 
     facet_failover mds
     sleep $TIMEOUT
-    facet_failover mds
-    df $MOUNT || return 1
+    fail mds
 
     unlinkmany $MOUNT1/$tfile- 25 || return 2
 
@@ -339,8 +363,7 @@ test_17() {
 
     facet_failover ost1
     sleep $TIMEOUT
-    facet_failover ost1
-    df $MOUNT || return 1
+    fail ost1
 
     unlinkmany $MOUNT1/$tfile- 25 || return 2
 
@@ -391,29 +414,42 @@ test_20() { #16389
     touch $MOUNT1/a
     touch $MOUNT2/b
     umount $MOUNT2
-    facet_failover mds
-    df $MOUNT1 || return 1
+    fail mds
     rm $MOUNT1/a
     zconf_mount `hostname` $MOUNT2 || error "mount $MOUNT2 fail"
     TIER1=$((`date +%s` - BEFORE))
-    echo "First recovery time is $TIER1 sec"
     BEFORE=`date +%s`
     replay_barrier mds
     touch $MOUNT1/a
     touch $MOUNT2/b
     umount $MOUNT2
-    facet_failover mds
-    df $MOUNT1 || return 1
+    fail mds
     rm $MOUNT1/a
     zconf_mount `hostname` $MOUNT2 || error "mount $MOUNT2 fail"
     TIER2=$((`date +%s` - BEFORE))
-    echo "Second recovery time is $TIER2 sec"
     [ $TIER2 -ge $((TIER1 * 2)) ] && \
         error "recovery time is growing $TIER2 > $TIER1"
     return 0
 }
 run_test 20 "recovery time is not increasing"
 
+test_22() { #bug 18927
+    multiop_bg_pause $MOUNT1/$tfile O_c || return 1
+    pid1=$!
+    multiop_bg_pause $MOUNT2/$tfile O_c || return 2
+    pid2=$!
+    rm -f $MOUNT1/$tfile
+    replay_barrier mds
+    fail mds
+    kill -USR1 $pid1
+    wait $pid1 || return 3
+    kill -USR1 $pid2
+    wait $pid2 || return 4
+    [ -e $MOUNT1/$tfile ] && return 5
+    return 0
+}
+run_test 22 "double open|creat in replay with open orphan from two mntp"
+
 equals_msg `basename $0`: test complete, cleaning up
 SLEEP=$((`date +%s` - $NOW))
 [ $SLEEP -lt $TIMEOUT ] && sleep $SLEEP
