@@ -3,8 +3,8 @@
 set -e
 
 ONLY=${ONLY:-"$*"}
-# bug number for skipped test: 3192 12652 15528/3811 16929 9977 15528/11549 18080
-ALWAYS_EXCEPT="                14b  14c   19         22    28   29          35    $SANITYN_EXCEPT"
+# bug number for skipped test: 3192 12652  15528/3811 16929 9977 15528/11549  18080
+ALWAYS_EXCEPT="                14b  14c    19         22    28   29           35    $SANITYN_EXCEPT"
 # UPDATE THE COMMENT ABOVE WITH BUG NUMBERS WHEN CHANGING ALWAYS_EXCEPT!
 
 # bug number for skipped test:                                                    12652 12652
@@ -18,13 +18,11 @@ PATH=$PWD/$SRCDIR:$SRCDIR:$SRCDIR/../utils:$PATH
 
 SIZE=${SIZE:-40960}
 CHECKSTAT=${CHECKSTAT:-"checkstat -v"}
-CREATETEST=${CREATETEST:-createtest}
 GETSTRIPE=${GETSTRIPE:-lfs getstripe}
 SETSTRIPE=${SETSTRIPE:-lstripe}
 MCREATE=${MCREATE:-mcreate}
 OPENFILE=${OPENFILE:-openfile}
 OPENUNLINK=${OPENUNLINK:-openunlink}
-TOEXCL=${TOEXCL:-toexcl}
 TRUNCATE=${TRUNCATE:-truncate}
 export TMP=${TMP:-/tmp}
 MOUNT_2=${MOUNT_2:-"yes"}
@@ -45,7 +43,7 @@ init_test_env $@
 [ "$SLOW" = "no" ] && EXCEPT_SLOW="12 16"
 
 SANITYLOG=${TESTSUITELOG:-$TMP/$(basename $0 .sh).log}
-FAIL_ON_ERROR=false
+FAIL_ON_ERROR=${FAIL_ON_ERROR:-false}
 
 SETUP=${SETUP:-:}
 TRACE=${TRACE:-""}
@@ -131,7 +129,7 @@ run_test 2e "check chmod on root is propagated to others"
 
 test_3() {
 	( cd $DIR1 ; ln -s this/is/good $tfile )
-	[ "this/is/good" = "`perl -e 'print readlink("'$DIR2/$tfile'");'`" ] ||\
+	[ "this/is/good" = "`perl -e 'print readlink("'$DIR2/$tfile'");'`" ] ||
 		error "link $DIR2/$tfile not as expected"
 }
 run_test 3 "symlink on one mtpt, readlink on another ==========="
@@ -425,17 +423,17 @@ test_22() { # Bug 9926
 	cat $DIR2/d21/no_joined || error "cat error"
 	rm -rf $DIR2/d21/no_joined || error "unlink normal file error"
 }
-run_test 22 " After joining in one dir,  open/close unlink file in anther dir" 
+run_test 22 " After joining in one dir,  open/close unlink file in anther dir"
 
 test_23() { # Bug 5972
 	echo "others should see updated atime while another read" > $DIR1/f23
-	
+
 	# clear the lock(mode: LCK_PW) gotten from creating operation
 	cancel_lru_locks osc
-	
-	time1=`date +%s`	
+
+	time1=`date +%s`
 	sleep 2
-	
+
 	multiop_bg_pause $DIR1/f23 or20_c || return 1
 	MULTIPID=$!
 
@@ -452,7 +450,7 @@ test_23() { # Bug 5972
 }
 run_test 23 " others should see updated atime while another read===="
 
-test_24() {
+test_24a() {
 	touch $DIR1/$tfile
 	lfs df || error "lfs df failed"
 	lfs df -ih || error "lfs df -ih failed"
@@ -460,14 +458,21 @@ test_24() {
 	lfs df -i $DIR2 || error "lfs df -i $DIR2 failed"
 	lfs df $DIR1/$tfile || error "lfs df $DIR1/$tfile failed"
 	lfs df -ih $DIR2/$tfile || error "lfs df -ih $DIR2/$tfile failed"
-	
+
 	OSC=`lctl dl | awk '/-osc-|OSC.*MNT/ {print $4}' | head -n 1`
 	lctl --device %$OSC deactivate
 	lfs df -i || error "lfs df -i with deactivated OSC failed"
 	lctl --device %$OSC recover
 	lfs df || error "lfs df with reactivated OSC failed"
 }
-run_test 24 "lfs df [-ih] [path] test ========================="
+run_test 24a "lfs df [-ih] [path] test ========================="
+
+test_24b() {
+	touch $DIR1/$tfile
+	fsnum=`lfs df | grep -c "filesystem summary:"`
+	[ $fsnum -eq 2 ] || error "lfs df shows $fsnum != 2 filesystems."
+}
+run_test 24b "lfs df should show both filesystems ==============="
 
 test_25() {
 	[ `lctl get_param -n mdc.*-mdc-*.connect_flags | grep -c acl` -lt 2 ] && \
@@ -504,25 +509,25 @@ test_26b() {
         chmod a+x $DIR2/$tfile
         mt1=`stat -c %Y $DIR1/$tfile`
         mt2=`stat -c %Y $DIR2/$tfile`
-        
-        if [ x"$mt1" != x"$mt2" ]; then 
+
+        if [ x"$mt1" != x"$mt2" ]; then
                 error "not equal mtime, client1: "$mt1", client2: "$mt2"."
         fi
 }
 run_test 26b "sync mtime between ost and mds"
 
 test_27() {
-	cancel_lru_locks OSC
+	cancel_lru_locks osc
 	lctl clear
 	dd if=/dev/zero of=$DIR2/$tfile bs=$((4096+4))k conv=notrunc count=4 seek=3 &
 	DD2_PID=$!
 	sleep 0.050s
 	log "dd 1 started"
-	
+
 	dd if=/dev/zero of=$DIR1/$tfile bs=$((16384-1024))k conv=notrunc count=1 seek=4 &
 	DD1_PID=$!
 	log "dd 2 started"
-	
+
 	sleep 1
 	dd if=/dev/zero of=$DIR1/$tfile bs=8k conv=notrunc count=1 seek=0
 	log "dd 3 finished"
@@ -540,7 +545,7 @@ test_28() { # bug 9977
 	dd if=/dev/zero of=$DIR1/$tfile bs=1024k count=2
 
 	$LCTL --device $ostID destroy "${tOBJID}"
-    
+
 	# reading of 1st stripe should pass
 	dd if=$DIR2/$tfile of=/dev/null bs=1024k count=1 || error
 	# reading of 2nd stripe should fail (this stripe was destroyed)
@@ -593,7 +598,7 @@ run_test 31a "voluntary cancel / blocking ast race=============="
 
 test_31b() {
         remote_ost || { skip "local OST" && return 0; }
-	remote_ost_nodsh && skip "remote OST w/o dsh" && return 0
+        remote_ost_nodsh && skip "remote OST w/o dsh" && return 0
         mkdir -p $DIR1/$tdir || error "Creating dir $DIR1/$tdir"
         lfs setstripe $DIR/$tdir/$tfile -i 0 -c 1
         cp /etc/hosts $DIR/$tdir/$tfile
@@ -605,6 +610,8 @@ test_31b() {
         cat $DIR2/$tdir/$tfile > /dev/null 2>&1
         lctl set_param fail_loc=0
         do_facet ost1 lctl set_param fail_loc=0
+        # cleanup: reconnect the client back
+        df $DIR2
 }
 run_test 31b "voluntary OST cancel / blocking ast race=============="
 
@@ -616,10 +623,12 @@ enable_lockless_truncate() {
 test_32a() { # bug 11270
         local p="$TMP/sanityN-$TESTNAME.parameters"
         save_lustre_params $HOSTNAME llite.*.lockless_truncate > $p
+        rm -f $DIR1/$tfile
         cancel_lru_locks osc
-        clear_llite_stats
         enable_lockless_truncate 1
+        lfs setstripe -c -1 -s 1m $DIR1/$tfile
         dd if=/dev/zero of=$DIR1/$tfile count=10 bs=1M > /dev/null 2>&1
+        clear_llite_stats
 
         log "checking cached lockless truncate"
         $TRUNCATE $DIR1/$tfile 8000000
@@ -659,7 +668,7 @@ test_32b() { # bug 11270
                 save_lustre_params $node "ldlm.namespaces.filter-*.contention_seconds" >> $p
         done
         clear_llite_stats
-        # agressive lockless i/o settings 
+        # agressive lockless i/o settings
         for node in $(osts_nodes); do
                 do_node $node 'lctl set_param -n ldlm.namespaces.filter-*.max_nolock_bytes 2000000; lctl set_param -n ldlm.namespaces.filter-*.contended_locks 0; lctl set_param -n ldlm.namespaces.filter-*.contention_seconds 60'
         done
@@ -668,7 +677,7 @@ test_32b() { # bug 11270
                 dd if=/dev/zero of=$DIR1/$tfile bs=4k count=1 conv=notrunc > /dev/null 2>&1
                 dd if=/dev/zero of=$DIR2/$tfile bs=4k count=1 conv=notrunc > /dev/null 2>&1
         done
-        [ $(calc_llite_stats lockless_write_bytes) -ne 0 ] || error "lockless i/o was not triggered" 
+        [ $(calc_llite_stats lockless_write_bytes) -ne 0 ] || error "lockless i/o was not triggered"
         # disable lockless i/o (it is disabled by default)
         for node in $(osts_nodes); do
                 do_node $node 'lctl set_param -n ldlm.namespaces.filter-*.max_nolock_bytes 0; lctl set_param -n ldlm.namespaces.filter-*.contended_locks 32; lctl set_param -n ldlm.namespaces.filter-*.contention_seconds 0'
@@ -682,7 +691,7 @@ test_32b() { # bug 11270
                 dd if=/dev/zero of=$DIR2/$tfile bs=4k count=1 conv=notrunc > /dev/null 2>&1
         done
         [ $(calc_llite_stats lockless_write_bytes) -eq 0 ] ||
-                error "lockless i/o works when disabled" 
+                error "lockless i/o works when disabled"
         rm -f $DIR1/$tfile
         restore_lustre_params <$p
         rm -f $p
@@ -712,14 +721,12 @@ test_33() { #16129
                 echo writing on client1
                 dd if=/dev/zero of=$DIR1/$tfile count=100 conv=notrunc > /dev/null 2>&1
                 sync &
-                # wait for the flush
-                sleep 1
                 echo reading on client2
                 dd of=/dev/null if=$DIR2/$tfile > /dev/null 2>&1
                 # wait for a lock timeout
                 sleep 4
                 lock_out=$(do_nodes $(osts_nodes) "lctl get_param -n ldlm.namespaces.filter-*.lock_timeouts" | calc_sum)
-                if [ $OPER == "timeout" ] ; then 
+                if [ $OPER == "timeout" ] ; then
                         if [ $lock_in == $lock_out ]; then
                                 error "no lock timeout happened"
                         else
@@ -763,11 +770,11 @@ test_35() { # bug 17645
                 createmany -o $MOUNT2/$tfile/a 4000 &
                 pid1=$!
                 sleep 1
-        
+
                 # Let's make conflict and bl_ast
                 ls -la $MOUNT1/$tfile > /dev/null &
                 pid2=$!
-                
+
                 log "Wait for $pid1 $pid2 for $timeout sec..."
                 sleep $timeout
                 kill -9 $pid1 $pid2 > /dev/null 2>&1
@@ -794,26 +801,36 @@ run_test 35 "-EINTR cp_ast vs. bl_ast race does not evict client"
 
 test_36() { #bug 16417
     local SIZE
-    mkdir -p $MOUNT1/$tdir
-    lfs setstripe -c -1 $MOUNT1/$tdir
+    local SIZE_B
+    local i
+
+    mkdir -p $DIR1/$tdir
+    $LFS setstripe -c -1 $DIR1/$tdir
     i=0
-    SIZE=100
+    SIZE=50
+    let SIZE_B=SIZE*1024*1024
 
     while [ $i -le 10 ]; do
-	lctl mark "start test"
-	before=$($LFS df | awk '{if ($1 ~/^filesystem/) {print $5; exit} }')
-	dd if=/dev/zero of=$MOUNT1/$tdir/file000 bs=1M count=$SIZE
-	dd if=$MOUNT2/$tdir/file000 of=/dev/null bs=1M count=$SIZE &
-	read_pid=$!
-	sleep 0.1
-	rm -f $MOUNT1/$tdir/file000
-	wait $read_pid
-	after=$($LFS df | awk '{if ($1 ~/^filesystem/) {print $5; exit} }')
-	if [ $before -gt $after ]; then
-	    error "space leaked"
-	    exit;
-	fi
-	let i=i+1
+        lctl mark "start test"
+        local before=$($LFS df | awk '{if ($1 ~/^filesystem/) {print $5; exit} }')
+        dd if=/dev/zero of=$DIR1/$tdir/file000 bs=1M count=$SIZE
+        sync
+        sleep 1
+        local after_dd=$($LFS df | awk '{if ($1 ~/^filesystem/) {print $5; exit} }')
+        multiop_bg_pause $DIR2/$tdir/file000 O_r${SIZE_B}c || return 3
+        read_pid=$!
+        rm -f $DIR1/$tdir/file000
+        kill -USR1 $read_pid
+        wait $read_pid
+        sleep 1
+        local after=$($LFS df | awk '{if ($1 ~/^filesystem/) {print $5; exit} }')
+        echo "*** cycle($i) *** before($before):after_dd($after_dd):after($after)"
+        # this free space! not used
+        if [ $after_dd -ge $after ]; then
+            error "space leaked"
+            return 1;
+        fi
+        let i=i+1
     done
 }
 run_test 36 "handle ESTALE/open-unlink corectly"
@@ -829,10 +846,66 @@ test_37() { # bug 18695
 	kill -USR1 $MULTIPID
 	nr_files=`lfs find $DIR1/$tdir -type f | wc -l`
 	[ $nr_files -eq 10000 ] || error "$nr_files != 10000 truncated directory?"
-
 }
 run_test 37 "check i_size is not updated for directory on close (bug 18695) =============="
 
+test_38() { # bug 18801, based on the code of test_32b
+        remote_ost_nodsh && skip "remote OST with nodsh" && return
+
+        local node
+        local p="$TMP/sanityN-$TESTNAME.parameters"
+        local random="$TMP/sanityN-$TESTNAME.random"
+        # 1. locked unaligned non-DIRECT_IO write of 8192 bytes to file A
+        # 2a. locked unaligned DIRECT_IO write of 4000 bytes to file B
+        # 2b. locked unaligned DIRECT_IO write of 4000 bytes to file B
+        # 3. unaligned "lockless DIRECT_IO" write of 192 bytes in the end of file B
+        # 4. compare A and B
+        log "creating the initial file"
+        multiop $random Ob4000b4000b192c || error "failed creating random file"
+        log "creating a file with the same contents"
+        multiop $DIR1/$tfile oO_CREAT:O_DIRECT:O_RDWR:b4000c || error "first multiop failed"
+        multiop $DIR1/$tfile oO_CREAT:O_DIRECT:O_RDWR:z4000b4000c || error "second multiop failed"
+        save_lustre_params $HOSTNAME "llite.*.contention_seconds" > $p
+        for node in $(osts_nodes); do
+                save_lustre_params $node "ldlm.namespaces.filter-*.max_nolock_bytes" >> $p
+                save_lustre_params $node "ldlm.namespaces.filter-*.contended_locks" >> $p
+                save_lustre_params $node "ldlm.namespaces.filter-*.contention_seconds" >> $p
+        done
+        log "enforcing lockless I/O"
+        clear_llite_stats
+        # agressive lockless i/o settings
+        for node in $(osts_nodes); do
+                do_node $node 'lctl set_param -n ldlm.namespaces.filter-*.max_nolock_bytes 2000000; lctl set_param -n ldlm.namespaces.filter-*.contended_locks 0; lctl set_param -n ldlm.namespaces.filter-*.contention_seconds 60'
+        done
+        lctl set_param -n llite.*.contention_seconds 60
+        multiop $DIR2/$tfile oO_DIRECT:O_RDWR:z8000b192c || error "the last multiop failed"
+        [ $(calc_llite_stats lockless_write_bytes) -ne 0 ] || error "lockless i/o was not triggered"
+        restore_lustre_params <$p
+        log "comparing"
+        cmp $DIR1/$tfile $random || error "O_DIRECT+lockless results do not match the original file"
+        rm -f $DIR1/$tfile
+        rm -f $p
+        rm -f $random
+}
+run_test 38 "lockless i/o with O_DIRECT and unaligned writes"
+
+test_39() {
+        local originaltime
+        local updatedtime
+        local delay=3
+
+        touch $DIR1/$tfile
+        originaltime=$(stat -c %Y $DIR1/$tfile)
+        log "original modification time is $originaltime"
+        sleep $delay
+        multiop $DIR1/$tfile oO_DIRECT:O_WRONLY:w$((10*1048576))c || error "multiop has failed"
+        updatedtime=$(stat -c %Y $DIR2/$tfile)
+        log "updated modification time is $updatedtime"
+        [ $((updatedtime - originaltime)) -ge $delay ] || error "invalid modification time"
+        rm -rf $DIR/$tfile
+}
+run_test 39 "direct I/O writes should update mtime ========="
+
 log "cleanup: ======================================================"
 
 check_and_cleanup_lustre
