@@ -64,7 +64,6 @@
 static char cmdname[512];
 static char target[64] = "";
 char nid[64] = "";
-static int live_threads = 0;
 static int sig_received = 0;
 static int o_verbose = 4; /* 0-5 */
 static int my_oss = 0;
@@ -78,7 +77,7 @@ static int jt_quit(int argc, char **argv) {
 static int loadgen_usage(int argc, char **argv)
 {
         if (argc == 1) {
-                fprintf(stderr, 
+                fprintf(stderr,
         "This is a test program used to simulate large numbers of\n"
         "clients.  The echo obds are used, so the obdecho module must\n"
         "be loaded.\n"
@@ -145,7 +144,9 @@ struct kid_t {
         int              k_dev;
 };
 
+static int live_threads = 0;
 static struct kid_t *kid_list = NULL;
+pthread_mutex_t m_list = PTHREAD_MUTEX_INITIALIZER;
 
 static struct kid_t *push_kid(int tnum)
 {
@@ -156,10 +157,12 @@ static struct kid_t *push_kid(int tnum)
                 return NULL;
         }
         kid->k_pthread = pthread_self();
+        pthread_mutex_lock(&m_list);
         kid->k_next = kid_list;
         kid->k_id = tnum;
         kid_list = kid;
         live_threads++;
+        pthread_mutex_unlock(&m_list);
         return kid;
 }
 
@@ -690,12 +693,14 @@ static void *run_one_child(void *threadvp)
 
         while(!(rc || sig_received)) {
                 pthread_mutex_lock(&m_trigger);
+    		pthread_mutex_lock(&m_list);
                 waiting_count++;
                 if ((waiting_count == live_threads) && timer_on) {
                         report_perf();
                         timer_on = 0;
                         all_done = 1;
                 }
+    		pthread_mutex_unlock(&m_list);
                 pthread_cond_wait(&cv_trigger, &m_trigger);
                 waiting_count--;
                 all_done = 0;
@@ -731,6 +736,19 @@ out:
         pthread_exit((void *)(long)rc);
 }
 
+/* 
+ * PTHREAD_STACK_MIN is 16K minimal stack for threads. This
+ * is stack consumed by one thread, which executes NULL procedure.
+ * We need some more here and 20k stack for one client thread
+ * is enough to not overflow. In same time it does not consume
+ * a lot of memory for large number of threads.
+ *
+ * 20K virtual clients will only consume 320M + 400M. Still to
+ * create this number of virtual clients we need to fix 8192
+ * OBDs limit.
+ */
+#define CLIENT_THREAD_STACK_SIZE (PTHREAD_STACK_MIN + (20 * 1024))
+
 static int loadgen_start_clients(int argc, char **argv)
 {
         int rc = 0, i, numt;
@@ -756,7 +774,7 @@ static int loadgen_start_clients(int argc, char **argv)
                         cmdname, rc, strerror(errno));
                 return -errno;
         }
-        pthread_attr_setstacksize (&attr, PTHREAD_STACK_MIN);
+        pthread_attr_setstacksize (&attr, CLIENT_THREAD_STACK_SIZE);
         pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);
 
         numt += live_threads;
@@ -861,13 +879,17 @@ static int loadgen_write(int argc, char **argv)
         if (argc < 3 || argc > 4)
                 return CMD_HELP;
         threads = atoi(argv[1]);
+        pthread_mutex_lock(&m_list);
         if (threads > live_threads) {
+    		pthread_mutex_unlock(&m_list);
                 fprintf(stderr, "requested %d threads but only %d are running. "
                         "Use 'start' to start some more.\n",
                         threads, live_threads);
                 return -EOVERFLOW;
+        } else {
+    		pthread_mutex_unlock(&m_list);
         }
-        trigger(C_WRITE, threads, atoi(argv[2]), 
+        trigger(C_WRITE, threads, atoi(argv[2]),
                 (argc == 4) ? atoi(argv[3]) : 0);
         return 0;
 }
@@ -916,7 +938,7 @@ static int loadgen_start_echosrv(int argc, char **argv)
                         cmdname, rc);
                 goto clean;
         }
-  
+
         /* Create an OSS to handle the communications */
         /* attach ost OSS OSS_UUID */
         args[1] = "ost";
@@ -955,7 +977,7 @@ clean:
 
 static int loadgen_wait(int argc, char **argv)
 {
-        /* Give scripts a chance to start some threads */   
+        /* Give scripts a chance to start some threads */
         sleep(1);
         while (!all_done) {
                 sleep(1);
@@ -965,7 +987,7 @@ static int loadgen_wait(int argc, char **argv)
 
 static int loadgen_init(int argc, char **argv)
 {
-        char *args[3];
+        char *args[4];
         int rc;
 
         sprintf(cmdname, "%s", argv[0]);
@@ -995,7 +1017,7 @@ static int loadgen_init(int argc, char **argv)
 static int loadgen_exit()
 {
         int rc;
-        
+
         printf("stopping %d children\n", live_threads);
         kill_kids();
         rc = wait_for_threads();
@@ -1037,7 +1059,7 @@ static int loadgen_main(int argc, char **argv)
 
 out:
         obd_finalize(argc, argv);
-        return rc;
+        return rc < 0 ? -rc : rc;
 }
 
 #ifndef LIBLUSTRE_TEST
