@@ -39,6 +39,7 @@
  * Author: Nathan Rutman <nathan@clusterfs.com>
  */
 
+
 #ifndef _GNU_SOURCE
 #define _GNU_SOURCE
 #endif
@@ -60,6 +61,7 @@
 
 #define MAX_HW_SECTORS_KB_PATH  "queue/max_hw_sectors_kb"
 #define MAX_SECTORS_KB_PATH     "queue/max_sectors_kb"
+#define STRIPE_CACHE_SIZE       "md/stripe_cache_size"
 #define MAX_RETRIES 99
 
 int          verbose = 0;
@@ -67,6 +69,7 @@ int          nomtab = 0;
 int          fake = 0;
 int          force = 0;
 int          retry = 0;
+int          md_stripe_cache_size = 2048;
 char         *progname = NULL;
 
 void usage(FILE *out)
@@ -82,7 +85,7 @@ void usage(FILE *out)
                 "\t<filesystem>: name of the Lustre filesystem (e.g. lustre1)\n"
                 "\t<mountpt>: filesystem mountpoint (e.g. /mnt/lustre)\n"
                 "\t-f|--fake: fake mount (updates /etc/mtab)\n"
-                "\t--force: force mount even if already in /etc/mtab\n"
+                "\t-o force|--force: force mount even if already in /etc/mtab\n"
                 "\t-h|--help: print this usage message\n"
                 "\t-n|--nomtab: do not update /etc/mtab after mount\n"
                 "\t-v|--verbose: print verbose config settings\n"
@@ -93,6 +96,8 @@ void usage(FILE *out)
                 "\t\texclude=<ostname>[:<ostname>] : colon-separated list of "
                 "inactive OSTs (e.g. lustre-OST0001)\n"
                 "\t\tretry=<num>: number of times mount is retried by client\n"
+                "\t\tmd_stripe_cache_size=<num>: set the raid stripe cache "
+                "size for the underlying raid if present\n"
                 );
         exit((out != stdout) ? EINVAL : 0);
 }
@@ -161,6 +166,11 @@ static char *convert_hostnames(char *s1)
         lnet_nid_t nid;
 
         converted = malloc(left);
+        if (converted == NULL) {
+                fprintf(stderr, "out of memory: needed %d bytes\n",
+                        MAXNIDSTR);
+                return NULL;
+        }
         c = converted;
         while ((left > 0) && (*s1 != '/')) {
                 s2 = strpbrk(s1, ",:");
@@ -272,14 +282,22 @@ int parse_options(char *orig_options, int *flagp)
                  * manner */
                 arg = opt;
                 val = strchr(opt, '=');
-                if (val != NULL && strncmp(arg, "retry", 5) == 0) {
+                /* please note that some ldiskfs mount options are also in the form
+                 * of param=value. We should pay attention not to remove those
+                 * mount options, see bug 22097. */
+                if (val && strncmp(arg, "md_stripe_cache_size", 20) == 0) {
+                        md_stripe_cache_size = atoi(val + 1);
+                } else if (val && strncmp(arg, "retry", 5) == 0) {
                         retry = atoi(val + 1);
                         if (retry > MAX_RETRIES)
                                 retry = MAX_RETRIES;
                         else if (retry < 0)
                                 retry = 0;
-                }
-                else if (parse_one_option(opt, flagp) == 0) {
+                } else if (strncmp(opt, "force", 5) == 0) {
+                        //XXX special check for 'force' option
+                        ++force;
+                        printf("force: %d\n", force);
+                } else if (parse_one_option(opt, flagp) == 0) {
                         /* pass this on as an option */
                         if (*options)
                                 strcat(options, ",");
@@ -300,7 +318,12 @@ int read_file(char *path, char *buf, int size)
         if (fd == NULL)
                 return errno;
 
-        fgets(buf, size, fd);
+        /* should not ignore fgets(3)'s return value */
+        if (!fgets(buf, size, fd)) {
+                fprintf(stderr, "reading from %s: %s", path, strerror(errno));
+                fclose(fd);
+                return 1;
+        }
         fclose(fd);
         return 0;
 }
@@ -321,12 +344,12 @@ int write_file(char *path, char *buf)
 /* This is to tune the kernel for good SCSI performance.
  * For that we set the value of /sys/block/{dev}/queue/max_sectors_kb
  * to the value of /sys/block/{dev}/queue/max_hw_sectors_kb */
-int set_tunables(char *source, int src_len)
+int set_blockdev_tunables(char *source)
 {
         glob_t glob_info;
         struct stat stat_buf;
         char *chk_major, *chk_minor;
-        char *savept, *dev, *s2 = 0;
+        char *savept = NULL, *dev;
         char *ret_path;
         char buf[PATH_MAX] = {'\0'}, path[PATH_MAX] = {'\0'};
         char real_path[PATH_MAX] = {'\0'};
@@ -344,52 +367,32 @@ int set_tunables(char *source, int src_len)
                 return -EINVAL;
         }
 
-        src_len = sizeof(real_path);
-
         if (strncmp(real_path, "/dev/loop", 9) == 0)
                 return 0;
 
-        if ((real_path[0] != '/') && ((s2 = strpbrk(real_path, ",:")) != NULL))
+        if ((real_path[0] != '/') && (strpbrk(real_path, ",:") != NULL))
                 return 0;
 
-        dev = real_path + src_len - 1;
-        while (dev > real_path && (*dev != '/')) {
-                if (isdigit(*dev))
-                        *dev = 0;
-                dev--;
-        }
-        snprintf(path, sizeof(path), "/sys/block%s/%s", dev,
-                 MAX_HW_SECTORS_KB_PATH);
-        rc = read_file(path, buf, sizeof(buf));
-        if (rc == 0 && (strlen(buf) - 1) > 0) {
-                snprintf(path, sizeof(path), "/sys/block%s/%s", dev,
-                         MAX_SECTORS_KB_PATH);
-                rc = write_file(path, buf);
-                if (rc && verbose)
-                        fprintf(stderr, "warning: opening %s: %s\n",
-                                path, strerror(errno));
-                return rc;
-        }
-
-        if (rc != ENOENT)
-                return rc;
+        snprintf(path, sizeof(path), "/sys/block%s", real_path + 4);
+        if (access(path, X_OK) == 0)
+                goto set_params;
 
         /* The name of the device say 'X' specified in /dev/X may not
          * match any entry under /sys/block/. In that case we need to
          * match the major/minor number to find the entry under
          * sys/block corresponding to /dev/X */
-        dev = real_path + src_len - 1;
-        while (dev > real_path) {
-                if (isdigit(*dev))
-                        *dev = 0;
-                dev--;
-        }
+        dev = real_path + strlen(real_path);
+        while (--dev > real_path && isdigit(*dev))
+                *dev = 0;
 
-        rc = stat(dev, &stat_buf);
+        if (strncmp(real_path, "/dev/md_", 8) == 0)
+                *dev = 0;
+
+        rc = stat(real_path, &stat_buf);
         if (rc) {
                 if (verbose)
                         fprintf(stderr, "warning: %s, device %s stat failed\n",
-                                strerror(errno), dev);
+                                strerror(errno), real_path);
                 return rc;
         }
 
@@ -423,31 +426,59 @@ int set_tunables(char *source, int src_len)
                 if (verbose)
                         fprintf(stderr,"warning: device %s does not match any "
                                 "entry under /sys/block\n", real_path);
-                rc = -EINVAL;
-                goto out;
+                globfree(&glob_info);
+                return -EINVAL;
         }
 
-        snprintf(path, sizeof(path), "%s/%s", glob_info.gl_pathv[i],
+        /* Chop off "/dev" from path we found */
+        path[strlen(glob_info.gl_pathv[i])] = '\0';
+        globfree(&glob_info);
+
+set_params:
+        if (strncmp(real_path, "/dev/md", 7) == 0) {
+                snprintf(real_path, sizeof(real_path), "%s/%s", path,
+                         STRIPE_CACHE_SIZE);
+
+                rc = read_file(real_path, buf, sizeof(buf));
+                if (rc) {
+                        if (verbose)
+                                fprintf(stderr, "warning: opening %s: %s\n",
+                                        real_path, strerror(errno));
+                        return rc;
+                }
+
+                if (atoi(buf) >= md_stripe_cache_size)
+                        return 0;
+
+                if (strlen(buf) - 1 > 0) {
+                        snprintf(buf, sizeof(buf), "%d", md_stripe_cache_size);
+                        rc = write_file(real_path, buf);
+                        if (rc && verbose)
+                                fprintf(stderr, "warning: opening %s: %s\n",
+                                        real_path, strerror(errno));
+                }
+                /* Return since raid and disk tunables are different */
+                return rc;
+        }
+
+        snprintf(real_path, sizeof(real_path), "%s/%s", path,
                  MAX_HW_SECTORS_KB_PATH);
-        rc = read_file(path, buf, sizeof(buf));
+        rc = read_file(real_path, buf, sizeof(buf));
         if (rc) {
                 if (verbose)
                         fprintf(stderr, "warning: opening %s: %s\n",
-                                path, strerror(errno));
-                goto out;
+                                real_path, strerror(errno));
+                return rc;
         }
 
         if (strlen(buf) - 1 > 0) {
-                snprintf(path, sizeof(path), "%s/%s",
-                         glob_info.gl_pathv[i], MAX_SECTORS_KB_PATH);
-                rc = write_file(path, buf);
+                snprintf(real_path, sizeof(real_path), "%s/%s", path,
+                         MAX_SECTORS_KB_PATH);
+                rc = write_file(real_path, buf);
                 if (rc && verbose)
                         fprintf(stderr, "warning: writing to %s: %s\n",
-                                path, strerror(errno));
+                                real_path, strerror(errno));
         }
-
-out:
-        globfree(&glob_info);
         return rc;
 }
 
@@ -513,7 +544,15 @@ int main(int argc, char *const argv[])
         }
 
         usource = argv[optind];
+        if (!usource) {
+                usage(stderr);
+        }
+
         source = convert_hostnames(usource);
+        if (!source) {
+                usage(stderr);
+        }
+
         target = argv[optind + 1];
         ptr = target + strlen(target) - 1;
         while ((ptr > target) && (*ptr == '/')) {
@@ -521,19 +560,18 @@ int main(int argc, char *const argv[])
                 ptr--;
         }
 
-        if (!usource || !source) {
-                usage(stderr);
-        }
-
         if (verbose) {
                 for (i = 0; i < argc; i++)
                         printf("arg[%d] = %s\n", i, argv[i]);
-                printf("source = %s (%s), target = %s\n",
-                       usource, source, target);
+                printf("source = %s (%s), target = %s\n", usource, source, target);
                 printf("options = %s\n", orig_options);
         }
 
         options = malloc(strlen(orig_options) + 1);
+        if (options == NULL) {
+                fprintf(stderr, "can't allocate memory for options\n");
+                return -1;
+        }
         strcpy(options, orig_options);
         rc = parse_options(options, &flags);
         if (rc) {
@@ -572,6 +610,10 @@ int main(int argc, char *const argv[])
            functions.  So we'll stick it on the end of the options. */
         optlen = strlen(options) + strlen(",device=") + strlen(source) + 1;
         optcopy = malloc(optlen);
+        if (optcopy == NULL) {
+                fprintf(stderr, "can't allocate memory to optcopy\n");
+                return -1;
+        }
         strcpy(optcopy, options);
         if (*optcopy)
                 strcat(optcopy, ",");
@@ -582,11 +624,12 @@ int main(int argc, char *const argv[])
                 printf("mounting device %s at %s, flags=%#x options=%s\n",
                        source, target, flags, optcopy);
 
-        if (!strstr(usource, ":/") && set_tunables(source, strlen(source)) &&
-            verbose)
-                fprintf(stderr, "%s: unable to set tunables for %s"
+        if (!strstr(usource, ":/") && set_blockdev_tunables(source)) {
+                if (verbose)
+                        fprintf(stderr, "%s: unable to set tunables for %s"
                                 " (may cause reduced IO performance)\n",
                                 argv[0], source);
+        }
 
         register_service_tags(usource, source, target);
 
@@ -669,8 +712,12 @@ int main(int argc, char *const argv[])
                 /* May as well try to clean up loop devs */
                 if (strncmp(usource, "/dev/loop", 9) == 0) {
                         char cmd[256];
+                        int ret;
                         sprintf(cmd, "/sbin/losetup -d %s", usource);
-                        system(cmd);
+                        if ((ret = system(cmd)) < 0)
+                                rc = errno;
+                        else if (ret > 0)
+                                rc = WEXITSTATUS(ret);
                 }
 
         } else if (!nomtab) {
